/************************************************************
 * QRNA - Comparative analysis of biological sequences 
 *        with pair hidden Markov models and pair stochastic context-free grammars
 * Copyright (C) 2000-Howard Hughes Medical Institute
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* codcreatealign.c
 *
 * ER, Wed Jun  2 09:23:28 CDT 1999 [St. Louis]
 * 
 * creates an alignment acoding to the codingmodel
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

#include "funcs.h"
#include "globals.h"
#include "squid.h"
#include "structs.h"

#ifdef MEMDEBUG
#include "dbmalloc.h"
#endif

static void pass_thru_COJ(FILE *ofp, SQINFO *sqinfoX, int *seqX, SQINFO *sqinfoY, int *seqY, int start, 
			  int *ret_nxt_pos, struct othmodel_s *oth, struct ali_s *ali, int traceback, 
			  int alignment, char *string_name);

/* Function: SimulateCodingSequences()
 * Date:     SRE, Thu Jun 11 10:06:28 1998 [St. Louis]
 *
 * Purpose:  Simulate an ungapped alignment of two coding
 *           sequences, given a pammodel
 *
 * Args:     L        - length of the alignment; will be made next lowest multiple of three
 *           pammodel - model, [64][64] joint probabilities
 *           ret_s1   - RETURN: first seq
 *           ret_s2   - RETURN: second seq
 *
 * Returns:  void
 */
void
SimulateCodingSequences(int L, double pammodel[64][64], char **ret_s1, char **ret_s2)
{
  char  *s1, *s2;
  int    pos;
  double p[4096];
  int    x,y, i;

  L = (L / 3) * 3;  /* make a multiple of three */

  s1 = MallocOrDie(sizeof(char) * (L+1));
  s2 = MallocOrDie(sizeof(char) * (L+1));
  
  for (i = 0; i < 4096; i++) 
    {
      x = i / 64;
      y = i % 64;
      p[i] = pammodel[x][y];
    }

  for (pos = 0; pos < L; pos += 3)
    {
      i = DChoose(p, 4096);
      x = i / 64;
      y = i % 64;

      s1[pos]   = DNAAlphabet[x/16];
      s1[pos+1] = DNAAlphabet[(x%16)/4];
      s1[pos+2] = DNAAlphabet[x%4];

      s2[pos]   = DNAAlphabet[y/16];
      s2[pos+1] = DNAAlphabet[(y%16)/4];
      s2[pos+2] = DNAAlphabet[y%4];
    }
  s1[L] = '\0';
  s2[L] = '\0';
  *ret_s1 = s1;
  *ret_s2 = s2;
  return;
}


/* Function:  SimulateCODAlign()
 * Date:      ER, Mon Feb  7 10:55:52 CST 2000 [St. Louis]
 *
 * Purpose:  Simulate an aligment of two sequences
 *           related by the COD model.
 *
 * Args:     s1, s2    - the sequences 
 *           L         - length of aligmnemt
 *           othmodel - the null model
 *
 * Returns:  (void)
 */
void
SimulateCODAlign(FILE *ofp, SQINFO *sqinfoX, int *seqX, SQINFO *sqinfoY, int *seqY, int start, 
		 int L, struct codmodel_s *cod, struct nullmodel_s *null, struct ali_s *ali, 
		 int traceback, int alignment, char *string_name)
{
  struct tracer_s      *tr;      /* the traceback tree under construction  */
  struct tracer_s      *cur_tr;  /* ptr to node of tr we're working on     */
  struct tracerstack_s *dolist;  /* pushdown stack of active tr nodes      */

  double pcod[4096];
  double pcod_hexa[4096];
  double pcod30[64];
  double pcod40[256];
  double pcod04[256];
  double pcod32[1024];
  double pcod34[16384];
  double pcod44[65536];
  double pcod42[4096];

  double hexa, extra_hexamer;
  double sum;
  int    flag;

  int    cur_x,   cur_y;         /* nucleotides at those positions         */
  int    cur_st,  nxt_st;
  int    cur_pos, nxt_pos;
  int    pick;
  int    i;
  int    x, y;
  int    x1, x2, x3, x4;
  int    y1, y2, y3, y4;
  int    xx1, xx2, xx3;
  int    yy1, yy2, yy3;

  for (i = 0; i < 4096; i++) 
    {
      x = i / 64;
      y = i % 64;
      pcod[i] = cod->pcodon[x][y];
    }

  for (i = 0; i < 64; i++) {
    pcod30[i] = average_codon(i, cod, null, FALSE);
  }

  for (i = 0; i < 256; i++) {
    pcod40[i] = average_codon(i/4, cod, null, FALSE) + null->xem[i%4];
    pcod04[i] = average_codon(i/4, cod, null, FALSE) + null->yem[i%4];
  }

  for (i = 0; i < 1024; i++) {
    x = i/16;
    y = i%16;

    pcod32[i] = average(3, y/4, y%4, x, cod, null, FALSE);
  }

  for (i = 0; i < 16384; i++) 
    {
      x = i / 256;
      y = i % 256;

      pcod34[i] = cod->pcodon[x][y/4] + null->yem[y%4];
    }

  for (i = 0; i < 65536; i++) 
    {
      x = i / 256;
      y = i % 256;
      pcod44[i] = cod->pcodon[x/4][y/4] + null->xem[x%4] + null->yem[y%4];
    }

  for (i = 0; i < 4096; i++) 
    {
      x = i / 16;
      y = i % 16;

      pcod42[i] = pcod32[(x/4)*16+y] + null->xem[x%4];
    }


 /* Initialize
   * Start at pos "0" with state "stB"
   */
  tr     = InitTracer();       /* start a trace tree */
  dolist = InitTracerstack();  /* start a stack for traversing the trace tree */
  
  cur_tr = AttachTracer(tr, start, stCb); 
  PushTracerstack(dolist, cur_tr);

  while ((cur_tr = PopTracerstack(dolist)) != NULL)
    {
      cur_st = cur_tr->type;
      cur_pos = cur_tr->emit;

      xx1 = (cur_pos > 2)? seqX[cur_pos - 3] : 4;
      xx2 = (cur_pos > 1)? seqX[cur_pos - 2] : 4;
      xx3 = (cur_pos > 0)? seqX[cur_pos - 1] : 4;

      yy1 = (cur_pos > 2)? seqY[cur_pos - 3] : 4;
      yy2 = (cur_pos > 1)? seqY[cur_pos - 2] : 4;
      yy3 = (cur_pos > 0)? seqY[cur_pos - 1] : 4;

      sum = 0.0;
      for (i = 0; i < 4096; i++) {
	hexa = 0.0;
	flag = 0;

	if ( !isitagap(xx1) && !isitagap(xx2) && !isitagap(xx3) ) 
	  {
	    hexa += EXP2(cod->phexa[CODON(xx1,xx2,xx3)][i/64]);
	    flag ++;
	  }

	if ( !isitagap(yy1) && !isitagap(yy2) && !isitagap(yy3) )
	  {
	    hexa += EXP2(cod->phexa[CODON(yy1,yy2,yy3)][i%64]);
	    flag ++;
	  }
	
	switch (flag) 
	  {
	  case 0:  extra_hexamer =  0.0;              break;
	  case 1:  extra_hexamer =        LOG2(hexa); break;
	  case 2:  extra_hexamer = -1.0 + LOG2(hexa); break;
	  default: Die ("impossible! flag = %d\n", flag);
	  }
	
	pcod_hexa[i] = pcod[i] + extra_hexamer;

	sum += EXP2(pcod_hexa[i]);
      }
      if (sum > 1.01 || sum < 0.99) Die ("pcod_hexa probs do not add up to one, %f", sum);
      
      switch (cur_st){
      case stCb:
	cur_x = -1;
	cur_y = -1;
	nxt_pos = cur_pos;

	/* do not allow any possible frameshit at the very begining
	 */
	if (cur_pos == 0) 
	  pick = DENChoose(cod->t + IdxTransCOD[cur_st], TransPerCODState[cur_st]-1);
	else 
	  pick = DENChoose(cod->t + IdxTransCOD[cur_st], TransPerCODState[cur_st]);

	nxt_st = cur_st + 1 + pick;
	break;

      case stCe:
	/* passing thru COJ may introduce additional frame shifts. Not sure whether the 
	 * model can handle so much flexibility.
	 *
	 * (March 19, 2000) I have reduced the probabilisties of c34,c43,.. 
	 * (ie all the possible transitions that produce frame shifts). 
	 * If those frameshiter parameters are set too high,
	 * the model picks condons from everywhere in random alignments.
	 *
	 * Also, ScoreWithCOD() now also passsed thru COJ without allowing any emissions,
	 * so it is only fair not to allow any COJ emission here.
	 *
	pass_thru_COJ(ofp, sqinfoX, seqX, sqinfoY, seqY, cur_pos, &nxt_pos, cod->COJ, ali, 
		      FALSE, FALSE, string_name);
	 */
	nxt_pos = cur_pos;
	if (nxt_pos < start+L) nxt_st = stCb;
	else nxt_st = -1;
	break;

      case stC33:
	i = DLog2Choose(pcod_hexa, 4096);

	cur_x = i / 64;
	cur_y = i % 64;

	x1 = cur_x/16;
	x2 = (cur_x%16)/4;
	x3 = cur_x%4;
	
	y1 = cur_y/16;
	y2 = (cur_y%16)/4;
	y3 = cur_y%4;
	
	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	
	ali->charX[cur_pos]   = DNAAlphabet[x1];
	ali->charX[cur_pos+1] = DNAAlphabet[x2];
	ali->charX[cur_pos+2] = DNAAlphabet[x3];
	
	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = DNAAlphabet[y3];

	nxt_pos = cur_pos + 3;
	nxt_st = stCe;
	break;

      case stC34:
	i = DLog2Choose(pcod34, 16384);
	
	cur_x = i / 256;
	cur_y = i % 256;
	
	x1 = cur_x/16;
	x2 = (cur_x%16)/4;
	x3 = cur_x%4;
	x4 = 4;
	
	y1 = (cur_y/4)/16;
	y2 = ((cur_y/4)%16)/4;
	y3 = (cur_y/4)%4;
	y4 = cur_y%4;
	
	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	seqX[cur_pos+3] = x4;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	seqY[cur_pos+3] = y4;
	
	ali->charX[cur_pos]   = DNAAlphabet[x1];
	ali->charX[cur_pos+1] = DNAAlphabet[x2];
	ali->charX[cur_pos+2] = DNAAlphabet[x3];
	ali->charX[cur_pos+3] = '.';
	
	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = DNAAlphabet[y3];
	ali->charY[cur_pos+3] = DNAAlphabet[y4];

	nxt_pos = cur_pos + 4;
	nxt_st = stCe;
	break;

      case stC32:
	i = DLog2Choose(pcod32, 1024);

	cur_x = i / 16;
	cur_y = i % 16;

	x1 = cur_x/16;
	x2 = (cur_x%16)/4;
	x3 = cur_x%4;
	
	y1 = cur_y/4;
	y2 = cur_y%4;
	y3 = 4;
	
	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	
	ali->charX[cur_pos]   = DNAAlphabet[x1];
	ali->charX[cur_pos+1] = DNAAlphabet[x2];
	ali->charX[cur_pos+2] = DNAAlphabet[x3];
	
	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = '.';

	nxt_pos = cur_pos + 3;
	nxt_st = stCe;
	break;

      case stC43:
	i = DLog2Choose(pcod34, 16384);
	
	cur_x = i % 256;
	cur_y = i / 256;
	
	x1 = (cur_x/4)/16;
	x2 = ((cur_x/4)%16)/4;
	x3 = (cur_x/4)%4;
	x4 = cur_x%4;

	y1 = cur_y/16;
	y2 = (cur_y%16)/4;
	y3 = cur_y%4;
	y4 = 4;

	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	seqX[cur_pos+3] = x4;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	seqY[cur_pos+3] = y4;
	
	ali->charX[cur_pos]   = DNAAlphabet[x1];
	ali->charX[cur_pos+1] = DNAAlphabet[x2];
	ali->charX[cur_pos+2] = DNAAlphabet[x3];
	ali->charX[cur_pos+3] = DNAAlphabet[x4];
	
	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = DNAAlphabet[y3];
	ali->charY[cur_pos+3] = '.';

	nxt_pos = cur_pos + 4;
	nxt_st = stCe;
	break;

       case stC44:
	i = DLog2Choose(pcod44, 65536);

	cur_x = i / 256;
	cur_y = i % 256;

	x = cur_x/4;
	y = cur_y/4;

	x1 = x/16;
	x2 = (x%16)/4;
	x3 = x%4;
	x4 = cur_x%4;
	
	y1 = y/16;
	y2 = (y%16)/4;
	y3 = y%4;
	y4 = cur_y%4;
	
	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	seqX[cur_pos+3] = x4;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	seqY[cur_pos+3] = y4;
	
	ali->charX[cur_pos]   = DNAAlphabet[x1];
	ali->charX[cur_pos+1] = DNAAlphabet[x2];
	ali->charX[cur_pos+2] = DNAAlphabet[x3];
	ali->charX[cur_pos+3] = DNAAlphabet[x4];
	
	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = DNAAlphabet[y3];
	ali->charY[cur_pos+3] = DNAAlphabet[y4];

	nxt_pos = cur_pos + 4;
	nxt_st = stCe;
	break;

      case stC42:
	i = DLog2Choose(pcod42, 4096);
	
	cur_x = i / 16;
	cur_y = i % 16;
	
	x1 = (cur_x/4)/16;
	x2 = ((cur_x/4)%16)/4;
	x3 = (cur_x/4)%4;
	x4 = cur_x%4;

	y1 = cur_y/4;
	y2 = cur_y%4;
	y3 = 4;
	y4 = 4;

	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	seqX[cur_pos+3] = x4;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	seqY[cur_pos+3] = y4;
	
	ali->charX[cur_pos]   = DNAAlphabet[x1];
	ali->charX[cur_pos+1] = DNAAlphabet[x2];
	ali->charX[cur_pos+2] = DNAAlphabet[x3];
	ali->charX[cur_pos+3] = DNAAlphabet[x4];
	
	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = '.';
	ali->charY[cur_pos+3] = '.';

	nxt_pos = cur_pos + 4;
	nxt_st = stCe;
	break;

       case stC23:
	i = DLog2Choose(pcod32, 1024);

	cur_x = i % 16;
	cur_y = i / 16;

	x1 = cur_x/4;
	x2 = cur_x%4;
	x3 = 4;
	
	y1 = cur_y/16;
	y2 = (cur_y%16)/4;
	y3 = cur_y%4;
	
	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	
	ali->charX[cur_pos]   = DNAAlphabet[x1];
	ali->charX[cur_pos+1] = DNAAlphabet[x2];
	ali->charX[cur_pos+2] = '.';
	
	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = DNAAlphabet[y3];

	nxt_pos = cur_pos + 3;
	nxt_st = stCe;
	break;

      case stC24:
	i = DLog2Choose(pcod42, 4096);
	
	cur_x = i % 16;
	cur_y = i / 16;
	
	x1 = cur_x/4;
	x2 = cur_x%4;
	x3 = 4;
	x4 = 4;

	y1 = (cur_y/4)/16;
	y2 = ((cur_y/4)%16)/4;
	y3 = (cur_y/4)%4;
	y4 = cur_y%4;

	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	seqX[cur_pos+3] = x4;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	seqY[cur_pos+3] = y4;
	
	ali->charX[cur_pos]   = DNAAlphabet[y1];
	ali->charX[cur_pos+1] = DNAAlphabet[y2];
	ali->charX[cur_pos+2] = '.';
	ali->charX[cur_pos+3] = '.';

	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = DNAAlphabet[y3];
	ali->charY[cur_pos+3] = DNAAlphabet[y4];
	
	nxt_pos = cur_pos + 4;
	nxt_st = stCe;
	break;

       case stC22:
	nxt_st = stCe;
	break;

        case stC30:
	i = DLog2Choose(pcod30, 64);

	cur_x = i;

	x1 = cur_x/16;
	x2 = (cur_x%16)/4;
	x3 = cur_x%4;
	
	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	
	seqY[cur_pos]   = 4;
	seqY[cur_pos+1] = 4;
	seqY[cur_pos+2] = 4;
	
	ali->charX[cur_pos]   = DNAAlphabet[x1];
	ali->charX[cur_pos+1] = DNAAlphabet[x2];
	ali->charX[cur_pos+2] = DNAAlphabet[x3];
	
	ali->charY[cur_pos]   = '.';
	ali->charY[cur_pos+1] = '.';
	ali->charY[cur_pos+2] = '.';

	nxt_pos = cur_pos + 3;
	nxt_st = stCe;
	break;

      case stC40:
	i = DLog2Choose(pcod40, 256);

	cur_x = i/4;

	x1 = cur_x/16;
	x2 = (cur_x%16)/4;
	x3 = cur_x%4;
	x4 = i%4;
	
	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	seqX[cur_pos+3] = x4;
	
	seqY[cur_pos]   = 4;
	seqY[cur_pos+1] = 4;
	seqY[cur_pos+2] = 4;
	seqY[cur_pos+3] = 4;
	
	ali->charX[cur_pos]   = DNAAlphabet[x1];
	ali->charX[cur_pos+1] = DNAAlphabet[x2];
	ali->charX[cur_pos+2] = DNAAlphabet[x3];
	ali->charX[cur_pos+3] = DNAAlphabet[x4];
	
	ali->charY[cur_pos]   = '.';
	ali->charY[cur_pos+1] = '.';
	ali->charY[cur_pos+2] = '.';
	ali->charY[cur_pos+3] = '.';

	nxt_pos = cur_pos + 4;
	nxt_st = stCe;
	break;

       case stC20:
	nxt_st = stCe;
	break;

       case stC03:
	i = DLog2Choose(pcod30, 64);

	cur_y = i;

	y1 = cur_y/16;
	y2 = (cur_y%16)/4;
	y3 = cur_y%4;
	
	seqX[cur_pos]   = 4;
	seqX[cur_pos+1] = 4;
	seqX[cur_pos+2] = 4;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	
	ali->charX[cur_pos]   = '.';
	ali->charX[cur_pos+1] = '.';
	ali->charX[cur_pos+2] = '.';

	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = DNAAlphabet[y3];
	
	nxt_pos = cur_pos + 3;
	nxt_st = stCe;
	break;

       case stC04:
	i = DLog2Choose(pcod04, 256);

	cur_y = i/4;

	y1 = cur_y/16;
	y2 = (cur_y%16)/4;
	y3 = cur_y%4;
	y4 = i%4;
	
	seqX[cur_pos]   = 4;
	seqX[cur_pos+1] = 4;
	seqX[cur_pos+2] = 4;
	seqX[cur_pos+3] = 4;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	seqY[cur_pos+3] = y4;
	
	ali->charX[cur_pos]   = '.';
	ali->charX[cur_pos+1] = '.';
	ali->charX[cur_pos+2] = '.';
	ali->charX[cur_pos+3] = '.';

	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = DNAAlphabet[y3];
	ali->charY[cur_pos+3] = DNAAlphabet[y4];
	
	nxt_pos = cur_pos + 4;
	nxt_st = stCe;
	break;

       case stC02:
	nxt_st = stCe;
	break;

       default:
	Die("invalid state (%d) in SimulateCODAlign()", cur_st);
      }

      if (traceback) fprintf(ofp,"emitting %s (%d)\n", cstNAME[cur_st], cur_pos);
      
      if (nxt_st != -1) {
	if (traceback) fprintf(ofp," %s->%s \n", cstNAME[cur_st], cstNAME[nxt_st]);
	PushTracerstack(dolist, AttachTracer(cur_tr, nxt_pos+start, nxt_st));
      }
    }
  
  if (alignment) PrintNewAlign(ofp, sqinfoX, sqinfoY, start, cur_pos, ali, string_name);
  
  FreeTracer(cur_tr);
  FreeTracerstack(dolist);
}


/* Function:  SimulateCODAlign()
 * Date:      ER, Tue Mar 21 15:02:44 CST 2000 [St. Louis]
 *
 * Purpose:   Given a sequence generate another at a given evolutionary distance 
 *           forming an alignmentrelated by the COD model.
 *
 * Args:     s1, s2    - the sequences 
 *           L         - length of aligmnemt
 *           othmodel - the null model
 *
 * Returns:  (void)
 */
void
SimulateCODSequence(FILE *ofp, int *seq, SQINFO *sqinfoX, int *seqX, SQINFO *sqinfoY, int *seqY, int start, 
		    int *L, struct codmodel_s *cod, struct nullmodel_s *null, struct ali_s *ali, 
		    int traceback, int alignment, char *string_name)
{
  struct tracer_s      *tr;      /* the traceback tree under construction  */
  struct tracer_s      *cur_tr;  /* ptr to node of tr we're working on     */
  struct tracerstack_s *dolist;  /* pushdown stack of active tr nodes      */

  double pccod33[64][64];
  double pccod44[256][256];
  double pccod32[64][16];
  double pccod23[16][64];
  double pccod34[64][256];
  double pccod43[265][64];
  double pccod42[256][16];
  double pccod24[16][256];

  double pcod30[64];
  double pcod40[256];
  double pcod04[256];
  double pcod20[16];

  int    ct, s[4];

  int    cur_x,   cur_y;         /* nucleotides at those positions         */
  int    cur_st,  nxt_st;
  int    cur_pos, nxt_pos, seq_pos = 0;
  int    len_seq;
  int    pick;
  int    i;
  int    x, y;
  int    x1, x2, x3, x4;
  int    y1, y2, y3, y4;

  len_seq = *L;

  for (i = 0; i < 16; i++) 
    pcod20[i] = 0.;

  for (i = 0; i < 64; i++) {
    pcod30[i] = average_codon(i, cod, null, FALSE);
    pcod20[i/4] += EXP2(pcod30[i]);
   }
  for (i = 0; i < 16; i++) 
    pcod20[i] = LOG2(pcod20[i]);


  for (i = 0; i < 256; i++) {
    pcod40[i] = average_codon(i/4, cod, null, FALSE) + null->xem[i%4];
    pcod04[i] = average_codon(i/4, cod, null, FALSE) + null->yem[i%4];
  }


  for (i = 0; i < 4096; i++) 
    {
      x = i / 64;
      y = i % 64;
      pccod33[x][y] = cod->pcodon[x][y] - pcod30[x];
    }


  for (i = 0; i < 1024; i++) {
    x = i/16;
    y = i%16;

    pccod32[x][y] = average(3, y/4, y%4, x, cod, null, FALSE) - pcod30[x];
    pccod23[y][x] = average(3, y/4, y%4, x, cod, null, FALSE) - pcod20[y];
  }

  for (i = 0; i < 16384; i++) 
    {
      x = i / 256;
      y = i % 256;

      pccod34[x][y] = pccod33[x][y/4] + null->yem[y%4];
      pccod43[y][x] = pccod33[y/4][x];
    }


  for (i = 0; i < 65536; i++) 
    {
      x = i / 256;
      y = i % 256;

      pccod44[x][y] = pccod33[x/4][y/4] + null->yem[y%4];
    }

  for (i = 0; i < 4096; i++) 
    {
      x = i / 16;
      y = i % 16;
      pccod42[x][y] = pccod32[x/4][y];
      pccod24[y][x] = pccod23[y][x/4] + null->yem[x%4];
    }

  /* check conditional probabilities
   */
  CheckSingleLog2Prob(pcod30, 64);
  CheckSingleLog2Prob(pcod40, 256);
  CheckSingleLog2Prob(pcod04, 256);
  CheckSingleLog2Prob(pcod20, 16);
  for (i = 0; i < 64; i++) {
    CheckSingleLog2Prob(pccod34[i], 256);
    CheckSingleLog2Prob(pccod33[i], 64);
    CheckSingleLog2Prob(pccod32[i], 16);
  }
  for (i = 0; i < 16; i++) {
    CheckSingleLog2Prob(pccod23[i], 64);
    CheckSingleLog2Prob(pccod24[i], 256);
  }
  for (i = 0; i < 256; i++) {
    CheckSingleLog2Prob(pccod44[i], 256);
    CheckSingleLog2Prob(pccod43[i], 64);
    CheckSingleLog2Prob(pccod42[i], 16);
  }

 /* Initialize
   * Start at pos "0" with state "stB"
   */
  tr     = InitTracer();       /* start a trace tree */
  dolist = InitTracerstack();  /* start a stack for traversing the trace tree */
  
  cur_tr = AttachTracer(tr, start, stCb); 
  PushTracerstack(dolist, cur_tr);

  
  while ((cur_tr = PopTracerstack(dolist)) != NULL)
    {
      cur_st = cur_tr->type;
      
      cur_pos = cur_tr->emit;

      switch (cur_st){
      case stCb:
	cur_x = -1;
	cur_y = -1;
	nxt_pos = cur_pos;

	/* do not allow any possible frameshit at the very begining
	 */
	if (cur_pos == 0) 
	  pick = DENChoose(cod->t + IdxTransCOD[cur_st], TransPerCODState[cur_st]-1);
	else 
	  pick = DENChoose(cod->t + IdxTransCOD[cur_st], TransPerCODState[cur_st]);
	
	nxt_st = cur_st + 1 + pick;
	break;

      case stCe:
	nxt_pos = cur_pos;
	if (nxt_pos < start+*L-2) nxt_st = stCb;
	else nxt_st = -1;
	break;

      case stC33:
	ct = 0;
	while (ct < 3) {
	  if (!isitagap(seq[seq_pos])) s[ct++] = seq[seq_pos];
	  if(++seq_pos > len_seq) break;
	}
	
	x1 = s[0];
	x2 = s[1];
	x3 = s[2];
	cur_x = CODON(x1,x2,x3);
	cur_y = DLog2Choose(pccod33[cur_x], 64);

	y1 = cur_y/16;
	y2 = (cur_y%16)/4;
	y3 = cur_y%4;
	
	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	
	ali->charX[cur_pos]   = DNAAlphabet[x1];
	ali->charX[cur_pos+1] = DNAAlphabet[x2];
	ali->charX[cur_pos+2] = DNAAlphabet[x3];
	
	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = DNAAlphabet[y3];

	nxt_pos = cur_pos + 3;
	nxt_st = stCe;
	break;

      case stC34:
	ct = 0;
	while (ct < 3) {
	  if (!isitagap(seq[seq_pos])) s[ct++] = seq[seq_pos];
	  if(++seq_pos > len_seq) break;
	}

	x1 = s[0];
	x2 = s[1];
	x3 = s[2];
	x4 = 4;

	cur_x =  CODON(x1,x2,x3);
	cur_y = DLog2Choose(pccod34[cur_x], 256);
	
	y1 = (cur_y/4)/16;
	y2 = ((cur_y/4)%16)/4;
	y3 = (cur_y/4)%4;
	y4 = cur_y%4;
	
	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	seqX[cur_pos+3] = x4;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	seqY[cur_pos+3] = y4;
	
	ali->charX[cur_pos]   = DNAAlphabet[x1];
	ali->charX[cur_pos+1] = DNAAlphabet[x2];
	ali->charX[cur_pos+2] = DNAAlphabet[x3];
	ali->charX[cur_pos+3] = '.';
	
	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = DNAAlphabet[y3];
	ali->charY[cur_pos+3] = DNAAlphabet[y4];

	nxt_pos = cur_pos + 3;
	nxt_st = stCe;
	break;

      case stC32:
	ct = 0;
	while (ct < 3) {
	  if (!isitagap(seq[seq_pos])) s[ct++] = seq[seq_pos];
	  if(++seq_pos > len_seq) break;
	}

	x1 = s[0];
	x2 = s[1];
	x3 = s[2];
	
	cur_x = CODON(x1,x2,x3);
	cur_y = DLog2Choose(pccod32[cur_x], 16);

	y1 = cur_y/4;
	y2 = cur_y%4;
	y3 = 4;
	
	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	
	ali->charX[cur_pos]   = DNAAlphabet[x1];
	ali->charX[cur_pos+1] = DNAAlphabet[x2];
	ali->charX[cur_pos+2] = DNAAlphabet[x3];
	
	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = '.';

	nxt_pos = cur_pos + 3;
	nxt_st = stCe;
	break;

      case stC43:
	ct = 0;
	while (ct < 4) {
	  if (!isitagap(seq[seq_pos])) s[ct++] = seq[seq_pos];
	  if(++seq_pos > len_seq) break;
	}
	
	x1 = s[0];
	x2 = s[1];
	x3 = s[2];
	x4 = s[3];
	
	cur_x = x1*64 + CODON(x2,x3,x4);
	cur_y = DLog2Choose(pccod43[cur_x], 64);
	
	y1 = cur_y/16;
	y2 = (cur_y%16)/4;
	y3 = cur_y%4;
	y4 = 4;

	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	seqX[cur_pos+3] = x4;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	seqY[cur_pos+3] = y4;
	
	ali->charX[cur_pos]   = DNAAlphabet[x1];
	ali->charX[cur_pos+1] = DNAAlphabet[x2];
	ali->charX[cur_pos+2] = DNAAlphabet[x3];
	ali->charX[cur_pos+3] = DNAAlphabet[x4];
	
	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = DNAAlphabet[y3];
	ali->charY[cur_pos+3] = '.';

	nxt_pos = cur_pos + 4;
	nxt_st = stCe;
	break;

       case stC44:
	ct = 0;
	while (ct < 4) {
	  if (!isitagap(seq[seq_pos])) s[ct++] = seq[seq_pos];
	  if(++seq_pos > len_seq) break;
	}
	
	x1 = s[0];
	x2 = s[1];
	x3 = s[2];
	x4 = s[3];
	
	cur_x = x1*64 + CODON(x2,x3,x4);
	cur_y = DLog2Choose(pccod44[cur_x], 256);

	y = cur_y/4;
	y1 = y/16;
	y2 = (y%16)/4;
	y3 = y%4;
	y4 = cur_y%4;
	
	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	seqX[cur_pos+3] = x4;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	seqY[cur_pos+3] = y4;
	
	ali->charX[cur_pos]   = DNAAlphabet[x1];
	ali->charX[cur_pos+1] = DNAAlphabet[x2];
	ali->charX[cur_pos+2] = DNAAlphabet[x3];
	ali->charX[cur_pos+3] = DNAAlphabet[x4];
	
	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = DNAAlphabet[y3];
	ali->charY[cur_pos+3] = DNAAlphabet[y4];

	nxt_pos = cur_pos + 4;
	nxt_st = stCe;
	break;

      case stC42:
	ct = 0;
	while (ct < 4) {
	  if (!isitagap(seq[seq_pos])) s[ct++] = seq[seq_pos];
	  if(++seq_pos > len_seq) break;
	}
	
	x1 = s[0];
	x2 = s[1];
	x3 = s[2];
	x4 = s[3];
	
	cur_x = x1*64 + CODON(x2,x3,x4);
	cur_y = DLog2Choose(pccod42[cur_x], 16);
	
	y1 = cur_y/4;
	y2 = cur_y%4;
	y3 = 4;
	y4 = 4;

	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	seqX[cur_pos+3] = x4;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	seqY[cur_pos+3] = y4;
	
	ali->charX[cur_pos]   = DNAAlphabet[x1];
	ali->charX[cur_pos+1] = DNAAlphabet[x2];
	ali->charX[cur_pos+2] = DNAAlphabet[x3];
	ali->charX[cur_pos+3] = DNAAlphabet[x4];
	
	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = '.';
	ali->charY[cur_pos+3] = '.';

	nxt_pos = cur_pos + 4;
	nxt_st = stCe;
	break;

       case stC23:
	ct = 0;
	while (ct < 2) {
	  if (!isitagap(seq[seq_pos])) s[ct++] = seq[seq_pos];
	  if(++seq_pos > len_seq) break;
	}
	
	x1 = s[0];
	x2 = s[1];
	x3 = 4;
	
	cur_x = x1*16 + x2;
	cur_y = DLog2Choose(pccod23[cur_x], 64);

	y1 = cur_y/16;
	y2 = (cur_y%16)/4;
	y3 = cur_y%4;
	
	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	
	ali->charX[cur_pos]   = DNAAlphabet[x1];
	ali->charX[cur_pos+1] = DNAAlphabet[x2];
	ali->charX[cur_pos+2] = '.';
	
	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = DNAAlphabet[y3];

	nxt_pos = cur_pos + 3;
	nxt_st = stCe;
	break;

      case stC24:
	ct = 0;
	while (ct < 2) {
	  if (!isitagap(seq[seq_pos])) s[ct++] = seq[seq_pos];
	  if(++seq_pos > len_seq) break;
	}
	
	x1 = s[0];
	x2 = s[1];
	x3 = 4;
	
	cur_x = x1*16 + x2;
	cur_y = DLog2Choose(pccod24[cur_x], 256);
	
	y1 = (cur_y/4)/16;
	y2 = ((cur_y/4)%16)/4;
	y3 = (cur_y/4)%4;
	y4 = cur_y%4;

	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	seqX[cur_pos+3] = x4;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	seqY[cur_pos+3] = y4;
	
	ali->charX[cur_pos]   = DNAAlphabet[y1];
	ali->charX[cur_pos+1] = DNAAlphabet[y2];
	ali->charX[cur_pos+2] = '.';
	ali->charX[cur_pos+3] = '.';

	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = DNAAlphabet[y3];
	ali->charY[cur_pos+3] = DNAAlphabet[y4];
	
	nxt_pos = cur_pos + 4;
	nxt_st = stCe;
	break;

       case stC22:
	nxt_st = stCe;
	break;

        case stC30:
	ct = 0;
	while (ct < 3) {
	  if (!isitagap(seq[seq_pos])) s[ct++] = seq[seq_pos];
	  if(++seq_pos > len_seq) break;
	}

	x1 = s[0];
	x2 = s[1];
	x3 = s[2];

	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	
	seqY[cur_pos]   = 4;
	seqY[cur_pos+1] = 4;
	seqY[cur_pos+2] = 4;
	
	ali->charX[cur_pos]   = DNAAlphabet[x1];
	ali->charX[cur_pos+1] = DNAAlphabet[x2];
	ali->charX[cur_pos+2] = DNAAlphabet[x3];
	
	ali->charY[cur_pos]   = '.';
	ali->charY[cur_pos+1] = '.';
	ali->charY[cur_pos+2] = '.';

	nxt_pos = cur_pos + 3;
	nxt_st = stCe;
	break;

      case stC40:
	ct = 0;
	while (ct < 4) {
	  if (!isitagap(seq[seq_pos])) s[ct++] = seq[seq_pos];
	  if(++seq_pos > len_seq) break;
	}
	
	x1 = s[0];
	x2 = s[1];
	x3 = s[2];
	x4 = s[3];

	seqX[cur_pos]   = x1;
	seqX[cur_pos+1] = x2;
	seqX[cur_pos+2] = x3;
	seqX[cur_pos+3] = x4;
	
	seqY[cur_pos]   = 4;
	seqY[cur_pos+1] = 4;
	seqY[cur_pos+2] = 4;
	seqY[cur_pos+3] = 4;
	
	ali->charX[cur_pos]   = DNAAlphabet[x1];
	ali->charX[cur_pos+1] = DNAAlphabet[x2];
	ali->charX[cur_pos+2] = DNAAlphabet[x3];
	ali->charX[cur_pos+3] = DNAAlphabet[x4];
	
	ali->charY[cur_pos]   = '.';
	ali->charY[cur_pos+1] = '.';
	ali->charY[cur_pos+2] = '.';
	ali->charY[cur_pos+3] = '.';

	nxt_pos = cur_pos + 4;
	nxt_st = stCe;
	break;

       case stC20:
	nxt_st = stCe;
	break;

       case stC03:
	i = DLog2Choose(pcod30, 64);

	cur_y = i;

	y1 = cur_y/16;
	y2 = (cur_y%16)/4;
	y3 = cur_y%4;
	
	seqX[cur_pos]   = 4;
	seqX[cur_pos+1] = 4;
	seqX[cur_pos+2] = 4;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	
	ali->charX[cur_pos]   = '.';
	ali->charX[cur_pos+1] = '.';
	ali->charX[cur_pos+2] = '.';

	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = DNAAlphabet[y3];
	
	nxt_pos = cur_pos + 3;
	nxt_st = stCe;
	break;

       case stC04:
	i = DLog2Choose(pcod04, 256);

	cur_y = i/4;

	y1 = cur_y/16;
	y2 = (cur_y%16)/4;
	y3 = cur_y%4;
	y4 = i%4;
	
	seqX[cur_pos]   = 4;
	seqX[cur_pos+1] = 4;
	seqX[cur_pos+2] = 4;
	seqX[cur_pos+3] = 4;
	
	seqY[cur_pos]   = y1;
	seqY[cur_pos+1] = y2;
	seqY[cur_pos+2] = y3;
	seqY[cur_pos+3] = y4;
	
	ali->charX[cur_pos]   = '.';
	ali->charX[cur_pos+1] = '.';
	ali->charX[cur_pos+2] = '.';
	ali->charX[cur_pos+3] = '.';

	ali->charY[cur_pos]   = DNAAlphabet[y1];
	ali->charY[cur_pos+1] = DNAAlphabet[y2];
	ali->charY[cur_pos+2] = DNAAlphabet[y3];
	ali->charY[cur_pos+3] = DNAAlphabet[y4];
	
	nxt_pos = cur_pos + 4;
	nxt_st = stCe;
	break;

       case stC02:
	nxt_st = stCe;
	break;

       default:
	Die("invalid state (%d) in SimulateCODSequence()", cur_st);
      }

      if (traceback) fprintf(ofp,"emitting %s (%d)\n", cstNAME[cur_st], cur_pos);
      
      if (seq_pos < len_seq && nxt_st != -1) {
	if (traceback) fprintf(ofp," %s->%s \n", cstNAME[cur_st], cstNAME[nxt_st]);
	PushTracerstack(dolist, AttachTracer(cur_tr, nxt_pos+start, nxt_st));
      }
    }
  
  if (alignment) PrintNewAlign(ofp, sqinfoX, sqinfoY, start, cur_pos, ali, string_name);

  *L = nxt_pos;
  
  FreeTracer(cur_tr);
  FreeTracerstack(dolist);
}


/* Function:  pass_thru_COJ()
 * Date:      ER, Sat Feb  5 14:13:23 CST 2000 [St. Louis]
 *
 * Purpose:  Goes through the COJ model maybe generating a few positions.
 *           Very similar to SimulateOTHSequence() apart from the fact that
 *           it exits the model when convenient but emits at least 1 nts.
 *
 * Args:     s1, s2    - the sequences 
 *           L         - length of aligmnemt
 *           othmodel - the null model
 *
 * Returns:  (void)
 */
void
pass_thru_COJ(FILE *ofp, SQINFO *sqinfoX, int *seqX, SQINFO *sqinfoY, int *seqY, int start, 
		 int *ret_nxt_pos, struct othmodel_s *oth, struct ali_s *ali, int traceback, int alignment, 
		 char *string_name)
{
  struct tracer_s      *tr;      /* the traceback tree under construction  */
  struct tracer_s      *cur_tr;  /* ptr to node of tr we're working on     */
  struct tracerstack_s *dolist;  /* pushdown stack of active tr nodes      */

  int    cur_x,   cur_y;           /* nucleotides at those positions         */
  int    cur_st,  nxt_st;
  int    cur_pos, nxt_pos;
  int    t, x;

  /* Initialize
   * Start at pos "0" with state "stB"
   */
  tr     = InitTracer();       /* start a trace tree */
  dolist = InitTracerstack();  /* start a stack for traversing the trace tree */
  
  cur_tr = AttachTracer(tr, start, stB); 
  PushTracerstack(dolist, cur_tr);

  while ((cur_tr = PopTracerstack(dolist)) != NULL)
    {
      cur_st = cur_tr->type;
      
      t = DENChoose(oth->t + IdxTransOTH[cur_st], TransPerOTHState[cur_st]);
      if      (t == 0) nxt_st = stM;
      else if (t == 1) nxt_st = stX;
      else if (t == 2) nxt_st = stY;
      else if (t == 3) nxt_st = stE;
      else Die ("invalid transition in SimulateOTHAlign()");

      cur_pos = cur_tr->emit;

      switch (cur_st){
      case stB:
	cur_x = -1;
	cur_y = -1;
	nxt_pos = cur_pos;
	break;

      case stM:
	x = DLog2Choose(oth->mem, 16);

	cur_x = x/4;
	cur_y = x%4;
	nxt_pos = cur_pos + 1;
	ali->charX[cur_pos] = DNAAlphabet[cur_x];
	ali->charY[cur_pos] = DNAAlphabet[cur_y];
	break;

      case stX: 
	cur_x = DLog2Choose(oth->xem, 4);
	cur_y = 4;
	nxt_pos = cur_pos + 1;
	ali->charX[cur_pos] = DNAAlphabet[cur_x];
	ali->charY[cur_pos] = '.';
	break;

      case stY: 
	cur_x = 4;
	cur_y = DLog2Choose(oth->yem, 4);
	nxt_pos = cur_pos + 1;
	ali->charX[cur_pos] = '.';
	ali->charY[cur_pos] = DNAAlphabet[cur_y];
	break;

      case stE: 
	*ret_nxt_pos = cur_pos;
	break;

      default:
	Die("invalid state in pass_thru_COJ()");
      }

      if (traceback) fprintf(ofp,"emitting %s (%d) [%d %d] \n", ostNAME[cur_st], cur_pos, cur_x, cur_y);
      
      seqX[cur_pos] = cur_x;
      seqY[cur_pos] = cur_y;
      
      if (cur_st != stE) {
	if (traceback) fprintf(ofp," %s->%s \n", ostNAME[cur_st], ostNAME[nxt_st]);
	PushTracerstack(dolist, AttachTracer(cur_tr, nxt_pos, nxt_st));
      }
    }
  
  if (alignment) PrintNewAlign(ofp, sqinfoX, sqinfoY, start, *ret_nxt_pos-start+1, ali, string_name);
  
  FreeTracer(cur_tr);
  FreeTracerstack(dolist);
}



