/************************************************************
 * QRNA - Comparative analysis of biological sequences 
 *        with pair hidden Markov models and pair stochastic context-free grammars
 * Copyright (C) 2000-Howard Hughes Medical Institute
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* codmodel.c
 *
 * E. Rivas [St. Louis]
 * 
 * 9 april 1999.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

#include "funcs.h"
#include "globals.h"
#include "squid.h"
#include "structs.h"

#ifdef MEMDEBUG
#include "dbmalloc.h"
#endif

static double normalizecodparam(int tfactor);

/* Function: AllocCODModel()
 * Date:     ER,  Mon Jun  7 11:46:48 CDT 1999 [St. Louis]
 *
 * Purpose:  Allocates memory for the transition and emission probs of the codingmodel
 *
 * Args:     codmodel - codmodel structure
 *
 * Returns:  void
 *           allocates cod->t[], cod->pcodon[], which must be free'd by caller.
 */
struct codmodel_s *
AllocCODModel(void)
{
  struct codmodel_s *cod;    /* codmodel structure           */
  int                codx;   /* symbols for codon prob's     */

  cod              = (struct codmodel_s *)  MallocOrDie (sizeof(struct codmodel_s));
  cod->t           = (double  *)            MallocOrDie (sizeof(double  ) * CTRANS);
  cod->pmut        = (double  *)            MallocOrDie (sizeof(double  ) * 16);
  cod->pcodon      = (double **)            MallocOrDie (sizeof(double *) * 64);
  cod->pcodon[0]   = (double  *)            MallocOrDie (sizeof(double  ) * 64 * 64);
  cod->phexa       = (double **)            MallocOrDie (sizeof(double *) * 64);
  cod->phexa[0]    = (double  *)            MallocOrDie (sizeof(double  ) * 64 * 64);

  cod->COB = AllocOTHModel();
  cod->COJ = AllocOTHModel();
  cod->COE = AllocOTHModel();

  for (codx = 1; codx < 64; codx++) {
    cod->pcodon[codx] = cod->pcodon[0] + codx * 64;
    cod->phexa[codx]  = cod->phexa[0]  + codx * 64;
  }

  PatternCODModel(cod);
  
  return cod;
}

struct pamcond_s *
AllocPamCond(void)
{
  struct pamcond_s *pamcond;
  int               codx, cody;

  pamcond = (struct pamcond_s *)  MallocOrDie (sizeof(struct pamcond_s));

  pamcond->pXY    = (double **) MallocOrDie (sizeof(double *) * 64);
  pamcond->pYX    = (double **) MallocOrDie (sizeof(double *) * 64);
  pamcond->pXY[0] = (double  *) MallocOrDie (sizeof(double  ) * 64 * 64);
  pamcond->pYX[0] = (double  *) MallocOrDie (sizeof(double  ) * 64 * 64);

  for (codx = 1; codx < 64; codx++) {
    pamcond->pXY[codx] =  pamcond->pXY[0] + codx * 64;
    pamcond->pYX[codx] =  pamcond->pYX[0] + codx * 64;
  }
  
  /* Initialize */
  for (codx = 0; codx < 64; codx++) 
    for (cody = 0; cody < 64; cody++) {
      pamcond->pXY[codx][cody] = 0.0;
      pamcond->pYX[codx][cody] = 0.0;
    }

  return pamcond;
}

/* Function: CheckCODProbs()
 * Date:     ER, Mon Jun  7 12:18:45 CDT 1999 [St. Louis]
 *
 * Purpose:  Verify that transition and emission prob's of a codmodel add up to one
 *
 * Args:     cod - the structure for an codmodel
 *
 * Returns:  void. 
 */
void
CheckCODProbs(struct codmodel_s *cod)
{
  int    idx;               /* index for transition prob's      +*/
  int    st;                /* state we are at                  +*/
  double sum;

  /* check transition prob's add up to one
   */
  for (st = 0; st < CSTATES; st++) {
    sum = 0.0; 
    for (idx = IdxTransCOD[st]; idx < IdxTransCOD[st]+TransPerCODState[st]; idx++)
      sum += cod->t[idx];
    
    if (sum > 2.-accuracy || sum < accuracy) 
      Die("transition prob's sum for state %d(%s) is %f \n", st, cstNAME[st], sum);
  }
  
  /* check emission prob's add up to one
   */
  CheckSingleProb(cod->pmut, 16);
  CheckDoubleProb(cod->pcodon, 64, 64);
}

/* Function: CheckPAMModel()
 * Date:     SRE, Wed Jun 10 13:40:01 1998 [St. Louis]
 *
 * Purpose:  Verify that a PAM model is OK
 *
 * Args:     pammodel - model to check [64][64]
 *
 * Returns:  void. prints stuff.
 */
void
CheckPAMModel(double **pammodel)
{
  int x, y;
  double sum = 0.0;

  for (x = 0; x < 64; x++)
    for (y = 0; y < 64; y++)
      sum += pammodel[x][y];

  if (sum > 1.01 || sum < 0.99) Die ("pam sum is %f\n", sum);
}

/* Function: CODToLog2()
 * Date:     ER, Mon Jun  7 12:24:32 CDT 1999 [St. Louis]
 *
 * Purpose:  Converts transition and emission prob's of a codmodel to log2 form
 *
 * Args:     cod - the structure for a codmodel
 *
 * Returns:  void. 
 */
void
CODToLog2(struct codmodel_s *cod)
{
  int idx;   /* index for transition prob's      +*/
  int x,y;   /* symbols for emission prob's      +*/

  /* transition prob's 
   */
  for (idx = 0; idx < CTRANS; idx++) 
    cod->t[idx] = LOG2(cod->t[idx]);
  
  /* emission prob's 
   */
  for (x = 0; x < 16; x++) 
    cod->pmut[x] = LOG2(cod->pmut[x]);
  for (x = 0; x < 64; x++) 
    for (y = 0; y < 64; y++) 
      cod->pcodon[x][y] = LOG2(cod->pcodon[x][y]);

}

/* Function: CODLog2ToOdds()
 * Date:     ER, Mon Jun  7 12:22:42 CDT 1999 [St. Louis]
 *
 * Purpose:  Converts transition and emission prob's of a codingmodel
 *           from log2 to log2odds form
 *
 * Args:     coding - the structure for an codingmodel
 *
 * Returns:  void. 
 */
void
CODLog2ToOdds(struct codmodel_s *cod, struct nullmodel_s *null)
{
  int    st;          /* index for states                   +*/
  int    tr;          /* index for transitions              +*/
  int    emit;        /* number of emissions per state      +*/
  int    x, y;        /* symbols for emission prob's        +*/
  
  OTHLog2ToOdds(cod->COB, null);
  OTHLog2ToOdds(cod->COJ, null);
  OTHLog2ToOdds(cod->COE, null);

  /* transition prob's 
   */
  for (st = 0; st < CSTATES; st++) {

    emit = EmitsPerCODState[st];
    /* [emit =   1] ==> no emission    or 
     * [emit = 4^x] ==> x  emissions   
     */
    if (emit%4 == 0)
      for (tr = IdxTransCOD[st]; tr < IdxTransCOD[st]+TransPerCODState[st]; tr++)
	cod->t[tr] -= 0.5 * LOG2(emit) * null->meta;   
    else if (emit != 1)
      Die ("wrong number of emissions (%d) for state %s", emit, cstNAME[st]);
  }

  /* emission prob's 
   */
  for (x = 0; x < 16; x++) 
    cod->pmut[x] -= (null->xem[x/4] + null->yem[x%4]);
  for (x = 0; x < 64; x++) 
    for (y = 0; y < 64; y++) 
      cod->pcodon[x][y] -= (null->xem[x/16] + null->xem[(x%16)/4] + null->xem[x%4] +
			    null->yem[y/16] + null->yem[(y%16)/4] + null->yem[y%4]);

}

/* Function: ConstructCODModel()
 * Date:     ER, Sat Jun  5 08:54:20 CDT 1999  [St. Louis]
 *
 * Purpose:  Constructs a codingmodel_s
 *
 * Args:     codingparam - the list of parameters that define a codingmodel           
 *
 * Returns:  (void)
 *           fills all prob's for othermodel, log2 form 
 *           (allc'ed here, freed by caller)
 */
void
ConstructCODModel(double *mutpxy, double **pammodel, double **hexa, double tfactor, 
		  struct codmodel_s *cod, int add_hexamer, int pedantic)
{
  double             tr;              /* transition prob's                +*/
  int                idx;             /* index for transition prob's      +*/
  double             sumcb;           /* normalization factor for codon transitions +*/

  ConstructOTHModel(mutpxy, COBparam, tfactor, cod->COB, pedantic);
  ConstructOTHModel(mutpxy, COJparam, tfactor, cod->COJ, pedantic);
  ConstructOTHModel(mutpxy, COEparam, tfactor, cod->COE, pedantic);

  /* normalize transition for state Cb (codon begins) */
  sumcb = (tfactor > 1.0)? normalizecodparam(tfactor) : normalizecodparam(1.0); 

  /* Calculate transition prob's: cod.t[]
   */
  for (idx = 0; idx < CTRANS; idx++) {
    switch (idx) {
    case TCOBCb:  tr = CODparam.psi;                                       break;
    case TCOBCOE: tr = 1.0 - CODparam.psi;                                 break;
    case TCbCe:   tr = (tfactor > 1.0)? 
		    EXP2(tfactor*LOG2(CODparam.tau)) : CODparam.tau;       break;
    case TCbC33:  tr = (tfactor > 1.0)? 
		    EXP2(tfactor*LOG2(CODparam.xi[0][0]))/(1. * sumcb) : CODparam.xi[0][0]/(1. * sumcb); break;
    case TCbC34:  tr = (tfactor > 1.0)? 
		    EXP2(tfactor*LOG2(CODparam.xi[0][1]))/(1. * sumcb) : CODparam.xi[0][1]/(1. * sumcb); break;
    case TCbC32:  tr = (tfactor > 1.0)? 
		    EXP2(tfactor*LOG2(CODparam.xi[0][2]))/(1. * sumcb) : CODparam.xi[0][2]/(1. * sumcb); break;
    case TCbC43:  tr = (tfactor > 1.0)? 
		    EXP2(tfactor*LOG2(CODparam.xi[1][0]))/(1. * sumcb) : CODparam.xi[1][0]/(1. * sumcb); break;
    case TCbC44:  tr = (tfactor > 1.0)? 
		    EXP2(tfactor*LOG2(CODparam.xi[1][1]))/(1. * sumcb) : CODparam.xi[1][1]/(1. * sumcb); break;
    case TCbC42:  tr = (tfactor > 1.0)? 
		    EXP2(tfactor*LOG2(CODparam.xi[1][2]))/(1. * sumcb) : CODparam.xi[1][2]/(1. * sumcb); break;
    case TCbC23:  tr = (tfactor > 1.0)? 
		    EXP2(tfactor*LOG2(CODparam.xi[2][0]))/(1. * sumcb) : CODparam.xi[2][0]/(1. * sumcb); break;
    case TCbC24:  tr = (tfactor > 1.0)? 
		    EXP2(tfactor*LOG2(CODparam.xi[2][1]))/(1. * sumcb) : CODparam.xi[2][1]/(1. * sumcb); break;
    case TCbC22:  tr = (tfactor > 1.0)? 
		    EXP2(tfactor*LOG2(CODparam.xi[2][2]))/(1. * sumcb) : CODparam.xi[2][2]/(1. * sumcb); break;
    case TCbC30:  tr = (tfactor > 1.0)? 
		    EXP2(tfactor*LOG2(CODparam.xi[0][3]))/(1. * sumcb) : CODparam.xi[0][3]/(1. * sumcb); break;
    case TCbC40:  tr = (tfactor > 1.0)? 
		    EXP2(tfactor*LOG2(CODparam.xi[1][3]))/(1. * sumcb) : CODparam.xi[1][3]/(1. * sumcb); break;
    case TCbC20:  tr = (tfactor > 1.0)? 
		    EXP2(tfactor*LOG2(CODparam.xi[2][3]))/(1. * sumcb) : CODparam.xi[2][3]/(1. * sumcb); break;
    case TCbC03:  tr = (tfactor > 1.0)? 
		    EXP2(tfactor*LOG2(CODparam.xi[3][0]))/(1. * sumcb) : CODparam.xi[3][0]/(1. * sumcb); break;
    case TCbC04:  tr = (tfactor > 1.0)? 
		    EXP2(tfactor*LOG2(CODparam.xi[3][1]))/(1. * sumcb) : CODparam.xi[3][1]/(1. * sumcb); break;
    case TCbC02:  tr = (tfactor > 1.0)? 
		    EXP2(tfactor*LOG2(CODparam.xi[3][2]))/(1. * sumcb) : CODparam.xi[3][2]/(1. * sumcb); break;
    case TC33Ce:  tr = 1.0;                                                break;
    case TC34Ce:  tr = 1.0;                                                break;
    case TC32Ce:  tr = 1.0;                                                break;
    case TC43Ce:  tr = 1.0;                                                break;
    case TC44Ce:  tr = 1.0;                                                break;
    case TC42Ce:  tr = 1.0;                                                break;
    case TC23Ce:  tr = 1.0;                                                break;
    case TC24Ce:  tr = 1.0;                                                break;
    case TC22Ce:  tr = 1.0;                                                break;
    case TC30Ce:  tr = 1.0;                                                break;
    case TC40Ce:  tr = 1.0;                                                break;
    case TC20Ce:  tr = 1.0;                                                break;
    case TC03Ce:  tr = 1.0;                                                break;
    case TC04Ce:  tr = 1.0;                                                break;
    case TC02Ce:  tr = 1.0;                                                break;
    case TCeCOJ:  tr = 1.0 - CODparam.eta;                                 break;
    case TCeCOE:  tr = CODparam.eta;                                       break;
    case TCOJCb:  tr = 1.0;                                                break;
    case TCOECOE: tr = 1.0;                                                break;
    default:  Die("transition (%d) does not exits", idx);
    }

    if (tr > 1.0 || tr < 0.0) Die ("transition %s in COD model is wrong", ctrNAME[idx]);
    else  cod->t[idx] = tr;
  }

  /* Calculate emission prob's: model.mem[], model.xem[], model.yem[]
   */
  ConstructCODEmitProbs(pammodel, cod->pmut, cod->pcodon);
  
  /* check prob's add up to one
   */
  CheckCODProbs(cod);

  /* convert to Log2 form
   */
  CODToLog2(cod);

  ConstructHexamerProbs(hexa, cod->phexa, add_hexamer);
}


/* Function: ConstructCODEmitProbs()
 * Date:     ER, Mon Jun  7 12:03:23 CDT 1999 [St. Louis]
 *
 * Purpose:  construct codon probs
 *
 * Args:     pammodel  - 64x64 AAA..UUUxAAA..UUU joint prob matrix (prealloc)
 *           pm        - 4x4 A..UxA..U joint prob matrix (prealloc)           
 *
 * Returns:  (void)
 *           Fills in other->mem. (already allocated)
 */
void
ConstructCODEmitProbs(double **pammodel, double *pmut, double **pcodon)
{
  int x, y;
  int x1, x2, x3;
  int y1, y2, y3;

  /* Marginalize and average over three positions
   */
  for (x1 = 0; x1 < 4; x1++)
    for (x2 = 0; x2 < 4; x2++)
      for (x3 = 0; x3 < 4; x3++)
	for (y1 = 0; y1 < 4; y1++)
	  for (y2 = 0; y2 < 4; y2++)
	    for (y3 = 0; y3 < 4; y3++)
	      {
		pmut[idx(x1,y1)] += pammodel[CODON(x1,x2,x3)][CODON(y1,y2,y3)] / 3.0;
		pmut[idx(x2,y2)] += pammodel[CODON(x1,x2,x3)][CODON(y1,y2,y3)] / 3.0;
		pmut[idx(x3,y3)] += pammodel[CODON(x1,x2,x3)][CODON(y1,y2,y3)] / 3.0;
	      }

  for (x = 0; x < 64; x++)
    for (y = 0; y < 64; y++)
      pcodon[x][y] = pammodel[x][y];

  CheckSingleProb(pmut, 16);
  CheckDoubleProb(pcodon, 64, 64);
}

/* Function: ConstructESTModel()
 * Date:     SRE, Wed Jun 10 08:46:06 1998 [St. Louis]
 *
 * Purpose:  Return an estmodel: a 20x64 matrix for P(xyz | a),
 *           for converting an amino acid into a distribution
 *           over possible codons. Incorporates an error
 *           model as well as a codon bias model.
 *
 * Derivation:
 *           call observed codon x
 *           call actual codon y
 *           call actual amino acid a.
 *           want: P(x | a)
 *            
 *           P(x,a) = \sum_y P(x,y,a)     
 *           P(x,a) = \sum_y P(x | y,a) P(y,a)
 *           P(x,a) = \sum_y P(x | y,a) P(y | a) P(a)
 *                           ^ assume base errors independent of a
 *           P(x,a) = 
 *           P(x | a) P(a) =  \sum_y P(x | y) P(y | a) P(a)                 
 *           P(x | a) = \sum_y P(x | y) P(y | a) 
 *                   error model ^        ^ codon bias model
 *                   
 * Args:     accuracy - per-base sequence accuracy
 *           aacode   - genetic code, 0..63 lookup for aa indices 0..19
 *           codebias - 0..63 codon bias probabilities per aa
 *           estmodel - RETURN: [20][64] A..YxAAA..UUU P(x | a)
 *
 * Returns:  (void)
 *           Fills in the estmodel.
 */
void
ConstructESTModel(double **estmodel)
{
  int    *aacode;       /*          */
  double *codebias;     /*          */
  double pxy;		/* P(x | y) */
  int    x, y, a;	/* indices  */

  aacode   = (int    *) MallocOrDie (sizeof(int)    * 64);
  codebias = (double *) MallocOrDie (sizeof(double) * 64);

  DefaultGeneticCode(aacode);
  DefaultCodonBias(codebias);

  for (a = 0; a < 20; a++)
    for (x = 0; x < 64; x++)
      {
	estmodel[a][x] = 0.0;
	for (y = 0; y < 64; y++)
	  {
	    if (aacode[y] != a) continue; /* genetic code, P(y|a) = 0.0 */

				/* count mutations from y to x brute force */
	    pxy =   (x / 16 == y / 16)     ? accuracy : 1.0 - accuracy;
	    pxy *=  ((x%16)/4 == (y%16)/4) ? accuracy : 1.0 - accuracy;
	    pxy *=  (x%4 == y%4)           ? accuracy : 1.0 - accuracy;
    
	    estmodel[a][x] += pxy * codebias[y];
	  }
      }

  free(aacode);
  free(codebias);
}

/* Function: ConstructHexamerProbs()
 * Date:     ER, Fri Sep 15 17:55:30 CDT 2000 [St. Louis]
 *
 * Purpose:  construct codon probs
 *
 * Args:     hexa  - 64x64 AAA..UUUxAAA..UUU joint hexamer prob matrix (prealloc)
 *           phexa -   phexa(c1,c2) = hexa(c1,c2) / [p(c1)*p(c2)] (prealloc)           
 *
 * Returns:  (void)
 *           Fills in phexa. (already allocated)
 */
void
ConstructHexamerProbs(double **hexa, double **phexa, int add_hexamer)
{
  double *pc;
  int     i;
  int     c1, c2;

  pc = (double *) MallocOrDie (sizeof(double)*64);
  for (i = 0; i < 64; i++) pc[i] = 0.0;
  
  if (!add_hexamer) 
    {
      for (c1 = 0; c1 < 64; c1++)
	for (c2 = 0; c2 < 64; c2++)
	  phexa[c1][c2] = 0.0;
      return;
    }

  /* initialization 
   */
  for (c1 = 0; c1 < 64; c1++)
    pc[c1] = 0.0;

  /* marginalize 
   */
  for (c1 = 0; c1 < 64; c1++)
    for (c2 = 0; c2 < 64; c2++) 
      pc[c1] += 0.5*EXP2(hexa[c2][c1]) + 0.5*EXP2(hexa[c1][c2]);
  
  /* check
   */
  CheckSingleProb(pc, 64);
 
  /* Transform hexamer frecuencies to 
   *
   *  phexa(c1,c2)/p(c1)*p(c2) in log2 form
   *  (these guys are odds-indepentent)
   *
   */
  for (c1 = 0; c1 < 64; c1++)
    for (c2 = 0; c2 < 64; c2++)
      phexa[c1][c2] = hexa[c1][c2] - LOG2(pc[c1]) - LOG2(pc[c2]);

  free(pc);
}


/* Function: ConstructPAMModel()
 * Date:     ER, Thu Mar  2 22:47:21 CST 2000  [St. Louis]
 *           Modified from: SRE, Wed Jun 10 08:13:35 1998 [St. Louis]
 *
 * Purpose:  Given a substitution matrix, return a
 *           PAMmodel: a 64x64 joint probability matrix
 *           for codons aligned to codons.
 *           Modification: it allows to change the "time"
 *              
 *           Sean used P(a,b, t_o) = q(a) * q(b) e^{\lambada*S_ab}
 *
 *           Here we are allowed to calculate:
 *
 *                                P(a,b, rt_o) = P(b|a, t_o)^r * P(a, t_o)
 *      
 *             where: P(b|a, t_o) = P(a,b, t_o) / P(a)
 *
 *                    P(a, t_o)   = \sum_b P(a,b, t_o)
 *
 * Args:     pam      - 27x27 integer score matrix, alphabetic
 *           scale    - scale of pam matrix           
 *           aafq     - background amino acid frequencies 0..19
 *           estmodel - 64x20 AAA..UUUxA..Y P(xyz | a) conversion to codons
 *           pammodel - 64x64 AAA..UUUxAAA..UUU joint prob matrix (alloc'ed here)
 *
 * Returns:  (void)
 *           fills in pammodel.
 */
void
ConstructPAMModel(int **pam, double scale, double tfactor, double ***ret_pammodel, int changefreq, 
		  int pedantic, int verbose)
{
  double **pammodel;
  double **pamcond;
  double  *pcod;
  double  *pm;
  double **estmodel;
  double  *aajoint;	        /* joint       aa probs P(ab)    */
  double  *Id;
  int      a, b;		/* 0..20 indices for amino acids */
  int      idxa, idxb;		/* 0..26 indices for amino acids */
  int      xyz, xyz1, xyz2;	/* indices for codons 0..63      */

   /* allocate pammodel[64][64]
   */
  estmodel    = (double **) MallocOrDie (sizeof(double *) * 20);
  pammodel    = (double **) MallocOrDie (sizeof(double *) * 64);
  pamcond     = (double **) MallocOrDie (sizeof(double *) * 64);
  estmodel[0] = (double  *) MallocOrDie (sizeof(double  ) * 20 * 64);
  pammodel[0] = (double  *) MallocOrDie (sizeof(double  ) * 64 * 64);
  pamcond[0]  = (double  *) MallocOrDie (sizeof(double  ) * 64 * 64);
  aajoint     = (double  *) MallocOrDie (sizeof(double  ) * 64 * 64);
  Id          = (double  *) MallocOrDie (sizeof(double  ) * 64 * 64);
  pcod        = (double  *) MallocOrDie (sizeof(double  ) * 64);
  pm          = (double  *) MallocOrDie (sizeof(double  ) * 4);

  Id = Cal_Id (20);

  for (a = 1; a < 20; a++) estmodel[a] = estmodel[0] + a * 64;

  for (xyz = 1; xyz < 64; xyz++) {
    pammodel[xyz] = pammodel[0] + xyz * 64;
    pamcond[xyz]  = pamcond[0]  + xyz * 64;
  }

  for (xyz = 1; xyz < 64; xyz++) 
    pcod[xyz] = 0.0;

  ConstructESTModel(estmodel);
  
 /* Convert integer pam matrix back to joint probabilities P(ab)
  */
  for (a = 0; a < 20; a++)
    {
      for (b = 0; b < 20; b++)
	{
	  idxa = Alphabet[a] - 'A';
	  idxb = Alphabet[b] - 'A';
	  aajoint[a*20+b] = aafq[a] * aafq[b] * exp((double) pam[idxa][idxb] * scale);
	}
    }
  
  /* Convert 
   *         joint probabilities aajoint(ab, t_o) 
   *   to 
   *         joint probabilities aajoint(ab, tfactor*t_o)
   */
  Joint2Joint(aajoint, Id, 20, tfactor, pedantic, verbose);
  
  /* Fill in pammodel
   *   P(xyz1,xyz2) = \sum_ab P(xyz1,xyz2 | ab) P(ab)
   *     independence assumption for converting aa to codon gives:
   *   P(xyz1,xyz2) = \sum_ab P(xyz1 | a) P(xyz2 | b) P(ab)
   *                             ^ estmodel  ^estmodel  ^aajoint  
   */
  for (xyz1 = 0; xyz1 < 64; xyz1++)
    for (xyz2 = 0; xyz2 < 64; xyz2++)
      {
	pammodel[xyz1][xyz2] = 0.0;
	for (a = 0; a < 20; a++)
	  for (b = 0; b < 20; b++)
	    pammodel[xyz1][xyz2] += 
	      estmodel[a][xyz1] * estmodel[b][xyz2] * aajoint[a*20+b];
      }

  /* Normalize it
   */
  NormalizePAMModel(pammodel);

  /* Check it
   */
  CheckPAMModel(pammodel); 

  /* To change the overal base composition,
   *
   * calculate:         P(xyz1|xyz2) = P(xyz1, xyz2) / P(xyz2)
   *
   * then use:         ^P(xyz) = f(x)*f(y)*f(z)
   *
   * to recalculate:   ^P(xyz1, xyz2) = P(xyz1|xyz2) * ^P(xyz2)
   */
  if (changefreq) {
    
    pm[0] = singlep.pa;
    pm[1] = singlep.pc;
    pm[2] = singlep.pg;
    pm[3] = 1. - pm[0] - pm[1] - pm[2];
    
    for (xyz1 = 0; xyz1 < 64; xyz1++)
      for (xyz2 = 0; xyz2 < 64; xyz2++)
	pcod[xyz1] += pammodel[xyz1][xyz2]/2.0 +  pammodel[xyz2][xyz1]/2.0;
    
    for (xyz1 = 0; xyz1 < 64; xyz1++)
      for (xyz2 = 0; xyz2 < 64; xyz2++)
	pamcond[xyz1][xyz2] = pammodel[xyz1][xyz2] / pcod[xyz2];

    for (xyz = 0; xyz < 64; xyz++)
      pcod[xyz] = pm[xyz/16] * pm[(xyz%16)/4] * pm[xyz%4];

    for (xyz1 = 0; xyz1 < 64; xyz1++)
      for (xyz2 = 0; xyz2 < 64; xyz2++)
	pammodel[xyz1][xyz2] = pamcond[xyz1][xyz2]*pcod[xyz2]/2.0 + pamcond[xyz2][xyz1]*pcod[xyz1]/2.0;
    
    NormalizePAMModel(pammodel);
    CheckPAMModel(pammodel); 
  }
  
  free(Id);
  free(pcod);
  free(aajoint);
  free(estmodel[0]);
  free(pamcond[0]);
  free(estmodel);
  free(pamcond);
  free(pm);
  
  *ret_pammodel = pammodel;
}

/* Function: ConstructPAMCondi()
 * Date:     ER, Fri Sep 22 13:56:23 CDT 2000  [St. Louis]
 *
 * Purpose:  Given a substitution matrix, return a
 *           PAMmodel: a 64x64 coditional probability matrix
 *           for codons aligned to codons.
 *
 * Args:     pam      - 27x27 integer score matrix, alphabetic
 *           scale    - scale of pam matrix           
 *           aafq     - background amino acid frequencies 0..19
 *           estmodel - 64x20 AAA..UUUxA..Y P(xyz | a) conversion to codons
 *           pammodel - 64x64 AAA..UUUxAAA..UUU joint prob matrix (alloc'ed here)
 *
 * Returns:  (void)
 *           fills in pammodel.
 */
void
ConstructPAMCondi(int **pam, double scale, double tfactor, struct pamcond_s **ret_pamcond, 
		  int pedantic, int verbose)
{
  struct pamcond_s  *pamcond;
  double            *pcodX;
  double            *pcodY;
  double           **pammodel;
  double           **estmodel;
  double            *aajoint;	        /* joint       aa probs P(ab)    */
  double            *Id;
  int                a, b;		/* 0..20 indices for amino acids */
  int                idxa, idxb;	/* 0..26 indices for amino acids */
  int                xyz, xyz1, xyz2;	/* indices for codons 0..63      */


  /* allocate estmodel[20][64]
   */
  estmodel    = (double **) MallocOrDie (sizeof(double *) * 20);
  estmodel[0] = (double  *) MallocOrDie (sizeof(double  ) * 20 * 64);

  for (a = 1; a < 20; a++) estmodel[a] = estmodel[0] + a * 64;

  ConstructESTModel(estmodel);
 
   /* allocate pamcond[64][64]
   */
  pamcond     = AllocPamCond();

  pammodel    = (double **) MallocOrDie (sizeof(double *) * 64);
  pammodel[0] = (double  *) MallocOrDie (sizeof(double  ) * 64 * 64);
  aajoint     = (double  *) MallocOrDie (sizeof(double  ) * 64 * 64);
  pcodX       = (double  *) MallocOrDie (sizeof(double  ) * 64);
  pcodY       = (double  *) MallocOrDie (sizeof(double  ) * 64);

  Id = Cal_Id (20);

  for (xyz = 1; xyz < 64; xyz++) 
    pammodel[xyz] = pammodel[0] + xyz * 64;

  for (xyz1 = 0; xyz1 < 64; xyz1++) 
    for (xyz2 = 0; xyz2 < 64; xyz2++) 
      pammodel[xyz1][xyz2] = 0.0;
  
  for (xyz = 0; xyz < 64; xyz++) {
    pcodX[xyz] = 0.0;
    pcodY[xyz] = 0.0;
  }

 /* Convert integer pam matrix back to joint probabilities P(ab)
  */
  for (a = 0; a < 20; a++)
    {
      for (b = 0; b < 20; b++)
	{
	  idxa = Alphabet[a] - 'A';
	  idxb = Alphabet[b] - 'A';
	  aajoint[a*20+b] = aafq[a] * aafq[b] * exp((double) pam[idxa][idxb] * scale);
	}
    }
  
  /* Convert 
   *         joint probabilities aajoint(ab, t_o) 
   *   to 
   *         joint probabilities aajoint(ab, tfactor*t_o)
   */
  Joint2Joint(aajoint, Id, 20, tfactor, pedantic, verbose);
  
  /* Fill in pammodel
   *   P(xyz1,xyz2) = \sum_ab P(xyz1,xyz2 | ab) P(ab)
   *     independence assumption for converting aa to codon gives:
   *   P(xyz1,xyz2) = \sum_ab P(xyz1 | a) P(xyz2 | b) P(ab)
   *                             ^ estmodel  ^estmodel  ^aajoint  
   */
  for (xyz1 = 0; xyz1 < 64; xyz1++)
    for (xyz2 = 0; xyz2 < 64; xyz2++)
      {
	pammodel[xyz1][xyz2] = 0.0;
	for (a = 0; a < 20; a++)
	  for (b = 0; b < 20; b++)
	    pammodel[xyz1][xyz2] += 
	      estmodel[a][xyz1] * estmodel[b][xyz2] * aajoint[a*20+b];
      }

  /* Normalize it
   */
  NormalizePAMModel(pammodel);

  /* Check it
   */
  CheckPAMModel(pammodel); 

  /* Calculate conditionals:   P(xyz1|xyz2) = P(xyz1, xyz2) / P(xyz2)
   *
   */
  for (xyz1 = 0; xyz1 < 64; xyz1++)
    for (xyz2 = 0; xyz2 < 64; xyz2++) {
      pcodX[xyz1] += pammodel[xyz1][xyz2];
      pcodY[xyz2] += pammodel[xyz1][xyz2];
    }
  
  for (xyz1 = 0; xyz1 < 64; xyz1++)
    for (xyz2 = 0; xyz2 < 64; xyz2++) {
      pamcond->pXY[xyz1][xyz2] = pammodel[xyz1][xyz2] / pcodY[xyz2];
      pamcond->pYX[xyz2][xyz1] = pammodel[xyz1][xyz2] / pcodX[xyz1];
    }
  
  free(Id);
  free(pcodX);
  free(pcodY);
  free(aajoint);
  free(estmodel[0]);
  free(estmodel);
  free(pammodel[0]);
  free(pammodel);

  *ret_pamcond = pamcond;
}

/* Function: ConstructPsingle()
 * Date:     ER, Mon Sep 25 17:43:48 CDT 2000  [St. Louis]
 *
 * Purpose:  Given a PAMmodel: a 64x64 coditional probability matrix
 *                             for codons aligned to codons.
 *
 * Args:     pammodel - 64x64 AAA..UUUxAAA..UUU joint prob matrix 
 *           psingle  - 4 X 4 single nucleotide frequencies (alloc'ed here)
 *
 * Returns:  (void)
 *           psingle allocated and filled here.
 */
void
ConstructPsingle (double **pammodel, double **ret_psingle)
{
  double *psingle;
  int     cod;
  int     x, y, z;

  /* allocate psingle[4]
   */
  psingle = (double *) MallocOrDie (sizeof(double) * 4);
  DSet(psingle, 4, 0.0);

  for (x = 0; x < 4; x++)
    for (y = 0; y < 4; y++)
      for (z = 0; z < 4; z++)
	for (cod = 0; cod < 64; cod++) 
	  psingle [x] += pammodel[CODON(x,y,z)][cod] + pammodel[CODON(y,x,z)][cod] +
	    pammodel[CODON(y,z,x)][cod];
  
  DScale(psingle, 4, 1./(64.*3.));

  *ret_psingle = psingle;
}

/* Function: ConstructPAMJoint()
 * Date:     ER, Fri Sep 22 14:51:09 CDT 2000 [St. Louis]
 *
 * Purpose: 
 *
 * Args:     pam      - 27x27 integer score matrix, alphabetic
 *           scale    - scale of pam matrix           
 *           aafq     - background amino acid frequencies 0..19
 *           estmodel - 64x20 AAA..UUUxA..Y P(xyz | a) conversion to codons
 *           pammodel - 64x64 AAA..UUUxAAA..UUU joint prob matrix (alloc'ed here)
 *
 * Returns:  (void)
 *           fills in pammodel.
 */
void
ConstructPAMJoint(struct pamcond_s *pamcond, double ***ret_pammodel, double *freqX, double *freqY, int pedantic, int verbose)
{
  double  *pcodX;
  double  *pcodY;
  double **pammodel;
  int      xyz, xyz1, xyz2;	/* indices for codons 0..63      */

  pammodel    = (double **) MallocOrDie (sizeof(double  *) * 64);
  pammodel[0] = (double  *) MallocOrDie (sizeof(double   ) * 64 * 64);
  pcodX       = (double  *) MallocOrDie (sizeof(double   ) * 64);
  pcodY       = (double  *) MallocOrDie (sizeof(double   ) * 64);

  for (xyz = 1; xyz < 64; xyz++) 
    pammodel[xyz] = pammodel[0] + xyz*64;
  
  
  for (xyz1 = 0; xyz1 < 64; xyz1++)  
    {
      pcodX[xyz1] = 0.0;
      pcodY[xyz1] = 0.0;

      for (xyz2 = 0; xyz2 < 64; xyz2++) 
	pammodel[xyz1][xyz2] = 0.0;
       
    }
  
  for (xyz = 0; xyz < 64; xyz++) {
    pcodX[xyz] = freqX[xyz/16] * freqX[(xyz%16)/4] * freqX[xyz%4];
    pcodY[xyz] = freqY[xyz/16] * freqY[(xyz%16)/4] * freqY[xyz%4];
  }
  
  for (xyz1 = 0; xyz1 < 64; xyz1++)
    for (xyz2 = 0; xyz2 < 64; xyz2++)
      pammodel[xyz1][xyz2] = 0.5 * pamcond->pXY[xyz1][xyz2] * pcodY[xyz2] + 0.5 * pamcond->pYX[xyz2][xyz1] * pcodX[xyz1];
  
  NormalizePAMModel(pammodel);
  CheckPAMModel(pammodel); 

  *ret_pammodel = pammodel;
  free(pcodX);
  free(pcodY);
}


/* Function: DefaultCodonBias()
 * 
 * Purpose:  Configure a codonbias table, mapping triplets to
 *           probability of using the triplet for the amino acid
 *           it represents: P(triplet | aa).
 *           The default is to assume codons are used equiprobably.
 *           
 * Args:     codebias:  0..63 array of P(triplet|aa), preallocated.
 * 
 * Return:   (void)
 */
void
DefaultCodonBias(double codebias[64])
{
  codebias[0]  = 1./2.; /* AAA Lys 2 */
  codebias[1]  = 1./2.; /* AAC Asn 2 */
  codebias[2]  = 1./2.; /* AAG Lys 2 */
  codebias[3]  = 1./2.; /* AAU Asn 2 */
  codebias[4]  = 1./4.; /* ACA Thr 4 */
  codebias[5]  = 1./4.; /* ACC Thr 4 */
  codebias[6]  = 1./4.; /* ACG Thr 4 */
  codebias[7]  = 1./4.; /* ACU Thr 4 */
  codebias[8]  = 1./6.; /* AGA Ser 6 */
  codebias[9]  = 1./6.; /* AGC Arg 6 */
  codebias[10] = 1./6.; /* AGG Ser 6 */
  codebias[11] = 1./6.; /* AGU Arg 6 */
  codebias[12] = 1./3.; /* AUA Ile 3 */
  codebias[13] = 1./3.; /* AUC Ile 3 */
  codebias[14] = 1.;    /* AUG Met 1 */
  codebias[15] = 1./3.; /* AUU Ile 3 */
  codebias[16] = 1./2.; /* CAA Gln 2 */
  codebias[17] = 1./2.; /* CAC His 2 */
  codebias[18] = 1./2.; /* CAG Gln 2 */
  codebias[19] = 1./2.; /* CAU His 2 */
  codebias[20] = 1./4.; /* CCA Pro 4 */
  codebias[21] = 1./4.; /* CCC Pro 4 */
  codebias[22] = 1./4.; /* CCG Pro 4 */
  codebias[23] = 1./4.; /* CCU Pro 4 */
  codebias[24] = 1./6.; /* CGA Arg 6 */
  codebias[25] = 1./6.; /* CGC Arg 6 */
  codebias[26] = 1./6.; /* CGG Arg 6 */
  codebias[27] = 1./6.; /* CGU Arg 6 */
  codebias[28] = 1./6.; /* CUA Leu 6 */
  codebias[29] = 1./6.; /* CUC Leu 6 */
  codebias[30] = 1./6.; /* CUG Leu 6 */
  codebias[31] = 1./6.; /* CUU Leu 6 */
  codebias[32] = 1./2.; /* GAA Glu 2 */
  codebias[33] = 1./2.; /* GAC Asp 2 */
  codebias[34] = 1./2.; /* GAG Glu 2 */
  codebias[35] = 1./2.; /* GAU Asp 2 */
  codebias[36] = 1./4.; /* GCA Ala 4 */
  codebias[37] = 1./4.; /* GCC Ala 4 */
  codebias[38] = 1./4.; /* GCG Ala 4 */
  codebias[39] = 1./4.; /* GCU Ala 4 */
  codebias[40] = 1./4.; /* GGA Gly 4 */
  codebias[41] = 1./4.; /* GGC Gly 4 */
  codebias[42] = 1./4.; /* GGG Gly 4 */
  codebias[43] = 1./4.; /* GGU Gly 4 */
  codebias[44] = 1./4.; /* GUA Val 4 */
  codebias[45] = 1./4.; /* GUC Val 4 */
  codebias[46] = 1./4.; /* GUG Val 4 */
  codebias[47] = 1./4.; /* GUU Val 4 */
  codebias[48] = 0.;    /* UAA och - */
  codebias[49] = 1./2.; /* UAC Tyr 2 */
  codebias[50] = 0.;    /* UAG amb - */
  codebias[51] = 1./2.; /* UAU Tyr 2 */
  codebias[52] = 1./6.; /* UCA Ser 6 */
  codebias[53] = 1./6.; /* UCC Ser 6 */
  codebias[54] = 1./6.; /* UCG Ser 6 */
  codebias[55] = 1./6.; /* UCU Ser 6 */
  codebias[56] = 0.;    /* UGA opa - */
  codebias[57] = 1./2.; /* UGC Cys 2 */
  codebias[58] = 1.;    /* UGG Trp 1 */
  codebias[59] = 1./2.; /* UGU Cys 2 */
  codebias[60] = 1./6.; /* UUA Leu 6 */
  codebias[61] = 1./2.; /* UUC Phe 2 */
  codebias[62] = 1./6.; /* UUG Leu 6 */
  codebias[63] = 1./2.; /* UUU Phe 2 */
}


/* Function: DefaultGeneticCode()
 * 
 * Purpose:  Configure aacode, mapping triplets to amino acids.
 *           Triplet index: AAA = 0, AAC = 1, ... UUU = 63.
 *           AA index: alphabetical: A=0,C=1... Y=19
 *           Stop codon: -1. 
 *           Uses the stdcode1[] global translation table from SQUID.
 *           
 * Args:     aacode  - preallocated 0.63 array for genetic code
 *                     
 * Return:   (void)
 */
void
DefaultGeneticCode(int aacode[64])
{
  int x;

  for (x = 0; x < 64; x++) {
    if (*(stdcode1[x]) == '*') aacode[x] = -1;
    else                       aacode[x] = SYMIDX(*(stdcode1[x]));
  }
}


void
FreeCODModel(struct codmodel_s *cod)
{
  FreeOTHModel(cod->COB);
  FreeOTHModel(cod->COJ);
  FreeOTHModel(cod->COE);

  free(cod->t);
  free(cod->pmut);

  free(cod->phexa[0]);
  free(cod->phexa);

  free(cod->pcodon[0]);
  free(cod->pcodon);

  free(cod);
}

void     
FreePamCond(struct pamcond_s *pamcond)
{

  free(pamcond->pXY[0]);
  free(pamcond->pYX[0]);

  free(pamcond->pXY);
  free(pamcond->pYX);

  free(pamcond);
}

/* Function: NormalizePAMModel()
 * Date:     SRE, Wed Jun 10 13:41:57 1998 [St. Louis]
 *
 * Purpose:  Normalize a PAM model; because of integer->double
 *           conversion, can expect some rough edges.
 *
 * Args:     pammodel[64][64]
 *
 * Returns:  void
 */
void
NormalizePAMModel(double **pammodel)
{
  int x, y;
  double sum = 0.0;

  for (x = 0; x < 64; x++)
    for (y = 0; y < 64; y++)
      sum += pammodel[x][y];
  for (x = 0; x < 64; x++)
    for (y = 0; y < 64; y++)
      pammodel[x][y] /= sum;
}


void     
PatternCODModel(struct codmodel_s *cod)
{
  int idx;
  int x;
  int codx, cody;

  PatternOTHModel(cod->COB);
  PatternOTHModel(cod->COJ);
  PatternOTHModel(cod->COE);

 /* Initialize all prob's to zero
   */
  for (idx = 0; idx < CTRANS; idx++)
    cod->t[idx] = 0.0;

  for (x = 0; x < 16; x++)
    cod->pmut[x] = 0.0;

  for (codx = 0; codx < 64; codx++) 
    for (cody = 0; cody < 64; cody++) {
      cod->pcodon[codx][cody] = 0.0;
      cod->phexa[codx][cody]  = 0.0;
    }

}


/* Function: PrintCodonString()
 * Date:     SRE, Wed Jun 10 14:00:28 1998 [St. Louis]
 */
void
PrintCodonString(int x)
{
  char *s;

  s = (char *) MallocOrDie (sizeof(char)*4);

  s[0] = DNAAlphabet[x/16];
  s[1] = DNAAlphabet[(x%16)/4];
  s[2] = DNAAlphabet[x%4];
  s[3] = '\0';
  printf("%3s ", s);

  free(s);
}


/* Function: PrintCODModel()
 * Date:     ER, Mon Jun  7 17:44:21 CDT 1999 [St. Louis]
 *
 * Purpose:  Print a cod model
 *
 * Args:     codmodel -- the codmodel prob's, in log2 form
 *
 * Returns:  void. prints transition and emission probs for cod model, in [0,1] form.
 */
void
PrintCODModel(struct codmodel_s *cod)
{
  int codx, cody;   /* symbols for emission prob's      +*/
  int idx;          /* index for transition prob's      +*/

  printf("\nCOB MODEL \n");
  PrintOTHModel(cod->COB);

  printf("\nCOJ MODEL \n");
  PrintOTHModel(cod->COJ);

  printf("\nCOE MODEL \n");
  PrintOTHModel(cod->COE);

  printf("\nCOD MODEL -- Transition probabilities\n");
  for (idx = 0; idx < CTRANS; idx++)
    printf("t[%s]\t = %f\n", ctrNAME[idx], EXP2(cod->t[idx])); 

  printf("\nCOD MODEL -- Emission probabilities\n");
  for (codx = 0; codx < 64; codx++)
    for (cody = 0; cody < 64; cody++) {
      PrintCodonString(codx);
      PrintCodonString(cody);
      printf("%3s %3s %12f \n",
	     stdcode3[codx], stdcode3[cody],
	     EXP2(cod->pcodon[codx][cody]));
    }
}


/* Function: PrintPAMModel()
 * Date:     SRE, Wed Jun 10 13:51:53 1998 [St. Louis]
 *
 * Purpose:  Print a pam model
 *
 * Args:     pammodel[64][64]
 *           mutpxy[4][4]
 *
 * Returns:  void. prints stuff.
 */
void
PrintPAMModel(double **pammodel, double *mutpxy)
{
  int      x,y;
  double   mutp;
  double   info = 0.0;
  double   expect = 0.0;

  printf("PAM probabilities\n");
  for (x = 0; x < 64; x++)
    for (y = 0; y < 64; y++)
      {
	mutp  = mutpxy[idx((x/16),y/16)];
	mutp *= mutpxy[idx(((x%16)/4),((y%16)/4))];
	mutp *= mutpxy[idx((x%4),(y%4))];

	PrintCodonString(x);
	PrintCodonString(y); 
	printf("%3s %3s %12f %12f %12f\n",
	       stdcode3[x], stdcode3[y], pammodel[x][y], mutp,
	       LOG2(pammodel[x][y] / mutp));
	
	info += pammodel[x][y] * LOG2(pammodel[x][y] / mutp);
	expect += mutp * LOG2(pammodel[x][y] / mutp);
      }

  printf("info content per codon: %f bits\n", info);
  printf("expectation per codon:  %f bits\n", expect);
}


double
normalizecodparam(int tfactor)
{
  int    i,j;
  double sum = 0.;

  for (i = 0; i < 4; i++)
    for (j = 0; j < 4; j++)
      sum += EXP2(tfactor*LOG2(CODparam.xi[i][j]));
  
  sum /= (1.0 - EXP2(tfactor*LOG2(CODparam.tau)));
  return sum;
}


