 /************************************************************
 * QRNA - Comparative analysis of biological sequences 
 *        with pair hidden Markov models and pair stochastic context-free grammars
 * Copyright (C) 2000-Howard Hughes Medical Institute
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* er_math.c
 * 
 * should go to a library
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <sys/types.h>
#include <time.h>

#include "funcs.h"
#include "globals.h"
#include "squid.h"
#include "structs.h"

#ifdef MEMDEBUG
#include "dbmalloc.h"
#endif

int
ILogSum(int *logp, int n)
{
  int x;
  float max = -1.e30;
  float sum;
  
  for (x = 0; x < n; x++)
    if ((float)logp[x] > max) max = (float)logp[x];

  sum = 0.0;
  for (x = 0; x < n; x++)
    if ((float)logp[x] > max - 50.)
      sum += exp((float)logp[x] - max);
  sum = log(sum) + max;

  return (int)sum;
}

int
ILog2Sum(int *log2p, int n)
{
  int x;
  float max = -1.e30;
  float sum;
  
  for (x = 0; x < n; x++)
    if ((float)log2p[x] > max) max = (float)log2p[x];

  sum = 0.0;
  for (x = 0; x < n; x++)
    if ((float)log2p[x] > max - 50.)
      sum += EXP2((float)log2p[x] - max);

  sum = LOG2(sum) + max;

  return (int)sum;
}

/* Functions: DChoose(), FChoose()
 *
 * Purpose:   Make a random choice from a normalized distribution.
 *            DChoose() is for double-precision vectors;
 *            FChoose() is for single-precision float vectors.
 *            Returns the number of the choice.
 */
int
DENChoose(double *p, int N)
{
  double *vec;                   /* vector to store exponetiated an normalized probs */
  double  sum = 0.;              /* integrated prob */
  int     i;                     /* counter over the probs */


  vec = (double *)MallocOrDie(sizeof(double) * N);

  /* exponentiate */
  for (i = 0; i < N; i++)
    {
      vec[i] = EXP2(p[i]);
      sum += vec[i];
    }

  /* normalize */
  for (i = 0; i < N; i++) 
    vec[i] /= sum; 

  return DChoose(vec, N);
}

int
DLog2Choose(double *p, int N)
{
  double roll;                  /* random fraction */
  double sum;                   /* integrated prob */
  int    i;                     /* counter over the probs */

  roll    = sre_random();
  sum     = 0.0;
  for (i = 0; i < N; i++)
    {
      if ((sum += EXP2(p[i])) > 1.00001) Die ("DLog2Choose() sum larger than one");
      if (roll < sum) return i;
    }
  return (int) (sre_random() * N);   /* bulletproof */
}

double
DLog2Sum(double *log2p, int n)
{
  int x;
  double max = -1.e30;
  double sum;
  
  for (x = 0; x < n; x++)
    if (log2p[x] > max) max = log2p[x];

  sum = 0.0;
  for (x = 0; x < n; x++)
    if (log2p[x] > max - 50.)
      sum += EXP2(log2p[x] - max);

  sum = LOG2(sum) + max;

  return sum;
}

void
FLog(float *vec, int n)
{
  int i;
  
  for (i = 0; i < n; i++)
    if (vec[i] > 0.0) vec[i] = log(vec[i]);
    else Die("invalid data for FLog");
}

void
FLog2(float *vec, int n)
{
  int i;
  
  for (i = 0; i < n; i++)
    if (vec[i] > 0.0) vec[i] = LOG2(vec[i]);
    else Die("invalid data for FLog2");
}

void
FExp(float *vec, int n)
{
  int i;
  
  for (i = 0; i < n; i++)
    vec[i] = exp(vec[i]);
}

void
FExp2(float *vec, int n)
{
  int i;
  
  for (i = 0; i < n; i++)
    vec[i] = exp(LN2*vec[i]);
}

void
FAddCOnst(float *vec, float plus, int n)
{
  int i;
  
  for (i = 0; i < n; i++)
    vec[i] = vec[i] + plus;
}

void
IAddConst(int *vec, int plus, int n)
{
  int i;
  
  for (i = 0; i < n; i++)
    vec[i] = vec[i] + plus;
}






