/************************************************************
 * QRNA - Comparative analysis of biological sequences 
 *        with pair hidden Markov models and pair stochastic context-free grammars
 * Copyright (C) 2000-Howard Hughes Medical Institute
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* evolve.c
 *
 * ER, Wed Mar  1 09:53:39 CST 2000 [St. Louis]
 * 
 * Purpose:  Given:
 *                       - a set of joint probabilities P(a,b| t) at a given evolutionary time t,
 *
 *                       - set of conditional probabilities at time zero Q_0
 *         
 *           Calculate:
 *                       - P(a,b| r*t) using a Markov chain model for evolution
 *  
 *
 * Method:   (1) conditionals: P(a,b | t) --> Q(a | b,t) 
 *            
 *           (2) Q(a | b, r*t) = Q_0 * exp{r * log[Q_0^{-1} * Q}
 *
 *           (3) Q(a | b, r*t) --> P(a,b | r*t)
 *
 *
 *
 * ConditionalsEvolved()  creates:    Q(a | b, r*t)   from    P(a,b | t)
 *
 * Joint2Joint()          converts:   P(a | b, t)     to      P(a,b | r*t)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <float.h>

#include "funcs.h"
#include "globals.h"
#include "squid.h"
#include "structs.h"

#ifdef MEMDEBUG
#include "dbmalloc.h"
#endif

static void isQconsistent(FILE *ofp, double *Q, int L, int pedantic, int verbose);
static void islogQconsistent(FILE *ofp, double *Q_0, double *K, int L, int pedantic, int verbose);
static void islogQconsistent_2(FILE *ofp, double *Q_0, double *A, double *K, int L, int pedantic, int verbose);

/* Function: Cal_Id()
 * Date:     ER, Thu Apr 13 10:28:16 CDT 2000 [St. Louis]
 *
 * Purpose:  Creates a Id(LxL) matrix
 *         
 * Args:    L - dimension          
 *
 * Returns: Id(LxL) = \delta_{ij}
 *          Id is allocated here, freed by caller.
 *           
 */
double *
Cal_Id(int L)
{
  double *Id;
  int     i, j;

  Id = (double *) MallocOrDie (sizeof(double) * L * L);

  for (i = 0; i < L; i++)
    for (j = 0; j < L; j++) 
      Id[i*L+j] = ((i == j) ? 1. : 0.);

  return Id;
}

/* Function: Cal_LUP_Decomp()
 * Date:     ER, Mon Apr 17 13:43:07 CDT 2000 [St. Louis]
 *
 * Purpose:  given M calculate a LUP decomposition, stored in LU
 *  
 *           (Implemented after reading chapter 31 of "introduction to algorithms"
 *            by Cormen et al.)
 *            Find matrices P, L(lower triangular), U(uppper triangular)
 *            such that P*M = L*U
 *
 *           if l_ij with i<j  are the elements of L  (l_ii = 1)
 *              u_ij with i>=j are the elements of U  (u_ii = m_ii)
 *
 *                 | l_ij for i  > j (we forget about the one's in the diagonal of L)
 *           LU  = |
 *                 | u_ij for i <= j
 *
 * Method: 
 *           Inspect by column k  
 *           Inspect rows from k on, until u_ik is a max value.
 *           Store the permutation in P.
 *           Permute rows k with i both in m and LU matrices.
 *
 *           u_ik =            m_ik - \sum_{j=1}^{j=i-1} [ l_ij * u_jk ]
 *
 *           l_ik = 1/u_kk * { m_ik - \sum_{j=1}^{j=k-1} [ l_ij * u_jk ]}
 *
 *
 *
 * Args:     M - LxL non-singular matrix.
 *
 * Returns:  LUP(LxL) = M^{-1}(LxL)
 *           LUP is alocated here, freed by caller.
 */
double *
Cal_LUP_Decomp(FILE *ofp, double *M, int *perm, int L, int verbose)
{
  double *LU;
  double *m;
  double  val;
  double  exchange;
  int     new_row, new_perm;
  int     L2;
  int     i, j, k;

  L2 = L * L;

  /* allocate LUP[L2]
   */
  LU   = (double *) MallocOrDie (sizeof(double) * L2);
  m    = (double *) MallocOrDie (sizeof(double) * L2);

  for (i = 0; i < L; i++) 
    for (j = 0; j < L; j++) {
      m[i*L+j]  = M[i*L+j];
      LU[i*L+j] = 0.;
    }

  for (i = 0; i < L; i++) 
    perm[i] = i; /* initializa to no permutations */

  /* start moving by columns
   */
  for (k = 0; k < L; k++) {

    val = 0.;
    /* For a given column move rows to have the largest possible value
     * in the diagonal.
     */
    for (i = k; i < L; i++) {
      if (fabs(m[i*L+k]) > val) {
	val = fabs(m[i*L+k]);
	new_row = i;
      }
    }
    if (val == 0.) Die( "this matrix is singular. No LUP decomposition possible");

    /* register in the P matrix that you did a permutation
     */
    new_perm = perm[k];
    perm[k]       = perm[new_row];
    perm[new_row] = new_perm;
    
    /* exchange values of the two rows both for LU and m
     */
    for (i = 0; i < L; i++) {
      exchange = m[k*L+i];
      m[k*L+i]       = m[new_row*L+i];
      m[new_row*L+i] = exchange;
      exchange = LU[k*L+i];
      LU[k*L+i]       = LU[new_row*L+i];
      LU[new_row*L+i] = exchange;
    }
    
    LU[k*L+k] = m[k*L+k];
    
    for (i = k+1; i < L; i++) {
      LU[i*L+k] = m[i*L+k]/LU[k*L+k];  /* L part of the LUP matrix */
      LU[k*L+i] = m[k*L+i];            /* U part of the LUP matrix */
    }
    
    /* redefine rest of matrix as m_ij -= l_ik * u_kj
     */
    for (i = k+1; i < L; i++) 
      for (j = k+1; j < L; j++) 
	m[i*L+j] -= LU[i*L+k]*LU[k*L+j];
	
  }
  if (verbose) {
    fprintf(ofp, "LU matrix\n");
    PrintProbs(ofp, LU, L);

    fprintf(ofp, "L * U matrix\n");
    for (i = 0; i < L; i++) 
      for (j = 0; j < L; j++) 
	m[i*L+j] = 0.;

    for (i = 0; i < L; i++) 
      for (k = 0; k < L; k++) {
	if (i <= k) m[i*L+k] += LU[i*L+k];

	for (j = 0; j <= ((i-1 < k)?i-1:k); j++) 
	  m[i*L+k] += LU[i*L+j] * LU[j*L+k];
      }
    
    PrintProbs(ofp, m, L);
    
    fprintf(ofp, "P matrix\n");
    for (i = 0; i < L; i++) {
      for (j = 0; j < L; j++) 
	fprintf(ofp, "%.4f ", ((perm[i] == j)? 1. : 0.));
      fprintf(ofp, "\n");
    }
  }

  free(m);
  return LU;
}


/* Function: Cal_M_Exp()
 * Date:     ER, Wed Mar  1 11:04:24 CST 2000 [St. Louis]
 *
 * Purpose:  Given a matrix M, calculate exp{r*M}
 *
 *            exp{rM} = \sum_{n=0}^{\infty} [ r^n * M^n / n!   ] 
 *
 * Args:     M  - LL joint prob matrix (prealloc)
 *
 * Returns:  Q(LxL) = exp{rM(LxL)}
 *           Q is alocated here, freed by caller.
 */
double *
Cal_M_Exp(FILE *ofp, double *M, int L, double r, int verbose)
{
  double *Qr;        /* Qr = exp{r*M} matrix */
  double *taylorQr;  /* next term for Qr in the taylor expansion */
  double *Mpower;    /* holds at a given n (M_I)^n */
  double  coeff;
  int     L2;
  int     stop = FALSE;
  int     i;
  int     n = 0;
  
  L2 = L * L;

  /* allocate M[L2]
   */
  Qr       = (double *) MallocOrDie (sizeof(double) * L2);
  taylorQr = (double *) MallocOrDie (sizeof(double) * L2);
  Mpower   = (double *) MallocOrDie (sizeof(double) * L2);

  /* inititalize Mpower to I and Qr to I
   */
  Mpower = Cal_Id(L);
  Qr     = Cal_Id(L);

  coeff  = 0.0; 
  while (!stop) {
    n++;

    coeff += log(r/n);
    
    /* calculate M^{n-1}*M
     */
    Mpower = Cal_M_N_Prod(ofp, Mpower, M, L, L, L, FALSE); /* multipL Mpower*M */
    
    for (i = 0; i < L2; i++) {
      taylorQr[i] = exp(coeff) * Mpower[i];
      if (abs(taylorQr[i]) > DBL_MAX || abs(Qr[i]-taylorQr[i]) > DBL_MAX) 
	Die("sorry I did not reach convergence before double-float limit in Cal_M_Exp()\n");
    }
    
    if (Check_Accuracy(taylorQr, L*L)) {
      for (i = 0; i < L2; i++) { 
	Qr[i] += taylorQr[i];
      }
    }
    else {
      stop = TRUE; 
      if (verbose) fprintf(ofp, "exp(r*K) convergence in %d terms\n", n);
    }   
  }
  
  if (verbose) {
    fprintf(ofp, "exp{r*M} matrix\n");
    PrintProbs(ofp, Qr, L);
  }

  free(Mpower);
  free(taylorQr);
  return Qr;
}

/* Function: Cal_M_Inv()
 * Date:     ER, Mon Apr 17 13:40:50 CDT 2000 [St. Louis]
 *
 * Purpose:  given M calculate M^{-1} using a LUP decomposition
 *
 * Method:   PA = LU => LUA^{-1} = P
 *
 * Args:     M^{-1} - LxL conditional prob matrix (prealloc)
 *
 * Returns:  Minv(LxL) = M^{-1}(LxL)
 *           Minv is alocated here, freed by caller.
 */
double *
Cal_M_Inv(FILE *ofp, double *M, int L, int verbose)
{
  double *Minv;     /* Minv  = M^{-1} matrix        */
  double *LU;       /* LU = L*U = PM matrix         */
  double *b;        /* LU = L*U = M matrix          */
  int    *perm;     /* P matrix                     */
  int     L2;
  int     i, j, k;
  double  sum;

  L2 = L * L;

  if (verbose ) {
    fprintf(ofp, "M matrix\n");
    PrintProbs(ofp, M, L);
 }

  /* allocate Minv[L2], LU
   */
  Minv  = (double *) MallocOrDie (sizeof(double) * L2);
  LU    = (double *) MallocOrDie (sizeof(double) * L2);
  b     = (double *) MallocOrDie (sizeof(double) * L);
  perm  = (int    *) MallocOrDie (sizeof(int)    * L);

  LU = Cal_LUP_Decomp(ofp, M, perm, L, verbose);

  for (j = 0; j < L; j++) {
    for (i = 0; i < L; i++) 
      if (j == perm[i]) b[i] = 1.;
      else b[i] = 0.;
    
    Solve_LU_Eq(ofp, LU, b, L, verbose);

    for (i = 0; i < L; i++) 
      Minv[i*L+j] = b[i];
  }
  
  if (verbose ) {
     fprintf(ofp, "M^{-1} matrix\n");
     PrintProbs(ofp, Minv, L);

    fprintf(ofp, "M * M^{-1} matrix\n");
    for (i = 0; i < L; i++) {
      for (k = 0; k < L; k++) {
	sum = 0.;
	for (j = 0; j < L; j++) 
	  sum += M[i*L+j] * Minv[j*L+k];
	
	fprintf(ofp, "%.4f ", sum);
      }
      fprintf(ofp, "\n");
    }
  }
  
  free(LU);
  free(perm);
  free(b);
  return Minv;
}

/* Function: Cal_M_Log()
 * Date:     ER, Wed Mar  1 11:04:24 CST 2000 [St. Louis]
 *
 * Purpose:  given M calculate K = log{M} using a taylor series.
 *
 *           K = \sum_{n=1}^{\infty} [ (-)^{n+1}(M-I)^n / n]
 *
 * Args:     Q  - LxL conditional prob matrix (prealloc)
 *
 * Returns:  K(LxL) = logM(LxL)
 *           K is alocated here, freed by caller.
 */
double *
Cal_M_Log(FILE *ofp, double *M, int L, int verbose)
{
  double *K;        /* K = logM matrix                         */
  double *taylorK;  /* next term for K in the taylor expansion */
  double *Mmin;     /* holds M - I                             */
  double *Mpower;   /* holds at a given n (M_I)^n              */
  double  coeff;
  int     L2;
  int     stop = FALSE;
  int     i;
  int     n = 0;

  L2 = L * L;

  /* allocate K[L2]
   */
  K       = (double *) MallocOrDie (sizeof(double) * L2);
  taylorK = (double *) MallocOrDie (sizeof(double) * L2);
  Mmin    = (double *) MallocOrDie (sizeof(double) * L2);
  Mpower  = (double *) MallocOrDie (sizeof(double) * L2);

  /* substract I from M
   */
  Mmin = Cal_M_MinId(M, L);  /* M --> Mmin = M-I */

  if (verbose) {
    fprintf(ofp, "Mmin matrix\n");
    PrintProbs(ofp, Mmin, L);
  }

  /* inititalize Mpower to I and K to zero
   */
  Mpower = Cal_Id(L);
  for (i = 0; i < L2; i++) 
    K[i] = 0.0;
  
  while (!stop) {
    n ++;
    
    Mpower = Cal_M_N_Prod(ofp, Mpower, Mmin, L, L, L, verbose); /* multiply Mpower*Mmin */
    
    coeff = (n%2==1)? 1./n : -1./n;

    for (i = 0; i < L2; i++) {
      taylorK[i] = coeff * Mpower[i];
      if (taylorK[i] > DBL_MAX) 
	Die("sorry I did not reach convergence before double-float limit in Cal_M_Log() \n");
    }
     
  if (verbose) {
    fprintf(ofp, "logQ taylor matrix\n");
    PrintProbs(ofp, taylorK, L);
  }

    if (Check_Accuracy(taylorK, L*L)) {
      for (i = 0; i < L2; i++) 
	K[i] += taylorK[i];
    }
    else {
      stop = TRUE; 
      if (verbose) fprintf(ofp, "logM convergence in %d terms\n", n);
    }   
  }
  
  if (verbose) {
    fprintf(ofp, "logM matrix\n");
    PrintProbs(ofp, K, L);
  }
  
  free(taylorK);
  free(Mmin);
  free(Mpower);
  return K;
}

/* Function: Cal_M_MinId()
 * Date:     ER, Thu Apr 13 10:30:38 CDT 2000 [St. Louis]
 *
 * Purpose:  given M calculate Mmin = M - Id.
 *
 * Args:     M  - LxL (prealloc)
 *
 * Returns:  Mmin(LxL) = M(LxL) - Id(LxL)
 *           Mmin is alocated here, freed by caller.
 */
double *
Cal_M_MinId(double *M, int L)
{
  double *MminId;
  int     i, j;

  MminId = (double *) MallocOrDie (sizeof(double) * L * L);
  
  for (i = 0; i < L; i++)
    for (j = 0; j < L; j++) 
      MminId[i*L+j] = M[i*L+j] - ((i == j) ? 1. : 0.);

  return MminId;
}

/* Function: Cal_M_N_Prod()
 * Date:     ER, Thu Apr 13 10:30:38 CDT 2000 [St. Louis]
 *
 * Purpose:  given M (LixLk) and N (LkxLj) calculate Mprod(LixLk) = M * N.
 *
 * Args:     M  - LixLk (prealloc)
 *           N  - LkxLj (prealloc)
 *
 * Returns:  Mprod(LixLj) = M(LixLk) * N(LkxLj)
 *           Mprod is alocated here, freed by caller.
 */
double *
Cal_M_N_Prod(FILE *ofp, double *M, double *N, int Li, int Lk, int Lj, int verbose)
{
  int     i, j, k;
  double  prod;
  double *Mprod;

  Mprod = (double *) MallocOrDie (sizeof(double) * Li * Lj);
  
  for (i = 0; i < Li; i++)
    for (j = 0; j < Lj; j++) {
      prod = 0.;
      for (k = 0; k < Lk; k++) {
	prod += M[i*Lk+k] * N[k*Lj+j];
	if (prod > DBL_MAX) 
	  Die("sorry, I am getting out of bounds in Cal_M_N_Prod() \n");
      }
      Mprod[i*Lj+j] = prod;
    }

  if (verbose) {
    fprintf(ofp, "M*N matrix\n");
    
    for (i = 0; i < Li; i++) {
      for (j = 0; j < Lj; j++) 
	fprintf(ofp, "%.4f ", Mprod[i*Lj+j]);
      fprintf(ofp, "\n");
    }
  }
  return Mprod;
}

/* Function: Cal_M_N_Sum()
 * Date:     ER, Mon May  8 10:44:33 CDT 2000 [St. Louis]
 *
 * Purpose:  given M (LixLj) and N (LixLj) calculate Msum(LixLj) = M + N.
 *
 * Args:     M  - LixLj (prealloc)
 *           N  - LixLj (prealloc)
 *
 * Returns:  Msum(LixLj) = M(LixLj) * N(LixLj)
 *           Msum is alocated here, freed by caller.
 */
double *
Cal_M_N_Sum(FILE *ofp, double *M, double *N, int Li, int Lj, int add, int verbose)
{
  int     i, j;
  double *Msum;
  
  Msum = (double *) MallocOrDie (sizeof(double) * Li * Lj);
  
  for (i = 0; i < Li; i++)
    for (j = 0; j < Lj; j++) 
      if (add)
	Msum[i*Lj+j] = M[i*Lj+j] + N[i*Lj+j];
      else
	Msum[i*Lj+j] = M[i*Lj+j] - N[i*Lj+j];
  
  if (verbose) {
    fprintf(ofp, "M+N matrix\n");
    
    for (i = 0; i < Li; i++) {
      for (j = 0; j < Lj; j++) 
	fprintf(ofp, "%.4f ", Msum[i*Lj+j]);
      fprintf(ofp, "\n");
    }
  }
  return Msum;
}


int
Check_Accuracy(double *vec, int L)
{
  int    i;
  int    issig;
  double m;

  issig = FALSE;
  
  for (i = 0; i < L; i ++) 
    {
      m = (vec[i] > 0)? vec[i]:-vec[i];
      if (m > (1.-accuracy)/(1.*L)) {
	issig = TRUE;
	break;
      }
    }
  
  return issig;
}

/* Function: Condi_From_Joint()
 * Date:     ER, Wed Mar  1 11:04:24 CST 2000 [St. Louis]
 *
 * Purpose:  Given       a LL matrix M(i,j) of joint probabilities, 
 *           calculate the LL matrix Q(i,j) of conditional probabilities.
 *
 *           Q[i*L+j] = Q(j|i) = P(i,j|t_o) / \sum_k  P(i,k|t_o) 
 *
 *           where P(i,j|t_o) are calculated according to 
 *           using a given pammodel t_o.
 *
 * Args:     P  - LxL joint prob matrix (prealloc)
 *
 * Returns:  Q(LxL), conditional probabilities.
 *           Q is alocated here, freed by caller.
 */
double *
Condi_From_Joint(FILE *ofp, double *P, int L, int verbose)
{
  double *Q;
  double *pm;
  double  sum;
  int     i, j;
  
  if (verbose) {
    fprintf(ofp, "P(i,j, t_o) probabilities\n");
    PrintProbs(ofp, P, L);
  }
  
  /* allocate Q[L*L]
   */
  Q  = (double *) MallocOrDie (sizeof(double) * L * L);
  pm = (double *) MallocOrDie (sizeof(double) * L * L);

  for (i = 0; i < L; i++)
    pm[i] = 0.0;
  
  for (i = 0; i < L; i++)
    for (j = 0; j < L; j++) 
      pm[i] += P[i*L+j];
  
  for (i = 0; i < L; i++) {
    sum = 0.;
    for (j = 0; j < L; j++) {
      Q[i*L+j] = P[i*L+j] / pm[i];
      sum += Q[i*L+j];
    }
    /* normalize */
    for (j = 0; j < L; j++) 
      Q[i*L+j] /= sum;
  }
 
  if (verbose) { 
    fprintf(ofp, "Q(j|i, t_o) probabilities\n");
    PrintProbs(ofp, Q, L);
  }
  
  /* consistency check, rows should add up to one
   */
  for (i = 0; i < L; i++)
    CheckSingleProb(Q+i*L, L);
  
  free (pm);
  return Q;
}

 
/* Function: ConditionalsEvolved()
 * Date:     ER, Wed Apr 12 18:11:23 CDT 2000 [St. Louis]
 *
 * Purpose:  Given a set of conditionals probabilities Q(a|b, t) 
 *           at a given evolutionary time t, 
 *           and set of conditional probabilities at time zero Q_0
 *         
 *           calculate Q(a,b| r*t) using a Markov chain model for evolution
 *  
 * Method:   Q(a | b, r*t) = Q_0 * exp{r * log[Q_0^{-1} * Q}
 *
 *
 *
 *
 * Args:              
 *
 * Returns: Qr(LxL). 
 *          Qr is allocated here, freed by caller.
 *           
 */
double *
ConditionalsEvolved(FILE *ofp, double *Q, double *Q_0, int L, double tfactor, int pedantic, int verbose)
{
  double  *Q_0_inv;
  double  *Qr;
  double  *K;
  int      i;

  /* calculate Q_0_inv
   */
  Q_0_inv = Cal_M_Inv(ofp, Q_0, L, verbose);
  
  /* construct K = log [Q_0_inv * Q]
   */
  Q = Cal_M_N_Prod(ofp, Q_0_inv, Q, L, L, L, verbose); /*  multiply Q_0_inv*Q, dump it in Q */
  
  isQconsistent(ofp, Q, L, pedantic, verbose);  /* check that Q is consistent with a markovian stationary 
						 * modelof evolution. 
						 * In plain words it checks whether logQ is going to exist.
						 *
						 * Basically, it checks whether  all entries of  
						 * are between -1 and 1. (this is not correct, not
						 * implemented right yet, april 24 2000)
						 */
  K = Cal_M_Log(ofp, Q, L, verbose);
  
  islogQconsistent(ofp, Q_0, K, L, pedantic, verbose); /*  check that K is consistent with a 
							*  markovian stationary model of evolution. 
							*
							*  Basically, it checks whether all entries of  
							*  Q_0 + epsilon*Q_0*K are none negatives.
							*/
  
  /* calculate Qr = Q_0 * exp{tfactor*K}
   */
  Qr = Cal_M_Exp(ofp, K, L, tfactor, verbose);
  
  Qr = Cal_M_N_Prod(ofp, Q_0, Qr, L, L, L, verbose); /* multiply Q_0*Qr, dump it in Qr */
  
  
  if (verbose) {
    fprintf(ofp, "Q(j|i, r*t_o)  probabilities\n");
    PrintProbs(ofp, Qr, L);
  }
  
  /* consistency check, rows should add up to one
   */	
  for (i = 0; i < L; i++)
    CheckSingleProb(Qr+i*L, L);
  
  free(K);
  free(Q_0_inv);
  
  return Qr;
}

/* Function: ConditionalsEvolved_2()
 * Date:     ER, Mon May  8 10:33:18 CDT 2000 [St. Louis]
 *
 * Purpose:  Given a set of conditionals probabilities Q(a|b, t) 
 *           at a given evolutionary time t, 
 *           and set of conditional probabilities at time zero Q_0
 *         
 *           calculate Q(a,b| r*t) 
 *           this is a more general model than the previous one:
 *  
 * Method:   Q(a | b, r*t) = Q_0 +  exp{ r * log[I + R^{-1}(Q-Q_0)] }
 *
 *           for once, the property Q(t+s) = Q(t)*Q(s) is not satisfied any more.
 *
 * Args:              
 *
 * Returns: Qr(LxL). 
 *          Qr is allocated here, freed by caller.
 *           
 */
double *
ConditionalsEvolved_2(FILE *ofp, double *Q, double *Q_0, double *R, int L, double tfactor, 
		      int pedantic, int verbose)
{
  double  *R_inv;
  double  *Qr;
  double  *K;
  double  *Id;
  int      i, j;

  Id  = (double *) MallocOrDie (sizeof(double) * L * L);
  for (i = 0; i < L; i++) 
    for (j = 0; j < L; j++) 
      Id[i*L+j] = ((i == j) ? 1. : 0.);
  
  /* calculate R_inv 
   */
  R_inv = Cal_M_Inv(ofp, R, L, verbose);
  
  /* calculate I + R_inv*(Q-Q_0)  dump it in Q 
   */
  Q = Cal_M_N_Sum(ofp, Q, Q_0, L, L, FALSE, verbose); /*  add:      Q - Q_0,   dump it in Q */
  Q = Cal_M_N_Prod(ofp, R_inv, Q, L, L, L, verbose);  /*  multiply: R_inv * Q, dump it in Q */
  Q = Cal_M_N_Sum(ofp, Id, Q, L, L, TRUE, verbose);   /*  add:      Id + Q,    dump it in Q */
  
  isQconsistent(ofp, Q, L, pedantic, verbose);  /* check that Q is consistent with a markovian stationary 
						 * modelof evolution. 
						 * In plain words it checks whether logQ is going to exist.
						 *
						 * Basically, it checks whether  all entries of  
						 * are between -1 and 1. (this is not correct, not
						 * implemented right yet, april 24 2000)
						 */
  /* construct K = log [I + R-inv*(Q-Q_0)]
   */
  K = Cal_M_Log(ofp, Q, L, verbose);
  
  islogQconsistent_2(ofp, Q_0, R, K, L, pedantic, verbose); /*  check that K is consistent with a 
							     *  markovian stationary model of evolution. 
							     *
							     *  Basically, it checks whether all entries of  
							     *  Q_0 + epsilon*R*K are none negatives.
							     */
  
  /* calculate Qr = Q_0 + R * [ exp{tfactor*K} - Id ]
   */
  Qr = Cal_M_Exp(ofp, K, L, tfactor, verbose);
  
  Qr = Cal_M_N_Sum(ofp, Qr, Id, L, L, FALSE, verbose); /* add:      Qr-Id,  dump it in Qr */
  Qr = Cal_M_N_Prod(ofp, R, Qr, L, L, L, verbose);     /* multiply: R*Qr,   dump it in Qr */
  Qr = Cal_M_N_Sum(ofp, Q_0, Qr, L, L, TRUE, verbose); /* add:      Q_0+Qr, dump it in Qr */
  
  
  if (verbose) {
    fprintf(ofp, "Q(j|i, r*t_o)  probabilities\n"); 
    PrintProbs(ofp, Qr, L);
  }
  
  /* consistency check, rows should add up to one
   */	
  for (i = 0; i < L; i++)
    CheckSingleProb(Qr+i*L, L);
  
  free(K);
  free(Id);
  free(R_inv);
  
  return Qr;
}

/* Function: HessenbergForm()
 *
 * Date:    ER, Fri Apr 21 16:15:23 CDT 2000 [St. Louis]
 *
 * Purpose: Given a real matrix M (LxL) obtain its Hessenberg form.
 *          This is an intermediate step to calculate the eigenvalues
 *          using the QL algorithm for real non-symmetric matrices.
 *
 *          The Hessenberg form has zeros everywhere below the diagonal,
 *          expect for the subdiagonal.
 *
 *
 *          Implemented after reading chapter 11 of "numerical recipies in C"
 *
 * Method:  -- pick a column k (k=0, k<L-1)
 *
 *          -- look at rows  i (i>k, i<L)
 *
 *                 find row i_o with larger value M(i_o,k)
 *
 *          -- exchange rows i_o <--> k+1
 *          -- exchange cols i_o <--> k+1
 *
 *          -- for rows  i (i=k+2, i<L)
 *
 *                 row_i <-- row_i - row_{k+1}*M(i,k)/M(k+1,k)
 *
 *             notice that for the new row_i:
 *                 M(i,k) = 0 so we have put zeros in column k, for all rows under the subdiagonal (i > k+1)
 *
 *         -- to make the elimination a similarity transformation, also reassign:
 *
 *               for rows  i (i=k+2, i<L)
 *                 col_{k+1} <-- col_{k+1} - col_{i}*M(i,k)/M(k+1,k)
 *
 * Args:              
 *
 * Returns:  H hessenberg form.
 *           H is allocated here, and freed by caller.
 *
 */
double *
HessenbergForm(FILE *ofp, double *M, int L)
{
  double *H;
  double  val;
  double  exchange;
  int     i, j, k, k1;
  int     new_row;


  H  = (double *) MallocOrDie (sizeof(double) * L * L);

  for (i = 0; i < L; i++) 
    for (j = 0; j < L; j++) 
      H[i*L+j] = M[i*L+j];

  /* start moving by columns
   */
  for (k = 0; k < L-1; k++) {

    k1 = k + 1;
    val = 0.;
    /* For a given column move rows to have the largest possible value
     * in the diagonal.
     */
    new_row = k1; /* initialize */
    for (i = k1; i < L; i++) {
      if (fabs(H[i*L+k]) > val) {
	val = fabs(H[i*L+k]);
	new_row = i;
      }
    }
    
    if (k1 != new_row) {
      for (i = 0; i < L; i++) {
	/* exchange values of the two rows (k+1 and new_row) 
	 */
	exchange = H[k1*L+i];
	H[k1*L+i]   = H[new_row*L+i];
	H[new_row*L+i] = exchange;
      }
      for (i = 0; i < L; i++) {
	/* also exchange values for the columns (k+1 and new_row)
	 */
	exchange = H[i*L+k1];
	H[i*L+k1] = H[i*L+new_row];
	H[i*L+new_row] = exchange;
      }
    }
    
    if (val != 0.) {
      for (i = k1+1; i < L; i++) {
	for (j = 0; j < L; j++) 
	  H[i*L+j]  -= H[i*L+k] * H[k1*L+j] / val;
	for (j = 0; j < L; j++) 
	  H[j*L+k1] += H[i*L+k] * H[j*L+i] / val;
     }
    }
  } /* for every column k */
  
  return H;
}

/* Function: Hessenberg2Eigenvalues()
 *
 * Date:    ER, Fri Apr 21 21:07:27 CDT 2000 [St. Louis]
 *
 * Purpose: Given a real non-symmetric matrix M (LxL) in its Hessenberg form
 *          calculate its eigenvalues using the QL algorithm.
 *
 *          Implemented after reading chapter 11 of "numerical recipies in C"
 *
 * Args:    H -- (LxL) Hessenberg matrix          
 *
 * Returns: eigen(2L) vector with real and imaginary part of eigenvalues
 *
 */
void
Hessenberg2Eigenvalues(FILE *ofp, double *H, int L, double **ret_vr, double **ret_vi)
{
  double *vr, *vi;
  int     i;

  vr = (double *) MallocOrDie (sizeof(double) * L);
  vi = (double *) MallocOrDie (sizeof(double) * L);

  for (i = 0; i < L; i++) {
    vr[i] = 0.;
    vi[i] = 0.;
  }
  
  *ret_vr = vr;
  *ret_vi = vi;
}

/* Function: Joint2Joint()
 * Date:     ER, Wed Apr 12 18:11:23 CDT 2000 [St. Louis]
 *
 * Purpose:  Given a set of joint probabilities P(a,b| t_o) 
 *           at a given evolutionary time t_o, 
 *           and set of conditional probabilities at time zer0 P^o
 *         
 *           calculate P(a,b| tfactor*t_o) using a Markov chain model for evolution
 *  
 * Method:   (1) conditionals: P(a,b | t_o) --> Q(a | b,t_o) 
 *            
 *           (2) Q(a | b, r*t_o) = Q_0 * exp{ r * log[Q_0^{-1} * Q] }
 *
 *           (3) Q(a | b, r*t_o) --> P(a,b | r*t_o)
 *
 * Args:              
 *
 * Returns:  (void)
 *           
 */
void
Joint2Joint(double *p_joint, double *q_nought, int L, double tfactor, int pedantic, int verbose)
{
  double *pm;
  double *p_cond;
  double  sum;
  int     i, j;
  
  if (tfactor == 1.) return;

  /* allocate p[L]
   */
  pm     = (double *) MallocOrDie (sizeof(double) * L);
  p_cond = (double *) MallocOrDie (sizeof(double) * L * L);

  for (i = 0; i < L; i++)
    pm[i] = 0.;
  
  sum = 0.;
  for (i = 0; i < L; i++) {
    for (j = 0; j < L; j++) 
      pm[i] += p_joint[i*L+j];
    sum += pm[i];
  }

  /* normalize */
  if (verbose) printf("marginalized probabilities\n");
  for (i = 0; i < L; i++) {
    pm[i] /= sum;
    if (verbose) printf("%f ", pm[i]);
  }
  if (verbose) printf("\n");
  
  /* calculate the matrix of conditional probabilities (Q) from joint probabilities (P)
   */
  p_cond = Condi_From_Joint(stdout, p_joint, L, verbose);
  
  /* evolve conditionals
   */
  p_cond = ConditionalsEvolved(stdout, p_cond, q_nought, L, tfactor, pedantic, verbose);
  
  for (i = 0; i < L; i++)
    for (j = 0; j < L; j++) 
      p_joint[i*L+j] = p_cond[i*L+j] * pm[i];

  CheckSingleProb(p_joint, L*L);
  
  free(pm);
  free(p_cond);
}

/* Function: Solve_LUP_Eq()
 * Date:     ER, Mon Apr 17 17:22:31 CDT 2000 [St. Louis]
 *
 * Purpose:  Given LU x = b solve for x
 *  
 *           (Implemented after reading chapter 31 of "introduction to algorithms"
 *            by Cormen et al.)
 *
 * Method:   (1) solve equation  Ly = b for y
 *                     trivial because L is a lower triangular matrix.
 *
 *                     y_0 = b_0/l_00 = b_0
 *
 *                     y_i = 1/l_ii [ b_i - \sum_{j=1}^{i-1} l_ij*y_j ] for i=2,...,L-1
 *
 *                         = b_i - \sum_{j=1}^{i-1} l_ij*y_j 
 *
 *           (2) solve equation  Ux = y for x
 *                     trivial because U is an upper triangular matrix.
 *
 *                     x_{L-1} = y_{L-1}/u_{L-1,L-1}
 *
 *                     x_i = 1/u_ii [ y_i - \sum_{j=i+1}^{L-1} u_ij*x_j ] for i=L-2,...,0
 *
 *
 * Args:     LU - LxL LU  matrix.
 *
 * Returns:  (void)
 *           values for x are dumped in b.
 */
void
Solve_LU_Eq(FILE *ofp, double *LU, double *b, int L, int verbose)
{
  int     i, j;
  double  val, add;

  /* calculate y (dump values into x)
   */
  for (i = 1; i < L; i++) {
    val = b[i];
    add = 0;
      for (j = 0; j < i; j++)
	{
	  add -= LU[i*L+j] * b[j];
	}
    val += add;
    b[i] = val;
  }

  /* calculate x (dump values into x)
   */
  b[L-1] /= LU[(L-1)*L+L-1];
  for (i = L-2; i >= 0; i--) {
    val = b[i];
    add = 0;
      for (j = i+1; j < L; j++)
	{
	  add -= LU[i*L+j] * b[j];
	}
    val += add;
    b[i] = val / LU[i*L+i];
  }
}

/* Function: isQconsistent()
 * Date:     ER, Thu Apr 20 14:19:42 CDT 2000 [St. Louis]
 *
 * Purpose:  Q(t) ---a matrix of conditional probabilities---
 *           is consistent with being derived from a markovian stationary
 *           model of evolution from:
 *                    Q_0 - a matrix of conditional probabilities at time zero
 *                    K - a matrix of rate of evolution
 *
 *           as
 *                     Q(t) = Q_0 * exp {t K}
 *
 *
 *           IF: Q_0 + epsilon * Q_0*K  can play the role of an "instant matrix of evolution"
 *
 *               that is, if all the entries are positive.
 *
 * Args:      Q_0 - LxL matrix of conditional prob matrix at time zero (prealloc)
 *           logQ - LxL log matrix of conditional prob matrix (prealloc)
 *
 * Returns:  1 if consistent, 0 otherwise.
 */
void
isQconsistent(FILE *ofp, double *Q, int L, int pedantic, int verbose)
{
  double *H;
  double *vr, *vi;
  int     i, j;

  H = HessenbergForm(ofp, Q, L);
  Hessenberg2Eigenvalues(ofp, H, L, &vr, &vi);

  if (verbose) {
      fprintf(ofp, "Q matrix\n"); 
      PrintProbs(ofp, Q, L);

      fprintf(ofp, "Hessenberg form\n"); 
      PrintProbs(ofp, H, L);
      
      fprintf(ofp, "Eigenvalues for Q\n"); 
      for (i = 0; i < L; i++) 
	fprintf(ofp, "%.4f %.4f \n", vr[i], vi[i]);
  }
  
  /* this is not true. the real condition is that the eigenvalues of Q are of norm smaller
   * than one. Still working on the implementation of Hessenberg2Eigenvalues()
   */
  for (i = 0; i < L; i++)
    for (j = 0; j < L; j++)
      if (pedantic && fabs(Q[i*L+j]) > 1.) 
	Warn ("Q not consistent with a markovian stationary model of evolution");
  
  free(H);
  free(vr);
  free(vi);
}
  
/* Function: islogQconsistent()
 * Date:     ER, Fri Apr 14 12:06:24 CDT 2000 [St. Louis]
 *
 * Purpose:  Q(t) ---a matrix of conditional probabilities---
 *           is consistent with being derived from a markovian stationary
 *           model of evolution from:
 *                    Q_0 - a matrix of conditional probabilities at time zero
 *                    K - a matrix of rate of evolution
 *
 *           as
 *                     Q(t) = Q_0 * exp {t K}
 *
 *
 *           IF: Q_0 + epsilon * Q_0*K  can play the role of an "instant matrix of evolution"
 *
 *               that is, if all the entries are positive.
 *
 * Args:      Q_0 - LxL matrix of conditional prob matrix at time zero (prealloc)
 *           logQ - LxL log matrix of conditional prob matrix (prealloc)
 *
 * Returns:  void.
 */
void
islogQconsistent(FILE *ofp, double *Q_0, double *K, int L, int pedantic, int verbose)
{
  double *R;
  double  epsilon = 0.3;
  int     i, j;

  /* allocate p[L]
   */
  R = (double *) MallocOrDie (sizeof(double) * L * L);

  R = Cal_M_N_Prod(ofp, Q_0, K, L, L, L, verbose); /*  multiply Q_0*K, dump it in R */

  for (i = 0; i < L; i++)
    for (j = 0; j < L; j++)
      R[i*L+j] = Q_0[i*L+j] + epsilon * R[i*L+j];

  for (i = 0; i < L; i++)
    for (j = 0; j < L; j++)
      if (R[i*L+j] < 0. && pedantic) 
	Warn ("Ah! log[Q_0^{-1}*Q] not consistent with a markovian stationary model of evolution (%f).", R[i*L+j]); 								     
  
  if (verbose) { 
    fprintf(ofp, "R Instant evolution probabilities\n");
    PrintProbs(ofp, R, L);
  }
  
  free(R);
}

/* Function: islogQconsistent_2()
 * Date:     ER, Mon May  8 12:55:29 CDT 2000 [St. Louis]
 *
 * Purpose:  Q(t) ---a matrix of conditional probabilities---
 *           is consistent with being derived from a markovian stationary
 *           model of evolution from:
 *                    Q_0 - a matrix of conditional probabilities at time zero
 *                    A - a matrix of rate of evolution
 *                    K - a matrix of rate of evolution
 *
 *           as
 *                     Q(t) = Q_0 + A[ exp{t K} - Id ]
 *
 *
 *           IF: Q_0 + epsilon * A*K  can play the role of an "instant matrix of evolution"
 *
 *               that is, if all the entries are positive.
 *
 * Args:     Q_0  - LxL matrix of conditional prob matrix at time zero (prealloc)
 *           A    - LxL matrix  (prealloc)
 *           logQ - LxL log matrix of conditional prob matrix (prealloc)
 *
 * Returns:  void.
 */
void
islogQconsistent_2(FILE *ofp, double *Q_0, double *A, double *K, int L, int pedantic, int verbose)
{
  double *R;
  double  epsilon = 0.3;
  int     i, j;

  /* allocate p[L]
   */
  R = (double *) MallocOrDie (sizeof(double) * L * L);

  R = Cal_M_N_Prod(ofp, A, K, L, L, L, verbose); /*  multiply A*K, dump it in R */

  for (i = 0; i < L; i++)
    for (j = 0; j < L; j++)
      R[i*L+j] = Q_0[i*L+j] + epsilon * R[i*L+j];

  for (i = 0; i < L; i++)
    for (j = 0; j < L; j++)
      if (R[i*L+j] < 0. && pedantic) 
	Warn ("Ah! logQ not consistent with a markovian stationary model of evolution (%f).", R[i*L+j]); 								     
  
  if (verbose) { 
    fprintf(ofp, "R Instant evolution probabilities\n");
    for (i = 0; i < L; i++) {
      for (j = 0; j < L; j++) {
	fprintf(ofp, "%.4f ", R[i*L+j]);
      }
      fprintf(ofp, "\n");
    }
  }
  
  free(R);
}


  
  

