/************************************************************
 * QRNA - Comparative analysis of biological sequences 
 *        with pair hidden Markov models and pair stochastic context-free grammars
 * Copyright (C) 2000-Howard Hughes Medical Institute
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* globals.c
 *
 *
 * E. Rivas [St. Louis]
 * 
 * 9 april 1999.
 */
                                            
#include <stdio.h>

#include "funcs.h"
#include "globals.h"
#include "squid.h"
#include "structs.h"

#ifdef MEMDEBUG
#include "dbmalloc.h"
#endif

char Alphabet[]    = AMINO_ALPHABET;
char DNAAlphabet[] = DNA_ALPHABET;
char SSAlphabet[]  = "123456789abcdefghijklmnopqrstuvwxyz";

/**********************************************************************
 * Base Composition
 **********************************************************************/
struct singlep_s 
singlep = { 
  0.33, /*+ p[a]   +*/
  0.17, /*+ p[c]   +*/
  0.17  /*+ p[g]   +*/
}; 

/**********************************************************************
 * NULL-model
 **********************************************************************/
struct nullparam_s 
Nullparam = { 0.35 /*+ eta   +*/}; 

/**********************************************************************
 * OTHER-model
 **********************************************************************/
struct othparam_s 
OTHparam = {
  {0.56, /*+ FLN eta   +*/},
  {0.56, /*+ FJN eta   +*/},
  {0.56, /*+ FRN eta   +*/},

  0.195,  /*+ delta   +*/
  0.33,  /*+ epsilon +*/ 
  0.90,  /*+ eta     +*/ 
  0.005, /*+ gamma   +*/  
  0.19,  /*+ kappa   +*/
  0.20,  /*+ tau     +*/  
  0.45   /*+ xi      +*/
};

/* EmitsPerOTHState tells how many emissions there are for each state 
 * (1 if no emissions)
 */
int EmitsPerOTHState[OSTATES] = {
/* stFL   stB   stM   stX   stY   stE   stFJ   stFR     */
     1,    1,    16,   4,    4,    1,    1,     1
};

/* IdxTransOTHState tells where the transition indices begin for a given state.
 * (1 if no emissions)
 */
int IdxTransOTH[OSTATES]= {
/* stFL   stB   stM   stX   stY   stE   stFJ   stFR     */
    0,     2,    5,    9,    13,   17,   19,    20   
};

/* OConnects defines which states connect to each other in othermodel.
 * The order must be the same as in cfg.h, as this is indexed
 * by stB, stN, etc. #define's.
 */
int OConnects[OSTATES][OSTATES] = {
/* from:  to: stFL  stB   stM   stX   stY   stE  stFJ  stFR      */
/* stFL  */ {  0,    1,    0,    0,    0,    0,   0,    1},
/* stB   */ {  0,    0,    1,    1,    1,    0,   0,    0},
/* stM   */ {  0,    0,    1,    1,    1,    1,   0,    0},
/* stX   */ {  0,    0,    1,    1,    1,    1,   0,    0},
/* stY   */ {  0,    0,    1,    1,    1,    1,   0,    0},
/* stE   */ {  0,    0,    0,    0,    0,    0,   1,    1},
/* stFJ  */ {  0,    1,    0,    0,    0,    0,   0,    0},
/* stFR  */ {  0,    0,    0,    0,    0,    0,   0,    0}
};

/* The NAME arrays convertsinteger index values to strings
 */
char *ostNAME[OSTATES] = { 
  "FL", "B", "M", "X", "Y", "E", "FJ",  "FR"
};
char *otrNAME[OTRANS] = { 
  "FL->B", "FL->FR",
  "B-> M", "B-> X",  "B-> Y", 
  "M-> M", "M-> X",  "M-> Y",  "M-> E", 
  "X-> M", "X-> X",  "X-> Y",  "X-> E", 
  "Y-> M", "Y-> X",  "Y-> Y",  "Y-> E", 
  "E->FJ", "E->FR",
  "FJ->B", 
  "FR->FR"
};

/* TransPerOTHState tells how many non-zero transitions there are for each state 
 * (1 if no emissions)
 */
int TransPerOTHState[OSTATES]= {
/* stFR   stB   stM   stX   stY   stE   stFJ   stFR      */    
    2,     3,    4,    4,    4,    2,    1,     1         
};


/**********************************************************************
 * COD-model
 **********************************************************************/
struct codparam_s 
CODparam = {
 0.55,  /*+ phi = COB -> Cb   +*/
 0.45,  /*+ eta = Ce  -> COE  +*/
 0.00,  /*+ tau = Cb  -> Ce   +*/

/* parameters for transitions to codons. 
 * We allow codons of either (3,3) (3,4) (3,2),..., and (3,0) (2,0) (4,0) nts
 * paramters will be normalized later on.
 */
 /*+ xi:          3         4          2          0         +*/
 /*+ -------------------------------------------------------+*/
 /*+ 3 +*/{{    110.00,     0.00,      0.50,      0.05      },
 /*+ 4 +*/ {      0.00,     0.00,      0.00,      0.00      },
 /*+ 2 +*/ {      0.50,     0.00,      0.00,      0.00      },
 /*+ 0 +*/ {      0.05,     0.00,      0.00,      0.00      }}
};

struct othparam_s 
COBparam = {
  {0.58, /*+ FLN eta   +*/},
  {0.58, /*+ FJN eta   +*/},
  {0.58, /*+ FRN eta   +*/},

  0.20,  /*+ delta   +*/
  0.33,  /*+ epsilon +*/ 
  0.90,  /*+ eta     +*/ 
  0.005, /*+ gamma   +*/  
  0.20,  /*+ kappa   +*/
  0.53,  /*+ tau     +*/  
  0.25   /*+ xi      +*/
};

struct othparam_s 
COJparam = {
  {0.60, /*+ FLN eta   +*/},
  {0.60, /*+ FJN eta   +*/},
  {0.60, /*+ FRN eta   +*/},

  0.10,  /*+ delta   +*/
  0.10,  /*+ epsilon +*/ 
  0.90,  /*+ eta     +*/ 
  0.005, /*+ gamma   +*/  
  0.50,  /*+ kappa   +*/
  0.79,  /*+ tau     +*/  
  0.05   /*+ xi      +*/
};

struct othparam_s 
COEparam = {
  {0.58, /*+ FLN eta   +*/},
  {0.58, /*+ FJN eta   +*/},
  {0.58, /*+ FRN eta   +*/},

  0.20,  /*+ delta   +*/
  0.33,  /*+ epsilon +*/ 
  0.90,  /*+ eta     +*/ 
  0.005, /*+ gamma   +*/  
  0.20,  /*+ kappa   +*/
  0.53,  /*+ tau     +*/  
  0.25   /*+ xi      +*/
};                

/* EmitsPerCODState tells how many emissions there are for each state 
 * (1 if no emissions)
 */
int EmitsPerCODState[CSTATES] = {
/* stCOB stCb stC33   stC34     stC32    stC43     stC44      stC42     stC23   stC24    stC22    */
    1,    1,   64*64,  64*256,   64*16,   256*64,   256*256,   256*16,   16*64,  16*256,  16*16,        
/* stC30  stC40   stC20  stC03   stC04    stC02  stCe  stCOJ  stCOE  */
    64,    256,    16,    64,     256,     16,    1,    1,     1 
};

/* IdxTransCODState tells where the transition indices begin for a given state.
 * (1 if no emissions)
 */
int IdxTransCOD[CSTATES] = {
/* stCOB stCb stC33  stC34   stC32  stC43   stC44    stC42   stC23  stC24   stC22    */
    0,    2,   18,    19,     20,    21,     22,      23,     24,    25,     26,        
/* stC30  stC40   stC20  stC03   stC04    stC02   stCe   stCOJ stCOE  */
    27,    28,     29,    30,     31,      32,     33,    35,   36   
};

/* The NAME arrays convert integer index values to strings
 */
char *cstNAME[CSTATES] = { 
  "COB", "Cb", 
  "C33", "C34", "C32", "C43", "C44", "C42", "C23", "C24", "C22", 
  "C30", "C40", "C20", "C03", "C04", "C02", 
  "Ce",  "COJ", "COE"
};
char *ctrNAME[CTRANS] = { 
  "COB->Cb", "COB->COE", 
  "Cb->C33", "Cb->C34", "Cb->C32", "Cb->C43", "Cb->C44", "Cb->C42", "Cb->C23", "Cb->C24", "Cb->C22", 
  "Cb->C30", "Cb->C40", "Cb->C20", "Cb->C03", "Cb->C04", "Cb->C02",
   "Cb->Ce",   
  "C33->Ce", "C34->Ce", "C32->Ce", "C43->Ce", "C44->Ce", "C42->Ce", "C23->Ce", "C24->Ce", "C22->Ce", 
  "C30->Ce", "C40->Ce", "C20->Ce", "C03->Ce", "C04->Ce", "C02->Ce",   
  "Ce->COJ", "Ce->COE", "COJ->Cb", "COE->COE" 
};

/* TransPerCODState tells how many non-zero transitions there are for each state 
 * (1 if no emissions)
 */
int TransPerCODState[CSTATES]= {
/* stCOB stCb stC33  stC34   stC32  stC43   stC44    stC42   stC23  stC24   stC22    */
    2,    16,  1,     1,      1,     1,      1,       1,      1,     1,      1,        
/* stC30  stC40   stC20  stC03   stC04    stC02   stCe stCOJ stCOE  */
    1,     1,      1,     1,      1,       1,      2,   1,    1
};


/**********************************************************************
 * RNA-model
 **********************************************************************/
struct rnaparam_s 
RNAparam = {
  0.65,  /*+ phi = ROB  -> RNA  +*/
  0.35,  /*+ eta = RNA  -> ROE  +*/
};

struct othparam_s 
Rloopparam = {
  {0.45, /*+ FLN eta   +*/},
  {0.45, /*+ FJN eta   +*/},
  {0.45, /*+ FRN eta   +*/},

  0.08,  /*+ delta   +*/
  0.13,  /*+ epsilon +*/ 
  0.90,  /*+ eta     +*/ 
  0.005, /*+ gamma   +*/  
  0.12,  /*+ kappa   +*/
  0.19,  /*+ tau     +*/  
  0.30   /*+ xi      +*/
};

struct othparam_s 
ROBparam = {
  {0.56, /*+ FLN eta   +*/},
  {0.56, /*+ FJN eta   +*/},
  {0.56, /*+ FRN eta   +*/},

  0.20,  /*+ delta   +*/
  0.33,  /*+ epsilon +*/ 
  0.90,  /*+ eta     +*/ 
  0.005, /*+ gamma   +*/  
  0.20,  /*+ kappa   +*/
  0.20,  /*+ tau     +*/  
  0.40   /*+ xi      +*/
};

struct othparam_s 
ROJparam = {
  {0.58, /*+ FLN eta   +*/},
  {0.48, /*+ FJN eta   +*/},
  {0.58, /*+ FRN eta   +*/},

  0.11,  /*+ delta   +*/
  0.10,  /*+ epsilon +*/ 
  0.90,  /*+ eta     +*/ 
  0.005, /*+ gamma   +*/  
  0.11,  /*+ kappa   +*/
  0.55,  /*+ tau     +*/  
  0.05   /*+ xi      +*/
};

struct othparam_s 
ROEparam = {
  {0.56, /*+ FLN eta   +*/},
  {0.56, /*+ FJN eta   +*/},
  {0.56, /*+ FRN eta   +*/},

  0.20,  /*+ delta   +*/
  0.33,  /*+ epsilon +*/ 
  0.90,  /*+ eta     +*/ 
  0.005, /*+ gamma   +*/  
  0.20,  /*+ kappa   +*/
  0.20,  /*+ tau     +*/  
  0.40   /*+ xi      +*/
};

/* IdxTransRNAState tells where the transition indices begin for a given state.
 * (1 if no emissions)
 */
int IdxTransRNA[RSTATES]= {
/* stROB  stRNA  stROJ  stROE    */
    0,     2,     4,     5    
};

/* The NAME arrays converts integer index values to strings
 */
char *rstNAME[RSTATES] = { 
  "ROB", "RNA", "ROJ", "ROE"
};
char *rtrNAME[RTRANS] = { 
  "ROB->RNA", "ROB->ROE", "RNA->ROJ", "RNA->ROE" , "ROJ->RNA" , "ROE->ROE" 
};

/* TransPerRNAState tells how many non-zero transitions there are for each state 
 * (1 if no emissions)
 */
int TransPerRNAState[RSTATES]= {
/* stROB  stRNA   stROJ   stROE       */
    2,     2,      1,      1  
};



/**********************************************************************
 * CFG_RNA-model
 **********************************************************************/
/* DpNodeidx tells where the node indices begin for a given state type.
 * This has to match up with idxP, etc. in cfg.h
 */
int DpNodeidx[NDPS] = {
/*  V     W     WB    IS1,  IS2         */
    0,    6,    6,    11,   11    
};


/* Idx tells how many states are included under each "state name"
 * (different from 1 for states that "remember" the previous emissions (ie V_{a,b}, VH_{a,b})
 */
int Idx[NDPS] = {
/*  V     W     WB     IS1   IS2      */
    16,   1,    1,     1,    1       
};

/* NodesPerDp tells how many nodes there are for each RNA-dp state
 *  (it does not include the dummy node)
 */
int NodesPerDp[NDPS] = {
/*  V     W     WB    IS1         IS2     */
    5,    4,    4,    MAXRNALOOP, MAXRNALOOP 
};

char *stNAME[NDPS] = { 
  "V", "W", "WB", "IS1", "IS2"
};

/* Ntrans tells how many transitions there are for each state
 */
int Ntrans[NDPS] = {
/*  V        W        WB       IS1       IS2     */
  NTRANSV, NTRANSW, NTRANSW, NTRANSIS, NTRANSIS
};


/* TransPerDp tells how many transitions there are for each dp state
 */
int TransPerDp[NDPS] = {
/*  V          W          WB         IS1         IS2     */
  NTRANSVst, NTRANSW-1, NTRANSW-1, NTRANSIS-1, NTRANSIS-1
};






