/************************************************************
 * QRNA - Comparative analysis of biological sequences 
 *        with pair hidden Markov models and pair stochastic context-free grammars
 * Copyright (C) 2000-Howard Hughes Medical Institute
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* nullmodel.c
 *
 * ER,  Fri Jun  4 12:11:35 CDT 1999 [St. Louis]
 * 
 * Allocation, free'ing, initialization of the nullmodel
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

#include "funcs.h"
#include "globals.h"
#include "squid.h"
#include "structs.h"

#ifdef MEMDEBUG
#include "dbmalloc.h"
#endif

/* Function: AllocNullModel()
 * Date:     ER,  Fri Jun  4 11:59:28 CDT 1999 [St. Louis]
 *
 * Purpose:  Allocates memory for the transition and emission probs of the nullmodel
 *
 * Args:     nullmodel - nullmodel structure
 *
 * Returns:  void
 *           allocates null->xem[], null->yem[], which must be free'd by caller.
 */
struct nullmodel_s *
AllocNullModel(void)
{
  struct nullmodel_s *null;    /* nullmodel structure         */

  null      = (struct nullmodel_s *) MallocOrDie (sizeof(struct nullmodel_s));
  null->xem = (double *) MallocOrDie (sizeof(double) * 4);
  null->yem = (double *) MallocOrDie (sizeof(double) * 4);
  
  PatternNullModel(null);
  
  return null;
}

/* Function: CheckNullProbs()
 * Date:     ER, Fri Jun  4 12:07:20 CDT 1999  [St. Louis]
 *
 * Purpose:  Verify that transition and emission prob's of a nullmodel add up to one
 *
 * Args:     other - the structure for an othermodel
 *
 * Returns:  void. 
 */
void
CheckNullProbs(struct nullmodel_s *null)
{
  /* check emission prob's add up to one
   */
  CheckSingleProb(null->xem, 4);
  CheckSingleProb(null->yem, 4);
}

/* Function: ConstructNullModel()
 * Date:     ER, Fri Jun  4 11:29:08 CDT 1999 [St. Louis]
 *
 * Purpose:  Constructs a nullmodel_s
 *
 * Args:     otherparam - the list of parameters that define a othermodel           
 *
 * Returns:  (void)
 *           fills all prob's for nullmodel, log2 form 
 *           (allc'ed here, freed by caller)
 */
void
ConstructNullModel(double *nullpxy, struct nullmodel_s *null, struct nullparam_s param)
{
  null->eta  = param.eta;
  null->meta = 1. - param.eta;

  /* Calculate emission prob's: null.xem[], null.yem[]
   */
  ConstructXemNullProb(nullpxy, null->xem);
  ConstructYemNullProb(nullpxy, null->yem);
  
  /* check prob's add up to one
   */
  CheckNullProbs(null);

  /* convert to LOG2 form
   */
  NullToLog2(null);
}

/* Function: ConstructXemNullProb()
 * Date:     ER, Fri Jun  4 11:54:37 CDT 1999 [St. Louis]
 *
 * Purpose:  Given a nullpxy[4][4] nullmodel joint probability distrubution, 
 *           marginalize to calculate null->xem[4].
 *
 * Args:    nullpxy - 4x4 A..UxA..U joint prob matrix (prealloc)
 *          xem     - 4 seqX null model emission prob (prealloc)
 *
 * Returns:  (void)
 *           Fills in null->xem (already allocated)
 */
void
ConstructXemNullProb(double *nullpxy, double *xem)
{
  int x;
  int y;

  /* Zero null model
   */
  for (x = 0; x < 4; x++)
    xem[x] = 0.0;

  /* Marginalize and average over Y positions
   */
  for (x = 0; x < 4; x++)
    for (y = 0; y < 4; y++)
      xem[x] += nullpxy[idx(x,y)];

  CheckSingleProb(xem, 4);
}

/* Function: ConstructYemNullProb()
 * Date:     ER, Fri Jun  4 11:54:37 CDT 1999 [St. Louis]
 *
 * Purpose:  Given a nullpxy[4][4] nullmodel joint probability distrubution, 
 *           marginalize to calculate null->yem[4].
 *
 * Args:    nullpxy - 4x4 A..UxA..U joint prob matrix (prealloc)
 *          yem     - 4 seqX null model emission prob (prealloc)
 *
 * Returns:  (void)
 *           Fills in null->yem (already allocated)
 */
void
ConstructYemNullProb(double *nullpxy, double *yem)
{
  int x;
  int y;

  /* Zero null model
   */
  for (x = 0; x < 4; x++)
    yem[x] = 0.0;

  /* Marginalize and average over Y positions
   */
  for (x = 0; x < 4; x++)
    for (y = 0; y < 4; y++)
      yem[x] += nullpxy[idx(y,x)];

  CheckSingleProb(yem, 4);
}

void
FreeNullModel(struct nullmodel_s *null)
{
  free(null->xem);
  free(null->yem);
  free(null);
}

double
NullAddOnePair(int cur_x, int cur_y, struct nullmodel_s *null)
{
  double sc;

  sc = 0.;

  if (cur_x < 4) sc += null->xem[cur_x] + null->meta;
  if (cur_y < 4) sc += null->yem[cur_y] + null->meta;
  
  return sc;
}

double
NullAddOneX(int cur_x, struct nullmodel_s *null)
{
  double sc;

  if (cur_x < 4) sc = null->xem[cur_x] + null->meta;
  
  return sc;
}

double
NullAddOneY(int cur_y, struct nullmodel_s *null)
{
  double sc;

  if (cur_y < 4) sc = null->yem[cur_y] + null->meta;
  
  return sc;
}

/* Function: OtherToLog2()
 * Date:     ER, Fri Jun  4 12:16:17 CDT 1999 [St. Louis]
 *
 * Purpose:  Converts transition and emission prob's of a nullmodel to log2 form
 *
 * Args:     null - the structure for an nullmodel
 *
 * Returns:  void. 
 */
void
NullToLog2(struct nullmodel_s *null)
{
  int sym;   /* symbols for emission prob's      +*/

  /* transition prob's 
   */
  null->eta  = LOG2(null->eta);
  null->meta = LOG2(null->meta);

  /* emission prob's 
   */
  for (sym = 0; sym < 4; sym++) {
      null->xem[sym] = LOG2(null->xem[sym]);
      null->yem[sym] = LOG2(null->yem[sym]);
  }
}

void
NullLog2ToOdds(struct nullmodel_s *null1, struct nullmodel_s *null2)
{
  int sym;   /* symbols for emission prob's      +*/

  /* transition prob's 
   *       (don't touch null->eta)
   */
  null1->meta -= null2->meta;
  
  /* emission prob's 
   */
  for (sym = 0; sym < 4; sym++) {
    null1->xem[sym] -= null2->xem[sym];
    null1->yem[sym] -= null2->yem[sym];
  }
}

void    
PatternNullModel(struct nullmodel_s *null) 
{
  int sym;

  /* Initialize all prob's to zero
   */
  for (sym = 0; sym < 4; sym++) {
    null->xem[sym] = 0.0;
    null->yem[sym] = 0.0;
  }
}

/* Function: PrintNullModel()
 * Date:     ER, Fri Jun  4 13:15:31 CDT 1999 [St. Louis]
 *
 * Purpose:  Print a null model
 *
 * Args:     nullmodel -- the nullmodel prob's, in log2 form
 *
 * Returns:  void. prints transition and emission probs for null model, in [0,1] form.
 */
void
PrintNullModel(struct nullmodel_s *null)
{
  int sym;   /* symbols for emission prob's      +*/

  printf("\nNull MODEL -- Transition probabilities\n");
  printf("eta   = %f\n", EXP2(null->eta)); 
  printf("1-eta = %f\n", EXP2(null->meta)); 
  
  printf("\nNull MODEL -- Emission probabilities\n");
  for (sym = 0; sym < 4; sym++) 
    printf("P^X(%d) \t\t= %f\n", sym, EXP2(null->xem[sym]));
  for (sym = 0; sym < 4; sym++) 
    printf("P^Y(%d) \t\t= %f\n", sym, EXP2(null->yem[sym]));
}


/* Function: ScoreWithNull()
 * Date:     ER, Fri Jun 11 10:27:34 CDT 1999 [St. Louis]
 *
 * Purpose:  Score a gapped sequence alignment with null model.
 *           gap = (-) = 4 are ignored.
 *
 * Args:     seqX, seqY    -- equal length sequences, ACGT /only/
 *           start         -- starting position in s1,s2 
 *           L             -- lengths of s1,s2 from start
 *           nullmodel     -- [4][4] substitution matrix
 *
 * Returns:  log likelihood, log P(seq1,seq2 | null)
 */
double
ScoreWithNull(int *seqX, int *seqY, int startX, int Lx, int startY, int Ly, struct nullmodel_s *null)
{
  int    x, y;        /* position in seqX seqY */
  double sc;

  sc = 2. * null->eta;

  for (x = 0; x < Lx; x++) if (seqX[x+startX] < 4) sc += null->xem[seqX[x+startX]] + null->meta;
  for (y = 0; y < Ly; y++) if (seqY[y+startY] < 4) sc += null->yem[seqY[y+startY]] + null->meta;

  return sc;
}
double
ScoreWithNullDiag(int *seqX, int *seqY, int start, int L, struct nullmodel_s *null)
{
  int    i, iabs;        /* position in seq1,seq2 */
  double sc;

  sc = 2. * null->eta;

  for (i = 0; i < L; i++) {
    iabs = i + start;

    if (seqX[iabs] < 4) sc += null->meta + null->xem[seqX[iabs]];
    if (seqY[iabs] < 4) sc += null->yem[seqY[iabs]] + null->meta;
 }

  return sc;
}
double
ScoreWithNullX(int *seqX, int startX, int Lx, struct nullmodel_s *null)
{
  int    x;        /* position in seq */
  double sc;

  sc = null->eta;

  for (x = 0; x < Lx; x++) 
    if (seqX[x+startX] <  4) sc += null->xem[seqX[x+startX]] + null->meta;

  return sc;
}
double
ScoreWithNullY(int *seqY, int startY, int Ly, struct nullmodel_s *null)
{
  int    y;        /* position in seq */
  double sc;

  sc = null->eta;

  for (y = 0; y < Ly; y++) 
    if (seqY[y+startY] <  4) sc += null->yem[seqY[y+startY]] + null->meta;

  return sc;
}


/* Function: ScoreWithNullRev()
 * Date:     ER,Thu Jul 8 11:44:11 CDT 1999 [St. Louis]
 *
 * Purpose:  Score the rev of a gapped sequence alignment with null model.
 *           gap = (-) = 4 are ignored.
 *           trick: you do not need to reverse the sequence.
 *                      0(A) --> 3 - 0 = 3 (U)
 *                      1(C) --> 3 - 1 = 2 (G)
 *                      2(G) --> 3 - 2 = 1 (C)
 *                      3(U) --> 3 - 3 = 0 (A)
 *
 * Args:     seqX, seqY    -- equal length sequences, ACGT /only/
 *           L             -- lengths of s1,s2 
 *           nullmodel     -- [4][4] substitution matrix
 *
 * Returns:  log likelihood, log P(seq1,seq2 | null)
 */
double
ScoreWithNullRev(int *seqX, int *seqY, int startX, int Lx, int startY, int Ly, struct nullmodel_s *null)
{
  int    x, y;        /* position in seqX seqY */
  double sc;

  sc = 2. * null->eta;

  for (x = 0; x < Lx; x++) if (seqX[x+startX] < 4) sc += null->xem[3-seqX[x+startX]] + null->meta;
  for (y = 0; y < Ly; y++) if (seqY[y+startY] < 4) sc += null->yem[3-seqY[y+startY]] + null->meta;

  return sc;
}
double
ScoreWithNullRevDiag(int *seqX, int *seqY, int start, int L, struct nullmodel_s *null)
{
  int    x, xabs;        /* position in seq1,seq2 */
  double sc;

  sc = 2. * null->eta;

  for (x = 0; x < L; x++) {
    xabs = x + start;

    if (seqX[xabs] < 4) sc += null->xem[3-seqX[xabs]] + null->meta;
    if (seqY[xabs] < 4) sc += null->yem[3-seqY[xabs]] + null->meta;
  }

  return sc;
}


/* Function:  SimulateNullAlign()
 * Date:     ER, Tue Feb  1 11:22:15 CST 2000 [St. Louis]
 *
 * Purpose:  Simulate an aligment of two sequences
 *           related by the nullmodel.
 *
 * Args:     s1, s2    - the sequences 
 *           L1        - length of s1
 *           L2        - length of s2
 *           nullmodel - the null model
 *
 * Returns:  s2, an alloc'ed simulated second sequence
 */
void
SimulateNullAlign(int *s1, int *s2, int L1, int L2, struct nullmodel_s *null)
{
  int    pos;
  int    x, y;

  for (pos = 0; pos < L1; pos++) {
    x = DLog2Choose(null->xem, 4);

    s1[pos] = x;
  }

  for (pos = 0; pos < L2; pos++) {
    y = DLog2Choose(null->yem, 4);

    s2[pos] = y;
  }

}




