/************************************************************
 * QRNA - Comparative analysis of biological sequences 
 *        with pair hidden Markov models and pair stochastic context-free grammars
 * Copyright (C) 2000-Howard Hughes Medical Institute
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* othmodel.c
 *
 * E. Rivas 9 april 1999 [St. Louis]
 * 
 * Allocation, free'ing, initialization of the Othermodel
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

#include "funcs.h"
#include "globals.h"
#include "squid.h"
#include "structs.h"

#ifdef MEMDEBUG
#include "dbmalloc.h"
#endif

static double *fill_OTH_tranfer(int dim, double *oth_trans);
static void    from_tranfer_2_transitions(double *T, int dim, double *oth_trans);
static double *rate_matrix(int dim);
static double *transfer_at_zero(double *T, int dim);

/* Function: AllocOTHModel()
 * Date:     ER,  Fri May 28 13:01:57 CDT 1999 CDT 1999  [St. Louis]
 *
 * Purpose:  Allocates memory for the transition and emission probs of the othmodel
 *
 * Args:     othmodel - othmodel structure
 *
 * Returns:  void
 *           allocates oth->t[], oth->em[], which must be free'd by caller.
 */
struct othmodel_s *
AllocOTHModel(void)
{
  struct othmodel_s *oth;    /* othmodel structure         */

  oth      = (struct othmodel_s *) MallocOrDie (sizeof(struct othmodel_s));
  oth->t   = (double *) MallocOrDie (sizeof(double) * OTRANS);
  oth->mem = (double *) MallocOrDie (sizeof(double) * 16);
  oth->xem = (double *) MallocOrDie (sizeof(double) * 4);
  oth->yem = (double *) MallocOrDie (sizeof(double) * 4);

  oth->FLN = AllocNullModel();
  oth->FJN = AllocNullModel();
  oth->FRN = AllocNullModel();
  
  PatternOTHModel(oth);
  
  return oth;
}

/* Function: CheckDoubleProb()
 * Date:     ER, Tue May 25 13:18:02 CDT 1999  [St. Louis]
 *
 * Purpose:  Verify that \sum_{x,y} P(x,y) = 1
 *
 * Args:     pdouble - double-argument probability distribution to check
 *
 * Returns:  void. 
 */
void
CheckDoubleProb(double **pdouble, int dimx, int dimy)
{
  int    x;
  int    y;
  double sum = 0.0;

  for (x = 0; x < dimx; x++)
    for (y = 0; y < dimy; y++){
      if (pdouble[x][y] < -0.00005) Die (" probabilities are getting too small here");
      sum += pdouble[x][y];
    }
  
  if (sum > 1.01 || sum < 0.99) Die("sum_{x,y} P(x,y) is %f\n", sum);
}

/* Function: CheckOTHProbs()
 * Date:     ER, Fri May 28 13:17:34 CDT 1999  [St. Louis]
 *
 * Purpose:  Verify that transition and emission prob's of a othmodel add up to one
 *
 * Args:     oth - the structure for an othmodel
 *
 * Returns:  void. 
 */
void
CheckOTHProbs(struct othmodel_s *oth)
{
  int    idx;               /* index for transition prob's      +*/
  int    st;                /* state we are at                  +*/
  double sum;

  /* check transition prob's add up to one
   */
  for (st = 0; st < OSTATES; st++) {
    sum = 0.0; 
    for (idx = IdxTransOTH[st]; idx < IdxTransOTH[st]+TransPerOTHState[st]; idx++)
      sum += oth->t[idx];
    
    if (sum > 1.01 || sum < 0.99) 
      Die("transition prob's sum for state %d(%s) is %f \n", st, ostNAME[st], sum);
  }
  
  /* check emission prob's add up to one
   */
  CheckSingleProb(oth->mem, 16);
  CheckSingleProb(oth->xem, 4);
  CheckSingleProb(oth->yem, 4);
}

/* Function: CheckSingleProb()
 * Date:     ER, Tue May 25 13:18:02 CDT 1999  [St. Louis]
 *
 * Purpose:  Verify that \sum_x P(x) = 1
 *
 * Args:     psingle - single-argument probability distribution to check
 *
 * Returns:  void. 
 */
void
CheckSingleProb(double *psingle, int size)
{
  int x;
  double sum = 0.0;

  for (x = 0; x < size; x++) {
    if (psingle[x] < -0.00005) Die (" probabilities are getting too small here");
    sum += psingle[x];
  }
  
  if (sum > 2. - accuracy || sum < accuracy) Die("sum_x P(x) is %f\n", sum);
}
void
CheckSingleLog2Prob(double *psingle, int size)
{
  int    x;
  double prob;
  double sum = 0.0;

  for (x = 0; x < size; x++) {
    prob = EXP2(psingle[x]);
    if (prob < -0.00005) Die (" probabilities are getting too small here");
      sum += prob;
  }

  if (sum > 2. - accuracy || sum < accuracy) Die("sum_x P(x) is %f\n", sum);
}


/* Function: ConstructMemOTHProbs()
 * Date:     ER, Mon May 31 15:09:08 CDT 1999 [St. Louis]
 *
 * Purpose:  Given a PAMModel, marginalize to calculate
 *           a random model.
 *
 * Args:     pammodel  - 64x64 AAA..UUUxAAA..UUU joint prob matrix (prealloc)
 *           mem       - 4x4 A..UxA..U joint prob matrix (prealloc)           
 *
 * Returns:  (void)
 *           Fills in oth->mem. (already allocated)
 */
void
ConstructMemOTHProbs(double *mutpxy, double *mem)
{
  int x1;
  int x2;
  
  /* Zero null model
   */
  for (x1 = 0; x1 < 4; x1++)
    for (x2 = 0; x2 < 4; x2++) 
      mem[idx(x1,x2)] = mutpxy[idx(x1,x2)];

  CheckSingleProb(mem, 16);
}



/* Function: ConstructOTHModel()
 * Date:     ER, Fri May 28 12:07:32 CDT 1999  [St. Louis]
 *
 * Purpose:  Constructs a othmodel_s
 *
 * Args:     othparam - the list of parameters that define a othmodel           
 *
 * Returns:  (void)
 *           fills all prob's for othmodel, log2 form 
 *           (allc'ed here, freed by caller)
 */
void
ConstructOTHModel(double *mutpxy, struct othparam_s param, double tfactor, struct othmodel_s *oth, 
		  int pedantic)
{
  double            *T, *T_0, *R;     /* transfer matrices at two evolutionary times +*/
  double             tr;              /* transition prob's                           +*/
  int                idx;             /* index for transition prob's                 +*/

  /* Calculate transition prob's: oth.t[]
   */
  for (idx = 0; idx < OTRANS; idx++) {
   switch (idx) {
    case TFLB:  tr = param.xi;                                                                 break;
    case TFLFR: tr = 1.0 - param.xi;                                                           break;
    case TBM:   if ((tr = 1.0 - 2.0*param.kappa) < 0.) Die("TBM < 0");                         break;
    case TBX:   tr = param.kappa;                                                              break;
    case TBY:   tr = param.kappa;                                                              break;
    case TMM:   if ((tr = 1.0 - 2.0*param.delta - param.tau) < 0.) Die("TMM < 0");             break;
    case TMX:   tr = param.delta;                                                              break;
    case TMY:   tr = param.delta;                                                              break;
    case TME:   tr = param.tau;                                                                break;
    case TXM:   if ((tr = 1.0 - param.epsilon - param.tau - param.gamma) < 0.) Die("TXM < 0"); break;
    case TXX:   tr = param.epsilon;                                                            break;
    case TXY:   tr = param.gamma;                                                              break;
    case TXE:   tr = param.tau;                                                                break;
    case TYM:   if ((tr = 1.0 - param.epsilon - param.tau - param.gamma) < 0.) Die("TYM < 0"); break;
    case TYX:   tr = param.gamma;                                                              break;
    case TYY:   tr = param.epsilon;                                                            break;
    case TYE:   tr = param.tau;                                                                break;
    case TEFJ:  tr = 1.0 - param.eta;                                                          break;
    case TEFR:  tr = param.eta;                                                                break;
    case TFJB:  tr = 1.0;                                                                      break;
    case TFRFR: tr = 1.0;                                                                      break;
    default:   Die("transition (%d) does not exits", idx);
    }

   if (tr > 1.0 || tr < 0.0) Die ("transition %s in OTH model is wrong", otrNAME[idx]);
   else oth->t[idx] = tr;
  }

  /* SCALE the transition probabilities with the evolutionary factor.
   *       [This is basically a way of making gaps "evolve".]
   *
   * Define a Transfer matrix as:    TBM  TBX  TBY  TBE
   *                                
   *                                 TMM  TMX  TMY  TME
   *                            T =                           
   *                                 TXM  TXX  TXY  TXE
   *                                
   *                                 TYM  TYX  TYY  TYE
   *                                
   * T is a matrix of conditional probabilities (rows add up to one).
   *
   * T corresponds to a given evolutionary time (be that blosum62, or any other)                   
   *                                
   * You have to define T_0 the transfer matrix at "time zero".
   * We'll assume no gaps at time zero, and is not singular (ie invertible)
   *
   *     for instance:       1.0  0.0  0.0  0.0
   *
   *                         0.8  0.0  0.0  0.2
   *                 T_0 =   
   *                         0.0  1.0  0.0  0.0
   *
   *                         0.0  0.0  1.0  0.0
   *
   *
   * Using T and T_0 we can extrapolate and calculate T(r) at any other time  
   *                                
   *        Rescale : T(r) = T_0 * exp{r * log[T_0^{-1}*T] }                               
   *                                
   * Redefine : TBB = T(0,0), TBM = T(0,1), ...                               
   *                                
   *                                
   */
  if (tfactor < 1.0) {
    T   = fill_OTH_tranfer(OSTATES-1, oth->t);   /* input1:    T   (transfer matrix at a given time) */
    T_0 = transfer_at_zero(T, OSTATES-1);        /* input2:    T_0 (calculated from T)               */
    R   = rate_matrix(OSTATES-1);                /* input3:    R   (assumed to be Id for now)        */
    
    /* calculate: T(tfactor) 
     */
    T = ConditionalsEvolved_2(stdout, T, T_0, R, OSTATES-1, tfactor, FALSE, FALSE); 
    
    from_tranfer_2_transitions(T, OSTATES-1, oth->t);  /* recalculate transition probabilities from T */
    free(T);
    free(R);
    free(T_0);
  }
  
  /* Calculate emission prob's: OTH->mem[], OTH->xem[], OTH->yem[]
   */
  ConstructMemOTHProbs(mutpxy,   oth->mem);
  ConstructXemOTHProbs(oth->mem, oth->xem);
  ConstructYemOTHProbs(oth->mem, oth->yem);
 
  /* Construct flanking null models (already in log2 form)
   */
  ConstructNullModel(oth->mem, oth->FLN, param.FLN);
  ConstructNullModel(oth->mem, oth->FJN, param.FJN);
  ConstructNullModel(oth->mem, oth->FRN, param.FRN);

  /* check prob's add up to one
   */
  CheckOTHProbs(oth);
  
  /* convert to Log2 form
   */
  OTHToLog2(oth);
}


/* Function: ConstructXemOTHProbs()
 * Date:     ER, Mon May 31 15:14:03  CDT 1999 [St. Louis]
 *
 * Purpose:  Given a mem[idx(4,4)] mutation probability distrubution, 
 *           marginalize to calculate xem[4].
 *
 * Args:    mem - 4x4 A..UxA..U joint prob matrix (prealloc)
 *          xem - 4 seqX emission prob (prealloc)
 *
 * Returns:  (void)
 *           Fills in oth->xem (already allocated)
 */
void
ConstructXemOTHProbs(double *mem, double *xem)
{
  int x;
  int y;

  /* Zero null model
   */
  for (x = 0; x < 4; x++)
    xem[x] = 0.0;

  /* Marginalize and average over Y positions
   */
  for (x = 0; x < 4; x++)
    for (y = 0; y < 4; y++)
      xem[x] += mem[idx(x,y)];

  CheckSingleProb(xem, 4);
}

/* Function: ConstructYemOTHProbs()
 * Date:     ER, Mon May 31 15:14:03  CDT 1999 [St. Louis]
 *
 * Purpose:  Given a mem[idx(4,4)] mutation probability distrubution, 
 *           marginalize to calculate yem[4].
 *
 * Args:    mem - 4x4 A..UxA..U joint prob matrix (prealloc)
 *          yem - 4 seqY emission prob (prealloc)
 *
 * Returns:  (void)
 *           Fills in oth->yem (already allocated)
 */
void
ConstructYemOTHProbs(double *mem, double *yem)
{
  int x;
  int y;

  /* Zero null model
   */
  for (x = 0; x < 4; x++)
    yem[x] = 0.0;

  /* Marginalize and average over Y positions
   */
  for (x = 0; x < 4; x++)
    for (y = 0; y < 4; y++)
      yem[x] += mem[idx(y,x)];

  CheckSingleProb(yem, 4);
}


void
FreeOTHModel(struct othmodel_s *oth)
{
  FreeNullModel(oth->FLN);
  FreeNullModel(oth->FJN);
  FreeNullModel(oth->FRN);

  free(oth->t);
  free(oth->mem);
  free(oth->xem);
  free(oth->yem);

  free(oth);
}

/* Function: Joint_From_Condi()
 * Date:     ER, Mon Jun 12 14:41:47 CDT 2000 [St. Louis]
 *
 * Purpose:  Given       a LxL matrix Q(i,j) of conditional probabilities, 
 *           calculate the LxL matrix P(i,j) of joint       probabilities.
 *
 *           P(i,j) = Q(j|i) * pm[j] 
 *
 *           where P(i,j|t_o) are calculated according to 
 *           using a given pammodel t_o.
 *
 * Args:     Q  - LxL conditional prob matrix (prealloc)
 *           pm - L   single emision probabilities (prealloc)
 *
 * Returns:  P(LxL), conditional probabilities.
 *           P is alocated here, freed by caller.
 */
double *
Joint_From_Condi(FILE *ofp, double *Q, double *pm, int L, int verbose)
{
  double *P;
  int     i, j;
  
  if (verbose) {
    fprintf(ofp, "Q(i|j,t) probabilities\n");
    PrintProbs(ofp, Q, L);
  }
  
  /* allocate P[L*L]
   */
  P  = (double *) MallocOrDie (sizeof(double) * L * L);
 
  for (i = 0; i < L; i++) 
    for (j = 0; j < L; j++) 
      P[i*L+j] = Q[i*L+j] * pm[i]; 
  
  if (verbose) { 
    fprintf(ofp, "P(i,j,t) probabilities\n");
    PrintProbs(ofp, P, L);
  }
  
  /* consistency check
   */
  CheckSingleProb(P, L*L);
  
  return P;
}

 
/* Function: OTHToLog2()
 * Date:     ER, Fri May 28 15:03:30 CDT 1999 [St. Louis]
 *
 * Purpose:  Converts transition and emission prob's of a othmodel to log2 form
 *
 * Args:     oth - the structure for an othmodel
 *
 * Returns:  void. 
 */
void
OTHToLog2(struct othmodel_s *oth)
{
  int idx;               /* index for transition prob's      +*/
  int sym, symx, symy;   /* symbols for emission prob's      +*/

  /* transition prob's 
   */
  for (idx = 0; idx < OTRANS; idx++) 
    oth->t[idx] = LOG2(oth->t[idx]);
  
  /* emission prob's 
   */
  for (symx = 0; symx < 4; symx++)
    for (symy = 0; symy < 4; symy++)
      oth->mem[idx(symx,symy)] = LOG2(oth->mem[idx(symx,symy)]);

  for (sym = 0; sym < 4; sym++) {
      oth->xem[sym] = LOG2(oth->xem[sym]);
      oth->yem[sym] = LOG2(oth->yem[sym]);
  }
}

/* Function: OTHLog2ToOdds()
 * Date:     ER, Fri Jun  4 11:19:49 CDT 1999 [St. Louis]
 *
 * Purpose:  Converts transition and emission prob's of a othmodel
 *           from log2 to log2odds form
 *
 * Args:     oth - the structure for an othmodel
 *
 * Returns:  void. 
 */
void
OTHLog2ToOdds(struct othmodel_s *oth, struct nullmodel_s *null)
{
  int tr;                /* index for transitions            +*/
  int sym, symx, symy;   /* symbols for emission prob's      +*/

  NullLog2ToOdds(oth->FLN, null);
  NullLog2ToOdds(oth->FJN, null);
  NullLog2ToOdds(oth->FRN, null);

  /* transition prob's 
   */
  for (tr = 0; tr < OTRANS; tr++) 
    switch (tr) {
      /* transitions to M, divide by (1-\eta)^2  */
    case TBM: oth->t[tr] -= 2. * null->meta;  break;
    case TMM: oth->t[tr] -= 2. * null->meta;  break;
    case TXM: oth->t[tr] -= 2. * null->meta;  break;
    case TYM: oth->t[tr] -= 2. * null->meta;  break;
      /* transitions to X, divide by (1-\eta)    */
    case TBX: oth->t[tr] -= 1. * null->meta;  break;
    case TMX: oth->t[tr] -= 1. * null->meta;  break;
    case TXX: oth->t[tr] -= 1. * null->meta;  break;
    case TYX: oth->t[tr] -= 1. * null->meta;  break;
      /* transitions to Y, divide by (1-\eta)    */
    case TBY: oth->t[tr] -= 1. * null->meta;  break;
    case TMY: oth->t[tr] -= 1. * null->meta;  break;
    case TXY: oth->t[tr] -= 1. * null->meta;  break;
    case TYY: oth->t[tr] -= 1. * null->meta;  break;
    }
  
  /* emission prob's 
   */
  for (symx = 0; symx < 4; symx++)
    for (symy = 0; symy < 4; symy++)
      oth->mem[idx(symx,symy)] -= (null->xem[symx] + null->yem[symy]);
  
  for (sym = 0; sym < 4; sym++) {
    oth->xem[sym] -= null->xem[sym];
    oth->yem[sym] -= null->yem[sym];
  }
}

void     
PatternOTHModel(struct othmodel_s *oth)
{
  int idx;
  int sym;

  PatternNullModel(oth->FLN);
  PatternNullModel(oth->FJN);
  PatternNullModel(oth->FRN);

  /* Initialize all prob's to zero
   */
  for (idx = 0; idx < OTRANS; idx++)
    oth->t[idx] = 0.0;

  for (sym = 0; sym < 16; sym++)
    oth->mem[sym] = 0.0;
  
  for (sym = 0; sym < 4; sym++) {
    oth->xem[sym] = 0.0;
    oth->yem[sym] = 0.0;
  }

}

/* Function: PrintMutProbs()
 * Date:     ER, Mon Aug  2 20:48:07 CDT 1999 [St. Louis]
 *
 * Purpose:  Print a mut probs model
 *
 * Args:     mutpxy[4][4]
 *           pnull[4]  (in log2 form)
 *
 * Returns:  void. prints stuff.
 */
void
PrintMutProbs(double *mutpxy, double *pnull)
{
  int      x;
  double   nullp;
  double   info = 0.0;
  double   expect = 0.0;

  printf("Mutation probabilities\n");
  for (x = 0; x < 16; x++) {
    nullp  = EXP2(pnull[x/4]) * EXP2(pnull[x%4]);
    
    if (x%4 == 0) printf("\n");
    printf("%12f ",  mutpxy[x]);

    if (mutpxy[x] > 0.) {
      info   += mutpxy[x]   * (LOG2(mutpxy[x]) - nullp);
      expect += EXP2(nullp) * (LOG2(mutpxy[x]) - nullp);
    }
    else if (mutpxy[x] > 0.) Die ("mutation probs have to be larger than zero, dear");

  }
  printf("\ninfo content per mutation: %f bits\n", info);
  printf("expectation per mutation:  %f bits\n", expect);
}

void
PrintMut5Probs(double *mut5pxy, double *pnull)
{
  int      x;
  double   nullp;
  double   info = 0.0;
  double   expect = 0.0;
  
  printf("Mutation probabilities\n");
  for (x = 0; x < 25; x++) {
    nullp  = (x/5<4)? EXP2(pnull[x/4]) : INDL;
    nullp *= (x%5<4)? EXP2(pnull[x%4]) : INDL;
    
    if (x%5 == 0) printf("\n");
    printf("%12f ",  mut5pxy[x]);
    
    if (mut5pxy[x] > 0.) {
      info   += mut5pxy[x]  * (LOG2(mut5pxy[x]) - nullp);
      expect += EXP2(nullp) * (LOG2(mut5pxy[x]) - nullp);
    }
    else if (mut5pxy[x] > 0.) Die ("mutation probs have to be larger than zero, dear");

  }
  printf("\ninfo content per mutation: %f bits\n", info);
  printf("expectation per mutation:  %f bits\n", expect);
}

/* Function: PrintOTHModel()
 * Date:     ER, Thu May 27 13:41:09 CDT 1999 [St. Louis]
 *
 * Purpose:  Print a oth model
 *
 * Args:     othmodel -- the othmodel prob's, in log2 form
 *
 * Returns:  void. prints transition and emission probs for oth model, in [0,1] form.
 */
void
PrintOTHModel(struct othmodel_s *oth)
{
  int sym, symx, symy;   /* symbols for emission prob's      +*/
  int idx;               /* index for transition prob's      +*/

  printf("\nOTH MODEL -- Flanking Left Model\n");
  PrintNullModel(oth->FLN);

  printf("\nOTH MODEL -- Flanking Middle Model\n");
  PrintNullModel(oth->FJN);

  printf("\nOTH MODEL -- Flanking Right Model\n");
  PrintNullModel(oth->FRN);

  printf("\nOTH MODEL -- Transition probabilities\n");
  for (idx = 0; idx < OTRANS; idx++)
    printf("t[%s]\t\t = %f\n", otrNAME[idx], EXP2(oth->t[idx])); 

  printf("\nOTH MODEL -- Emission probabilities\n");
  for (symx = 0; symx < 4; symx++)
    for (symy = 0; symy < 4; symy++)
      printf("P^M(%d,%d) \t= %f\n", symx, symy, EXP2(oth->mem[idx(symx,symy)]));
  for (sym = 0; sym < 4; sym++) 
    printf("P^X(%d) \t\t= %f\n", sym, EXP2(oth->xem[sym]));
  for (sym = 0; sym < 4; sym++) 
    printf("P^Y(%d) \t\t= %f\n", sym, EXP2(oth->yem[sym]));
}


/* Function: fill_OTH_tranfer()
 * Date:     ER, Thu Apr 20 10:51:47 CDT 2000 [St. Louis]
 *
 * Purpose:  Fill T the transfer matrix for the OTH model transition probabilities.
 *           T is a matrix of conditional probabilities.
 *
 * Args:     T         -- transfer matrix for the OTH model to be filled.
 *           dim       -- dimension. T(dim x dim)
 *           oth_trans -- transition probabilities of the OTH model
 *
 * Returns:  void. T is allocated and filled here. T freed by caller.
 */
double *
fill_OTH_tranfer(int dim, double *oth_trans)
{
  int     st1, st2;
  int     tr;       /* counter for non-null transitions of st1 state */
  int     row, col;
  int     idx, ntrans;
  double *T;
  double  sum;

  /* allocate T
   */
  T  = (double *) MallocOrDie (sizeof(double) * dim * dim);

  for (row = 0; row < dim; row++) 
    for (col = 0; col < dim; col++) 
      T[row*dim+col] = 0.;

  /* fill T.
   *           to:     B      M      X      Y     E    FJ    FR
   *          from
   *
   *           FL     TFLB    x      x      x     x     x    x
   *        
   *           B        x    TBM    TBX    TBY   TBE    x    x
   *                                
   *           M        x    TMM    TMX    TMY   TME    x    x
   * T =                           
   *           X        x    TXM    TXX    TXY   TXE    x    x
   *                                
   *           Y        x    TYM    TYX    TYY   TYE    x    x
   *              
   *           E        x     x      x      x     x   TEFJ TEFR
   * 
   *           FJ       x   TFJB     x      x     x     x    x
   * 
   * 
   */
  
  for (st1 = stFL, row = 0; st1 < stFR; st1++, row++) {
    idx    = IdxTransOTH[st1];      /* starting index for transitions for state st1 */
    ntrans = TransPerOTHState[st1]; /* number of non-null transitions for state st1 */

    for (st2 = stB, col = 0, tr = idx; st2 <= stFR; st2++, col++) 
      {
	if (col > dim || row > dim) 
	  Die("OTH transfer matrix badly dimensioned, state %s --> %s", ostNAME[st1], ostNAME[st2]);
	
	if (tr - idx > ntrans) 
	  Die("OTH transition off limits, tr %s", otrNAME[tr]);
	
	if (OConnects[st1][st2]) {
		T[row*dim+col] = oth_trans[tr];
		tr ++;
	}
      }
  }
  
  /* paranoia
   */
  for (row = 0; row  < dim; row ++) {
    sum = 0.;
    for (col = 0; col < dim; col++) {
      if (T[row*dim+col] < 0.) Die ("OTH-transfer-matrix is getting too small");
      sum += T[row*dim+col];
    }
    if (fabs(sum) < accuracy || fabs(sum) > 2. - accuracy)
      Die ("OTH-transfer-matrix row (%d) does not add up to one (sum =%f)", row, sum);
  }
 
  if(FALSE) {
    printf("T* - OTH transfer matrix\n");
    for (row = 0; row < dim; row++) {
      for (col = 0; col < dim; col++) {
	printf("%.4f ", T[row*dim+col]);
      }
      printf("\n");
    }
  }

  return T;
}
  
/* Function: from_tranfer_2_transitions()
 * Date:     ER, Thu Apr 20 11:01:39 CDT 2000 [St. Louis]
 *
 * Purpose:  Reverse of fill_OTH_tranfer(). 
 *           Given the transfer matrix for OTH, recalculate the transition probabilities.
 *
 * Args:     T         -- transfer matrix for the OTH model 
 *           dim       -- dimension. T(dim x dim)
 *           oth_trans -- transition probabilities of the OTH model
 *
 * Returns:  void. oth_trans are reasigned.
 */
void
from_tranfer_2_transitions(double *T, int dim, double *oth_trans)
{

  int    st1, st2, tr;
  int    row, col;
  int    idx, ntrans;

  if(FALSE) {
    printf("T(r) - OTH transfer matrix\n");
    for (row = 0; row < dim; row++) {
      for (col = 0; col < dim; col++) {
	printf("%.4f ", T[row*dim+col]);
      }
      printf("\n");
    }
  }
  
  for (st1 = stFL, row = 0; st1 < stFR; st1++, row++) {
    idx    = IdxTransOTH[st1];      /* starting index for transitions for state st1 */
    ntrans = TransPerOTHState[st1]; /* number of non-null transitions for state st1 */
    
    /*initialize counter for non-null transitions of st1 state 
     */
    tr = idx;
    
    for (st2 = stB, col = 0; st2 <= stFR; st2++, col++) 
      {
	if (col > dim || row > dim) 
	  Die("OTH transfer matrix badly dimensioned, state %s --> %s", ostNAME[st1], ostNAME[st2]);
	
	if (tr - idx > ntrans) 
	  Die("OTH transition off limits, tr %s", otrNAME[tr]);
	
	if (OConnects[st1][st2]) {
	  oth_trans[tr] = T[row*dim+col];
	  tr ++;
	}
      }
  }
}

/* Function: tranfer_at_zero()
 * Date:     ER, Thu Apr 20 10:26:24 CDT 2000 [St. Louis]
 *
 * Purpose:  Fill T_0 the transfer matrix for the OTH model at zero evolutionary distance
 *
 * Args:     T_0 -- transfer matrix for the OTH model (no gaps)
 *
 * Returns:  void. T_0 is allocated and filled here. T_0 freed by caller.
 */
double *
transfer_at_zero(double *T, int dim)
{
  double *T_0;
  int     row, col;
  int     st1, st2;
  int     pos;

  /* allocate T_0
   */
  T_0  = (double *) MallocOrDie (sizeof(double) * dim * dim);

  /* fill T_0.
   * this is done totally ad hoc, just assuming no gaps at zero evolutionary distance,
   * and that the matrix is not singular (ie invertible)
   */
  for (row = 0; row < dim; row++) 
    for (col = 0; col < dim; col++) 
      T_0[row*dim+col] = T[row*dim+col];

  for (st1 = stFL, row = 0; st1 < stFR; st1++, row++)       
    for (st2 = stB, col = 0; st2 <= stFR; st2++, col++) 
      {
	if (col > dim || row > dim) 
	  Die("OTH transfer matrix badly dimensioned");
	pos = row*dim + col;

  	     if(st1 == stB  && st2 == stM)  T_0[pos] = 1.0;
	else if(st1 == stB  && st2 == stX)  T_0[pos] = 0.0;
  	else if(st1 == stB  && st2 == stY)  T_0[pos] = 0.0;
  	else if(st1 == stB  && st2 == stE)  T_0[pos] = 0.0;

  	else if(st1 == stM  && st2 == stM)  T_0[pos] = 1.0 - T[row*dim+stE-1];
  	else if(st1 == stM  && st2 == stX)  T_0[pos] = 0.0;
  	else if(st1 == stM  && st2 == stY)  T_0[pos] = 0.0;
  	else if(st1 == stM  && st2 == stE)  T_0[pos] = T[row*dim+stE-1];

	     /*else if(st1 == stX  && st2 == stM)  T_0[pos] = 0.0;
  	else if(st1 == stX  && st2 == stX)  T_0[pos] = 0.0;
  	else if(st1 == stX  && st2 == stY)  T_0[pos] = 1.0 - T[row*dim+stE-1];
  	else if(st1 == stX  && st2 == stE)  T_0[pos] = T[row*dim+stE-1];

  	else if(st1 == stY  && st2 == stM)  T_0[pos] = 0.0;
  	else if(st1 == stY  && st2 == stX)  T_0[pos] = 1.0 - T[row*dim+stE-1];
  	else if(st1 == stY  && st2 == stY)  T_0[pos] = 0.0;
  	else if(st1 == stY  && st2 == stE)  T_0[pos] = T[row*dim+stE-1];*/
    }
  
  if(FALSE) {
    printf("T_0 - OTH transfer matrix\n");
    for (row = 0; row < dim; row++) {
      for (col = 0; col < dim; col++) {
	printf("%.4f ", T_0[row*dim+col]);
      }
      printf("\n");
    }
  }
  
  return T_0;
}

/* Function: rate_matrix()
 * Date:     ER, Mon May  8 11:20:52 CDT 2000 [St. Louis]
 *
 * Purpose:  Fill R the extra matrix
 *
 * Args:     R -- transfer matrix for the OTH model (no gaps)
 *
 * Returns:  void. R is allocated and filled here. R freed by caller.
 */
double *
rate_matrix(int dim)
{
  double *R;
  int     row, col;

  /* allocate R
   */
  R  = (double *) MallocOrDie (sizeof(double) * dim * dim);

  /* fill R.
   * this is done totally ad hoc.
   */
  for (row = 0; row < dim; row++) 
    for (col = 0; col < dim; col++) 
      if (row == 0)  R[row*dim+col] = (row == col+1-dim)? 1.0 : 0.0;
       else          R[row*dim+col] = (row == col+1)? 1.0 : 0.0;

  if (FALSE) {
    printf("R - OTH rate matrix\n");
    for (row = 0; row < dim; row++) {
      for (col = 0; col < dim; col++) {
	printf("%.4f ", R[row*dim+col]);
    }
      printf("\n");
    }
  }
  
  return R;
}





