/************************************************************
 * QRNA - Comparative analysis of biological sequences 
 *        with pair hidden Markov models and pair stochastic context-free grammars
 * Copyright (C) 2000-Howard Hughes Medical Institute
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* qrna.c
 *
 * E. Rivas [St. Louis]
 * 
 * 9 april 1999.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <time.h>

#include "funcs.h"
#include "globals.h"
#include "squid.h"
#include "structs.h"
#include "version.h"


static void analyze_2_sequences (FILE *ofp, int format, int num_align,
				 SQINFO sqinfoX, char *seqX, int *isegX, int *iseqX, int *ctX,
				 SQINFO sqinfoY, char *seqY, int *isegY, int *iseqY, int *ctY,
				 long int Lmax, int Lw, long int maxlenhit, int minlenhit,
				 int window, int slide,
				 struct pamcond_s  *pamcond,
				 double ***cfg_node,
				 double **hexa,
				 double *pair5prob,
				 struct dos_s d, 
				 struct model_s *model, 
				 struct dpd_s *dpd, 
				 struct dpf_s *dpf, 
				 struct rnascfg_s *mx, 
				 struct scores_s *sc, 
				 struct ali_s *ali,
				 int add_hexamer, int allow_pseudoknots, int alignment, int cyk, int changefreq, int changefreqwin, 
				 int logodds, int pedantic, int shuffle, int strand, int sweep, int tfactor, 
				 int traceback, int verbose, int rnass, int shtoo);
     
     
static char banner[] = "qrna -- scores an alignment with the 3 models";
                
static char usage[]  = "\
Usage: qrna [-options] <input_file.q> \n\
where options are:\n\
   -A             : do an all-to-all comparison between the two input files\n\
   -a             : print alignment \n\
   -b             : shuffle the alignment \n\
   -c <cfgfile>   : <cfgfile> to use to train the rna model (default = tRNA+rRNA)\n\
   -d             : log2 form (default = log2-odds space )\n\
   -e <num>       : number of sequen skipped in second file (for multiple comparisons). default 0. \n\
   -F             : change the overall base composition of the 3 models, based on nts frequencies in the whole alignment\n\
   -f             : use full dp for the probabilistic models (do not conserve the aligment-default is diagonal dp)\n\
   -G             : change the overall base composition of the 3 models, based on nts frequencies for each window\n\
   -g             : do forward (default is viterbi)\n\
   -H             : include a file of Hexamer frequencies for the coding model\n\
   -h             : print short help and usage info\n\
   -i             : evolutionary time factor (default i=1)\n\
   -j             : use semi-full dp for the probabilistic model (use the alignment created by OTH)\n\
   -k             : allow pseudoknots (not implemented)\n\
   -l <minlenhit> : change the minlenhit parameter (default 0)\n\
   -L <maxlenhit> : change the maxlenhit parameter (default provided by longuest sequence)\n\
   -m             : do Forward and Viterbi Diagonal dp\n\
   -n             : do Forward and Viterbi Full     dp\n\
   -o <outfile>   : direct structure-annotated sequence to <outfile>\n\
   -p <pamfile>   : <pamfile> to use (default = BLOSUM62)\n\
   -P             : pedantic, check your evolutionary models for inconsistencies\n\
   -q             : do Forward and Viterbi semi-full dp\n\
   -r             : do Nussinov rna model (default is a 3-state model) \n\
   -s             : do global (not dp)\n\
   -S             : sweep a collection of motifs(seqfile1) across another bunch of sequences(seqfile2)\n\
   -t             : print traceback\n\
   -v             : verbose debugging output\n\
   -w <num>       : scanning window (default is full length)\n\
   -x <num>       : slide positions (default is 50)\n\
   -y             : grab sequences at random from the second data file\n\
";

static char experts[] = "\
   --cyk   : use CYK algorithm to calculate RNA score (default is Inside).\n\
   --latte : I just called starbucks with your order...\n\
   --rnass : print the alignment with the predicted RNA secondary structure.\n\
   --shtoo : qrna the alignment and also give one shuffled score.\n\
\n";                                                                                                                        

static struct opt_s OPTIONS[] = {
  { "-A",        TRUE,  sqdARG_NONE},
  { "-a",        TRUE,  sqdARG_NONE},
  { "-b",        TRUE,  sqdARG_NONE},
  { "-c",        TRUE,  sqdARG_STRING},
  { "-d",        TRUE,  sqdARG_NONE},
  { "-e",        TRUE,  sqdARG_STRING},
  { "-F",        TRUE,  sqdARG_NONE},
  { "-f",        TRUE,  sqdARG_NONE},
  { "-G",        TRUE,  sqdARG_NONE},
  { "-g",        TRUE,  sqdARG_NONE},
  { "-h",        TRUE,  sqdARG_NONE},
  { "-H",        TRUE,  sqdARG_STRING},
  { "-i",        TRUE,  sqdARG_STRING},
  { "-j",        TRUE,  sqdARG_NONE},
  { "-k",        TRUE,  sqdARG_NONE},
  { "-l",        TRUE,  sqdARG_STRING},
  { "-L",        TRUE,  sqdARG_STRING},
  { "-m",        TRUE,  sqdARG_NONE},
  { "-n",        TRUE,  sqdARG_NONE},
  { "-o",        TRUE,  sqdARG_STRING},
  { "-p",        TRUE,  sqdARG_STRING},
  { "-P",        TRUE,  sqdARG_NONE},
  { "-q",        TRUE,  sqdARG_NONE},
  { "-r",        TRUE,  sqdARG_NONE},
  { "-S",        TRUE,  sqdARG_NONE},
  { "-s",        TRUE,  sqdARG_NONE},
  { "-t",        TRUE,  sqdARG_NONE},
  { "-v",        TRUE,  sqdARG_NONE},
  { "-w",        TRUE,  sqdARG_STRING},
  { "-x",        TRUE,  sqdARG_STRING},
  { "-y",        TRUE,  sqdARG_NONE},
  { "--cyk",     FALSE, sqdARG_NONE},
  { "--latte",   FALSE, sqdARG_NONE},
  { "--rnass",   FALSE, sqdARG_NONE},
  { "--shtoo",   FALSE, sqdARG_NONE},
};

#define NOPTIONS (sizeof(OPTIONS) / sizeof(struct opt_s))

int
main(int argc, char **argv)
{
  char    *seqfile;               /* input sequence file                                    */
  char    *seqfile_b;             /* input sequence file                                    */
  SQFILE  *sqfp;	          /* open sequence file                                     */
  SQFILE  *sqfp_b;	          /* open sequence file                                     */
  char    *seqX, *revX;	          /* sequence X                                             */
  char    *seqY;	          /* sequence to compare                                    */
  int     *iseqX, *isegX;	  /* sequence X integer form (without and with gaps)        */
  int     *iseqY, *isegY;	  /* sequence Y integer form (without and with gaps)        */
  SQINFO   sqinfoX;               /* info structures for seqX                               */
  SQINFO   sqinfoY;               /* info structures for seqY                               */
  int     *ctX;                   /* .ct notation for seqX's RNA structure (if any)         */
  int     *ctY;                   /* .ct notation for seqY's RNA structure (if any)         */
  long int curr_Lmax, Lmax;       /* length of the max alignment                            */
  int      format;                /* format of seq file                                     */

  char    *hexapfile;             /* Hexamer joint frequencies                              */
  char    *pamfile;               /* PAM substitution matrix                                */
  char    *cfgfile;               /* RNA grammar file                                       */

  double ***cfg_node;
  double  **hexa;
  double   *pair5prob;

  struct   pamcond_s   *pamcond;
  struct   model_s     *model;    /* transition + emission probs 3 models + null model                         */
  struct   rnascfg_s   *mx;       /* matrices for SCFG part of RNA dp calculation                              */
  struct   dpd_s       *dpd;      /* structure with dp matrices -- three models                                */
  struct   dpf_s       *dpf;      /* structure with dp matrices -- three models                                */
  struct   ali_s       *ali;      /* arrays to store the alignment created by full viterbi                     */ 
  struct   scores_s    *sc;       /* structure containing all possible scores                                  */
  struct   dos_s        d;        /* collection of flags to decide the scoring method(s)                       */

  double   tfactor;               /* evolutionary time factor                                                  */
  int      add_hexamer;	          /* TRUE adds hexamer frequencies for coding model                            */
  int      alignment;	          /* TRUE prints alignment                                                     */
  int      allow_pseudoknots;	  /* TRUE to logodds space                                                     */
  int      alltoall;	          /* TRUE all-to-all comparison between the two input files                    */
  int      changefreq;            /* TRUE change the baseline base composition of 3 models for whole alignment */
  int      changefreqwin;         /* TRUE change the baseline base composition of 3 models for each window     */
  int      cyk;                   /* TRUE use CYK algorithm to calculate RNA scores (default is Inside)        */
  int      logodds;               /* TRUE to be extremely verbose to debug                                     */
  int      pedantic;              /* TRUE do some checks for evolutionary models to debug                      */
  int      ran;                   /* TRUE to pick a random seq from the second data file                       */
  int      shuffle;               /* TRUE to shuffle alignment                                                 */
  int      sweep;                 /* TRUE sweep one motif along a sequence                                     */
  int      traceback;             /* TRUE to traceback alignment                                               */
  int      verbose;               /* TRUE to be extremely verbose to debug                                     */

  int      latte   = FALSE;       /* friday lunch 31 AUG 2001                                                       */  
  int      rnass   = FALSE;       /* TRUE prints the alignment with the predicted RNA secondary structure           */ 
  int      shtoo   = FALSE;       /* TRUE also gives one shuflled score                                             */ 


  char    *outfile;               /* where to send the output                               */
  FILE    *ofp;	                  /* open output file                                       */
  char    *optname;
  char    *optarg; 
  int      optind;	
  int      s, seed;
  int      num_seqs = 0;
  int      num = 0, skip;

  int      minlenhit = 0;         /* min length hit by default 0                                */
  long int maxlenhit = 0;         /* max length hit by default provided by the longer aligments */
  int      nseqs;                 /* total number of sequences                                  */
  int      nXseq = 0, nYseq = 0;
  int      num_ali = 0;
  int      slide, window, winop;
  int      Lw;

 
  /* re-seed the random number generator.
   */
  seed = (int) time ((time_t *) NULL);
  sre_srandom(seed); /* reinit sre_random each time you shuffle a sequence */

  /* Parse command line
   */
  add_hexamer       = FALSE;  /* TRUE  ==  adds hexamer frequencies for coding                           */
  alignment         = FALSE;  /* TRUE  ==  prints alignment                                              */
  allow_pseudoknots = FALSE;  /* TRUE  ==  allow pseudoknots                                             */
  alltoall          = FALSE;  /* TRUE  ==  all-to-all comparison                                         */
  changefreq        = FALSE;  /* TRUE  ==  change the base comp of 3 models for whole alignment          */
  changefreqwin     = FALSE;  /* TRUE  ==  change the base comp of 3 models for each window              */
  cyk               = FALSE;  /* TRUE  ==  use CYK algorithm to calculate RNA scores (default is Inside) */
  logodds           = TRUE;   /* FALSE ==  no logoods                                                    */
  pedantic          = FALSE;  /* TRUE  ==  check your evolutionary models                                */
  ran               = FALSE;  /* TRUE  ==  pick a random seq from the second file                        */
  shuffle           = FALSE;  /* TRUE  ==  shuffle alignment                                             */
  sweep             = FALSE;  /* TRUE  ==  sweep one motif along a sequence                              */
  traceback         = FALSE;  /* TRUE  ==  traceback alignment                                           */
  verbose           = FALSE;  /* TRUE  ==  for debuging                                                  */
  winop             = FALSE;  /* TRUE  ==  use a window; FALSE == use whole alignment                    */

  d.frdiag         = FALSE;   /* TRUE  ==  calculates forward diagonal                                   */
  d.frfull         = FALSE;   /* TRUE  ==  calculates forward full                                       */
  d.frsemi         = FALSE;   /* TRUE  ==  calculates forward semi-full                                  */
  d.fulldp         = FALSE;   /* TRUE  ==  calculates full dp scores                                     */
  d.global         = FALSE;   /* TRUE  ==  calculates global alignments                                  */
  d.nus            = FALSE;   /* TRUE  ==  uses NUS model for rna                                        */
  d.semidp         = FALSE;   /* TRUE  ==  calculates semi-full dp scores                                */
  d.frdiag         = FALSE;   /* TRUE  ==  calculates forward diagonal                                   */
  d.vifull         = FALSE;   /* TRUE  ==  calculates viterbi full                                       */
  d.visemi         = FALSE;   /* TRUE  ==  calculates viterbi semi-full                                  */
  d.vidiag         = TRUE;    /* FALSE ==  calculates viterbi dp scores                                  */
  d.forward        = FALSE;   /* TRUE  ==  calculates forward dp scores                                  */
  d.twodiag        = FALSE;   /* TRUE  ==  both Forward and Viterbi diagonal dp                          */
  d.twofull        = FALSE;   /* TRUE  ==  both Forward and Viterbi full dp                              */
  d.twosemi        = FALSE;   /* TRUE  ==  both Forward and Viterbi semi-full dp                         */
  
  tfactor   = 1.0;
  cfgfile   = NULL;
  hexapfile = NULL;
  pamfile   = "BLOSUM62";
  outfile   = NULL;
  seqfile   = NULL;
  seqfile_b = NULL;
  slide     = 50;
  skip      = 0;
  
  while (Getopt(argc, argv, OPTIONS, NOPTIONS, usage,
		&optind, &optname, &optarg))
    {
      if      (strcmp(optname, "-A") == 0)   alltoall          = TRUE;
      else if (strcmp(optname, "-a") == 0)   alignment         = TRUE;
      else if (strcmp(optname, "-b") == 0)   shuffle           = TRUE;
      else if (strcmp(optname, "-c") == 0)   cfgfile           = optarg;
      else if (strcmp(optname, "-d") == 0)   logodds           = FALSE;
      else if (strcmp(optname, "-e") == 0)   skip              = atoi(optarg);
      else if (strcmp(optname, "-F") == 0)   changefreq        = TRUE;
      else if (strcmp(optname, "-f") == 0)   d.fulldp          = TRUE;
      else if (strcmp(optname, "-G") == 0)   changefreqwin     = TRUE;
      else if (strcmp(optname, "-g") == 0)   d.forward         = TRUE;
      else if (strcmp(optname, "-H") == 0)   hexapfile         = optarg;
      else if (strcmp(optname, "-i") == 0)   tfactor           = atof(optarg);
      else if (strcmp(optname, "-j") == 0)   d.semidp          = TRUE;
      else if (strcmp(optname, "-k") == 0)   allow_pseudoknots = TRUE;
      else if (strcmp(optname, "-l") == 0)   minlenhit         = atoi(optarg);
      else if (strcmp(optname, "-L") == 0)   maxlenhit         = atoi(optarg);
      else if (strcmp(optname, "-m") == 0)   d.twodiag         = TRUE;
      else if (strcmp(optname, "-n") == 0)   d.twofull         = TRUE;
      else if (strcmp(optname, "-o") == 0)   outfile           = optarg;
      else if (strcmp(optname, "-p") == 0)   pamfile           = optarg;
      else if (strcmp(optname, "-P") == 0)   pedantic          = TRUE;
      else if (strcmp(optname, "-q") == 0)   d.twosemi         = TRUE;
      else if (strcmp(optname, "-r") == 0)   d.nus             = TRUE;
      else if (strcmp(optname, "-S") == 0)   sweep             = TRUE;
      else if (strcmp(optname, "-s") == 0)   d.global          = TRUE;
      else if (strcmp(optname, "-t") == 0)   traceback         = TRUE;
      else if (strcmp(optname, "-v") == 0)   verbose           = TRUE;
      else if (strcmp(optname, "-w") == 0) { window            = atoi(optarg); winop = TRUE; }
      else if (strcmp(optname, "-x") == 0)   slide             = atoi(optarg);
      else if (strcmp(optname, "-y") == 0)   ran               = TRUE;
      else if (strcmp(optname, "--cyk")    == 0)   cyk         = TRUE;
      else if (strcmp(optname, "--latte")  == 0)   latte       = TRUE;
      else if (strcmp(optname, "--rnass")  == 0)   rnass       = TRUE;
      else if (strcmp(optname, "--shtoo")  == 0)   shtoo       = TRUE;
      else if (strcmp(optname, "-h")       == 0) 
	{
	  puts(banner);
	  printf("          %s (%s)", RELEASE, RELEASEDATE);
	  printf(" using squid %s (%s)\n", squid_version, squid_date);
	  puts(usage);
	  puts(experts);                                                                                                      
	  exit(0);
	}
    }
  
  if (argc - optind == 1) {
    seqfile  = argv[optind]; 
  }
  else if (argc - optind == 2) {
    seqfile   = argv[optind++]; 
    seqfile_b = argv[optind]; 
  }
  else {
    puts(banner);
    printf("version: %s (%s)\n", RELEASE, RELEASEDATE);
    Die("Incorrect number of command line arguments.\n%s \n\n%s\n", usage, experts);
  }
  
  if (winop && window <= 0) Die("window has to be larger than zero\n");
  if (!winop) window = -1;

  if (cfgfile == NULL) /* SCFG is not provided */
    cfgfile = FileConcat("", "mix_tied_linux.cfg");
  
  if (seqfile == NULL) {
    puts(usage);
    puts(experts);                      
    exit(0);
  }
  
  if (hexapfile) add_hexamer = TRUE;
  else           add_hexamer = FALSE;

  /* Determine which DP we are asked to do
   *
   * (very very not elegant, but oh well, bite me!)
   */
  if (d.global)  { d.vidiag = FALSE; }
  if (d.forward && !d.fulldp && !d.semidp) { d.vidiag = FALSE; d.vifull = FALSE; d.frdiag = TRUE; }

  if (d.twodiag) { d.vidiag = TRUE; d.frdiag = TRUE; }

  if ((d.fulldp || d.semidp || d.twofull || d.twosemi) && !d.twodiag) d.vidiag = FALSE;

  if ((d.fulldp &&  d.forward) || d.twofull) d.frfull = TRUE;
  if ((d.fulldp && !d.forward) || d.twofull) d.vifull = TRUE;
  if ((d.semidp &&  d.forward) || d.twosemi) d.frsemi = TRUE;
  if ((d.semidp && !d.forward) || d.twosemi) d.visemi = TRUE;
  
  /* Open output file 
   */
  ofp = stdout;
  if (outfile != NULL && (ofp = fopen(outfile, "w")) == NULL)
    Die("Failed to open output file %s", outfile);
  
  /* Open sequence file(s)
   */
  if (! SeqfileFormat(seqfile, &format, NULL))
    Die("Failed to determine format of sequence file %s\n", seqfile);

        
  
  if (maxlenhit > 0) 
    {
      nseqs = 0;
      Lmax = maxlenhit;
    }
  else 
    {
      /* determine maximum length for sequence file (Lmax)
       * (this is too expensive to do)
       */
      if (winop) 
	maxlenhit = BIGINT;
      else /* if no window given, limit the max length of aligments for memoery reasons */
	maxlenhit = 1000;
      
      if ((sqfp = SeqfileOpen(seqfile, format, NULL)) == NULL)
	Die("Failed to open sequence file %s", seqfile);
      CheckMaxLength(sqfp, format, &curr_Lmax, &nseqs, maxlenhit); 
      Lmax = curr_Lmax; 
      SeqfileClose(sqfp);
    }

  if ((sqfp = SeqfileOpen(seqfile, format, NULL)) == NULL)
    Die("Failed to open sequence file %s", seqfile);
  
  
  if(seqfile_b != NULL) {
    if (! SeqfileFormat(seqfile_b, &format, NULL))
      Die("Failed to determine format of sequence file %s\n", seqfile_b);
    
    if (nseqs > 0) {
      /* determine maximum length for sequence file (Lmax)
       *
       */
      if ((sqfp_b = SeqfileOpen(seqfile_b, format, NULL)) == NULL)
	Die("Failed to open sequence file %s", seqfile_b);
      CheckMaxLength(sqfp_b, format, &curr_Lmax, &nseqs, maxlenhit);
      if (curr_Lmax > Lmax) Lmax = curr_Lmax;
      SeqfileClose(sqfp_b);
    }
    
    if ((sqfp_b = SeqfileOpen(seqfile_b, format, NULL)) == NULL)
      Die("Failed to open sequence file %s", seqfile_b);
  }

  if (alltoall && seqfile_b == NULL) {
    seqfile_b = seqfile;
    if ((sqfp_b = SeqfileOpen(seqfile_b, format, NULL)) == NULL)
      Die("Failed to open sequence file %s", seqfile_b);
  }
  
  /* Print banner
   */
   PrintBanner(ofp, pamfile, tfactor, cfgfile, seqfile, seqfile_b, allow_pseudoknots, nseqs, 
	       Lmax, maxlenhit, minlenhit, sweep, window, slide);
  

  /* Contruct the models [logodds = FALSE (probs in log2     form)]
   *                     [logodds = TRUE  (probs in log2odds form)]
   */
   if (changefreq || changefreqwin)
     ConstructModels_phase1(ofp, hexapfile, pamfile, cfgfile, &model, &pamcond, &cfg_node, &hexa,
			    &pair5prob, add_hexamer, logodds, pedantic, tfactor, verbose);
   else 
     ConstructModels(ofp, hexapfile, pamfile, cfgfile, &model, tfactor, logodds, pedantic, verbose);
   
  /* Allocate space for iseg's, these arrays keep the full alignments
   */
  AllocIntSeqs(Lmax, &isegX, &isegY);
  
  /* Allocate space for arrays used in the calculation [iseq's, dpd, dpf, mx, sc->ardiag, sc->arfull]
   */
  Lw = (window > -1)? window : Lmax; /* window of calculation */
  AllocIntSeqs(2*Lw, &iseqX, &iseqY);
  if (d.vidiag || d.vifull || d.visemi || d.frsemi) AllocAli(2*Lw, &ali);

  /* Allocate for DP
   */
  /* SCFG part of the RNA model 
   */
  mx = AllocScfgRNA(2*Lw, rnass);
  
  /* HMM-COMPONENT of 3 models 
   */
  if (d.global || d.frdiag || d.vidiag || d.frsemi || d.visemi) 
    dpd = AllocDpDiagMtx(2*Lw); 
  if (d.frsemi || d.visemi) 
    dpf = AllocDpSemiMtx(2*Lw); 
  if (d.frfull || d.vifull) 
    dpf = AllocDpFullMtx(Lw, Lw); 
  
  /* Allocate scores
   */
  sc = AllocScores();
  if (d.frdiag || d.frsemi) sc->ardiag = AllocDpDiagScores(2*Lw);
  if (d.frfull || d.frsemi) sc->arfull = AllocDpScores(2*Lw);
  
  if (ran) {
    if (seqfile_b == NULL) Die ("this ran version requires two files");
    while (ReadGapSeq(sqfp_b, format, &seqY, &sqinfoY)) num_seqs++;
    FreeSequence(seqY, &sqinfoY);  
  }
  else if (sweep) {
    if (seqfile_b == NULL) Die ("this motif-scanning version requires two files");
  }
  else {
    while (seqfile_b != NULL && num++ < skip) {
      ReadGapSeq(sqfp_b, format, &seqY, &sqinfoY);
      FreeSequence(seqY, &sqinfoY);  
    }
  }
  
  /* Read seqX (sequence may have gaps)
   */
  while (ReadGapSeq(sqfp, format, &seqX, &sqinfoX))
    {
      nXseq++;
      /* Read seqY (sequence may have gaps)
       */
      if (sweep || alltoall) 
	{
	  while (ReadGapSeq(sqfp_b, format, &seqY, &sqinfoY)) { 
	    nYseq++;
	    num_ali++;
	    analyze_2_sequences(ofp, format, num_ali,
				sqinfoX, seqX, isegX, iseqX, ctX, 
				sqinfoY, seqY, isegY, iseqY, ctY, 
				Lmax, Lw, maxlenhit, minlenhit, window, slide, pamcond, 
				cfg_node, hexa, pair5prob, 
				d, model, dpd, dpf, mx, sc, ali, 
				add_hexamer, allow_pseudoknots, alignment, cyk, changefreq, changefreqwin, logodds, 
				pedantic, shuffle, 0, sweep, tfactor, traceback, verbose, rnass, shtoo);
	    if (sweep) {
	      RevCompChar(revX, seqX, sqinfoX.len);
	      analyze_2_sequences(ofp, format, num_ali,
				  sqinfoX, revX, isegX, iseqX, ctX, 
				  sqinfoY, seqY, isegY, iseqY, ctY, 
				  Lmax, Lw, maxlenhit, minlenhit, window, slide, pamcond, 
				  cfg_node, hexa, pair5prob, 
				  d, model, dpd, dpf, mx, sc, ali, 
				  add_hexamer, allow_pseudoknots, alignment, cyk, changefreq, changefreqwin, logodds, 
				  pedantic, shuffle, 1, sweep, tfactor, traceback, verbose, rnass, shtoo);
	    }
	    FreeSequence(seqY, &sqinfoY); 
	  }
	  /* reopen the sqfp_b file to go over it with the next motif
	   */
	  SeqfileClose(sqfp_b);
	  nYseq = 0;
	  if ((sqfp_b = SeqfileOpen(seqfile_b, format, NULL)) == NULL)
	    Die("Failed to open sequence file %s", seqfile_b);
	  if (seqfile_b == seqfile)     
	    while (nYseq++ < nXseq) {
	      ReadGapSeq(sqfp_b, format, &seqY, &sqinfoY);
	      FreeSequence(seqY, &sqinfoY); 
	    }
	}
      else if (ran) {
	/* reopen the sqfp_b file to go over it with the next motif
	 */
	SeqfileClose(sqfp_b);
	if ((sqfp_b = SeqfileOpen(seqfile_b, format, NULL)) == NULL)
	  Die("Failed to open sequence file %s", seqfile_b);

	s = (int)(sre_random()*num_seqs);
	
	num = 0;
	while (seqfile_b != NULL && num++ < s) {
	  ReadGapSeq(sqfp_b, format, &seqY, &sqinfoY);
	  FreeSequence(seqY, &sqinfoY);  
	}
	ReadGapSeq(sqfp_b, format, &seqY, &sqinfoY);

	analyze_2_sequences(ofp, format, ++nYseq,
			    sqinfoX, seqX, isegX, iseqX, ctX, 
			    sqinfoY, seqY, isegY, iseqY, ctY, 
			    Lmax, Lw, maxlenhit, minlenhit, window, slide, pamcond, 
			    cfg_node, hexa, pair5prob, 
			    d, model, dpd, dpf, mx, sc, ali, 
			    add_hexamer, allow_pseudoknots, alignment, cyk, changefreq, changefreqwin, logodds, 
			    pedantic, shuffle, 0, sweep, tfactor, traceback, verbose, rnass, shtoo);
	FreeSequence(seqY, &sqinfoY);  
      }
      else if (!sweep && ReadGapSeq((seqfile_b == NULL)? sqfp:sqfp_b, format, &seqY, &sqinfoY)) 
	{ 
	  nYseq++;
	  
	  if (nXseq != nYseq) 
	    Die("seqs %c and %c are not in sync", sqinfoX.name, sqinfoY.name);
	  
	  analyze_2_sequences(ofp, format, nYseq,
			      sqinfoX, seqX, isegX, iseqX, ctX, 
			      sqinfoY, seqY, isegY, iseqY, ctY, 
			      Lmax, Lw, maxlenhit, minlenhit, window, slide, pamcond, 
			      cfg_node, hexa, pair5prob, 
			      d, model, dpd, dpf, mx, sc, ali, 
			      add_hexamer, allow_pseudoknots, alignment, cyk, changefreq, changefreqwin, logodds, 
			      pedantic, shuffle, 0, sweep, tfactor, traceback, verbose, rnass, shtoo);
	  
	  FreeSequence(seqY, &sqinfoY);  
	}
      FreeSequence(seqX, &sqinfoX);  
    }
  
  /* Cleanup */
  if (outfile   != NULL) fclose(ofp); 
  if (seqfile   != NULL) SeqfileClose(sqfp); 
  if (seqfile_b != NULL) SeqfileClose(sqfp_b); 
  
  if (sweep) { free(revX); }
  
  free(isegX); 
  free(isegY); 
  free(iseqX); 
  free(iseqY);  
  
  FreeScfgRNA(mx, rnass); 
  
  if (add_hexamer) free(hexa);
  
  if (d.vidiag || d.vifull) FreeAli(ali); 
  
  if (d.global || d.frdiag || d.vidiag || d.frsemi || d.visemi) 
    FreeDpDiagMtx(dpd);  
  if (d.frsemi || d.visemi) 
    FreeDpSemiMtx(dpf); 
  if (d.frfull || d.vifull) 
    FreeDpFullMtx(dpf); 
  
  if (d.frdiag || d.frsemi) FreeDpScores(sc->ardiag); 
  if (d.frfull || d.frsemi) FreeDpScores(sc->arfull);

  FreeScores(sc);
  FreeModels(model); 

  if (changefreq || changefreqwin) {
    free(pair5prob);
    FreePamCond(pamcond);
    FreeSCFGNode(cfg_node);
  }
  
   return EXIT_SUCCESS;
}






void
analyze_2_sequences (FILE *ofp, int format, int num_align,
		     SQINFO sqinfoX, char *seqX, int *isegX, int *iseqX, int *ctX,
		     SQINFO sqinfoY, char *seqY, int *isegY, int *iseqY, int *ctY,
		     long int Lmax, int Lw, long int maxlenhit, int minlenhit,
		     int window, int slide,
		     struct pamcond_s  *pamcond,
		     double ***cfg_node,
		     double **hexa,
		     double *pair5prob,
		     struct dos_s d, 
		     struct model_s *model, 
		     struct dpd_s *dpd, 
		     struct dpf_s *dpf, 
		     struct rnascfg_s *mx, 
		     struct scores_s *sc, 
		     struct ali_s *ali,
		     int add_hexamer, int allow_pseudoknots, int alignment, int cyk, int changefreq, int changefreqwin, 
		     int logodds, int pedantic, int shuffle, int strand, int sweep, int tfactor, 
		     int traceback, int verbose, int rnass, int shtoo)
{
  int     i;
  int     pos;
  int     dis;
  int     win;
  int     leg;
  char   *cc;
  double *freqX;  /* base composition of fragment X*/
  double *freqY;  /* base composition of fragment X*/

  if (minlenhit >= maxlenhit) Die ("You want alignments with positive length, right?\n");
  if ((sqinfoX.len > maxlenhit || sqinfoY.len > maxlenhit)) {
    Warn("Too long ++ skip hit # %d\n>%s (%d)\n>%s (%d)\n", num_align, sqinfoX.name, sqinfoX.len, 
	 sqinfoY.name, sqinfoY.len);
    return;
  }
  if ((sqinfoX.len < minlenhit || sqinfoY.len < minlenhit)) {
    Warn("Too short ++ skip hit # %d\n>%s (%d)\n>%s (%d)\n", num_align, sqinfoX.name, sqinfoX.len, 
	 sqinfoY.name, sqinfoY.len);
    return;
  }

  /* Print alignment banner
   */
  PrintHitBanner(ofp, sqinfoX, sqinfoY, num_align, shuffle, strand);
  
  /* Create arrays with given aligment:
   *    isegX[sqinfoX->len]         
   *    isegY[sqinfoY->len]    (int seqs with gaps) 
   */
  FormatSeqs(ofp, Lmax, format, seqX, seqY, &sqinfoX, &sqinfoY, &ctX, &ctY, isegX, isegY, 
	     allow_pseudoknots, verbose);

  
  /* (0) leg = minimum length of aligment 
   *
   * this function can be passed to ScoreWithModels() in "scorewithmodels.c" if 
   * one wants to remove the common gaps of the two alignments window by window.
   * The advantage of that is that you keep a common coordinate sistem,
   * the disadvantage is that the "actual" window used may be smaller than
   * the given one.
   */
  if (sweep) window = sqinfoX.len;
  RemoveJointGaps(sqinfoX.len, isegX, sqinfoY.len, isegY, &leg, sweep);
  if (window != -1) fprintf(ofp, "length of whole alignment after removing common gaps: %d \n\n", leg); 

  if (window == -1) { win = leg;    slide = leg; }
  else              { win = window;              }
  
  if (changefreq || changefreqwin) {
    freqX = (double *) MallocOrDie (sizeof(double)*4);
    freqY = (double *) MallocOrDie (sizeof(double)*4);
  }

  /* Calculate nucleotide freqs for the whole alignment
   */
  if (changefreq) {
    /* calculate single-nt frequencies for whole alignment
     */
    for (i = 0; i < 4; i++)         /* initialization         */
      {
	freqX[i] = 0.0;   
	freqY[i] = 0.0;   
      }
    BaseComp(ofp, isegX, leg-1, leg-1, freqX);    /* freqs for seqX         */
    BaseComp(ofp, isegY, leg-1, leg-1, freqY);    /* freqs for seqY         */
    
    /* Use the base-composition to construct the models
     */
    ConstructModels_phase2(ofp, pamcond, cfg_node, hexa, pair5prob, freqX, freqY, model, add_hexamer,  
			   TRUE, logodds, pedantic, tfactor, verbose);
  }   
  
  /* Calculate scores from pos to pos+win-1
   */
  pos = 0;
  while (pos < leg) {
    
    dis = (pos<leg-win)? win: leg-pos;

    if(changefreqwin) {
      /* calculate single-nt frequencies for window: [pos, pos+dis-1]
       */
      for (i = 0; i < 4; i++)         /* initialization         */
	{
	  freqX[i] = 0.0;   
	  freqY[i] = 0.0;   
	}
      BaseComp(ofp, isegX, pos+dis-1, dis-1, freqX);    /* freqs for seqX         */
      BaseComp(ofp, isegY, pos+dis-1, dis-1, freqY);    /* freqs for seqY         */
      
      /* Use the base-composition to construct the models
       */
      ConstructModels_phase2(ofp, pamcond, cfg_node, hexa, pair5prob, freqX, freqY, model, add_hexamer, 
			     TRUE, logodds, pedantic, tfactor, verbose);
    }
    
    /* the "given" RNA secondary structure (with gaps) if any goes to cc 
     */
    if ((format == kSquid || format == kSelex) && (sqinfoX.flags & SQINFO_SS)) {
      Ct2KHS_ER(pos, dis, ctX, &cc);
      sqinfoX.ss = cc;
      if (verbose) PrintCtSeqs(ofp, &sqinfoX, isegX, &sqinfoY, isegY, pos, dis, cc);
    }

    ScoreWithModels(ofp, sqinfoX, isegX, iseqX, sqinfoY, isegY, iseqY, pos, 
		    dis, Lw, d, model, dpd, dpf,
		    mx, sc, ali, alignment, cyk, logodds, rnass, shuffle, sweep, traceback, verbose);
    
    if (!shuffle && shtoo)
      ScoreWithModels(ofp, sqinfoX, isegX, iseqX, sqinfoY, isegY, iseqY, pos, 
		      dis, Lw, d, model, dpd, dpf,
		      mx, sc, ali, alignment, cyk, logodds, rnass, TRUE, sweep, traceback, verbose);
    
     pos = (pos<leg-win)? pos+slide : leg;
  }

  if (changefreq || changefreqwin) { 
    free(freqX); 
    free(freqY); 
  }
  if ((format == kSquid || format == kSelex) && (sqinfoX.flags & SQINFO_SS)) { 
    free(ctX);   
    free(ctY);   
    free(cc);
  }
}
