/************************************************************
 * QRNA - Comparative analysis of biological sequences 
 *        with pair hidden Markov models and pair stochastic context-free grammars
 * Copyright (C) 2000-Howard Hughes Medical Institute
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* qrna_sample.c
 * 
 * Creates alignments according to the different models
 *
 * ER, Tue Feb  1 12:50:24 CST 2000 [ST. Louis]
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <time.h>

#include "funcs.h"
#include "globals.h"
#include "squid.h"
#include "structs.h"
#include "version.h"


#ifdef MEMDEBUG
#include "dbmalloc.h"
#endif

static struct opt_s OPTIONS[] = {
  { "-a",        TRUE,  sqdARG_NONE},
  { "-C",        TRUE,  sqdARG_NONE},
  { "-c",        TRUE,  sqdARG_STRING},
  { "-F",        TRUE,  sqdARG_NONE},
  { "-h",        TRUE,  sqdARG_NONE},
  { "-i",        TRUE,  sqdARG_STRING},
  { "-l",        TRUE,  sqdARG_STRING},
  { "-m",        TRUE,  sqdARG_STRING},
  { "-n",        TRUE,  sqdARG_NONE},
  { "-o",        TRUE,  sqdARG_STRING},
  { "-p",        TRUE,  sqdARG_STRING},
  { "-q",        TRUE,  sqdARG_STRING},
  { "-R",        TRUE,  sqdARG_NONE},
  { "-t",        TRUE,  sqdARG_NONE},
  { "-v",        TRUE,  sqdARG_NONE},
};
                
#define NOPTIONS (sizeof(OPTIONS) / sizeof(struct opt_s))

static char usage[]  = "\
Usage: qrna_sample [-options] \n\
where options are:\n\
   -a             : print alignment \n\
   -C             : create a COD-like alignment (default is OTH)\n\
   -c <cfgfile>   : <cfgfile> to use to train the rna model (default = tRNA+rRNA)\n\
   -F             : change the overall base composition of the 3 models\n\
   -h             : print short help and usage info\n\
   -i <tinit>     : minimum evolutionary time factor (default tinit=0)\n\
   -l <maxlenhit> : change the maxlenhit parameter (default 1000)\n\
   -m <tmax>      : maximum evolutionary time factor (default tmax=12)\n\
   -n             : create a null-like alignment (default is OTH)\n\
   -o <outfile>   : direct structure-annotated sequence to <outfile>\n\
   -p <pamfile>   : <pamfile> to use (default = BLOSUM62)\n\
   -q <num>       : number of alingments to be created (default = 1)\n\
   -R             : create a RNA-like alignment (default is OTH)\n\
   -t             : print traceback\n\
   -v             : verbose debugging output\n\
";

static char banner[] = "qrna_sample";

int
main(int argc, char **argv)
{
  char    *seqX;	          /* sequence X                                             */
  char    *seqY;	          /* sequence to compare                                    */
  int     *iseqX, *isegX;	  /* sequence X integer form (without and with gaps)        */
  int     *iseqY, *isegY;	  /* sequence Y integer form (without and with gaps)        */
  SQINFO   sqinfoX;               /* info structures for seqX                               */
  SQINFO   sqinfoY;               /* info structures for seqY                               */
  int     *ct;                    /* .ct notation for seq's RNA structure (if any)         */
  int      leg;                   /* length of a given alignment                            */
  int      Lmax;                  /* length of the max alignment                            */

  char    *hexapfile;             /* Hexamer joint frequencies                              */
  char    *pamfile;               /* PAM substitution matrix                                */
  char    *cfgfile;               /* RNA grammar file                                       */

  struct   model_s     *model;    /* transition + emission probs 3 models + null model      */

  struct   ali_s *ali;            /* arrays to store the alignment created by full viterbi  */ 


  int     alignment;	          /* TRUE prints alignment                                  */
  int     changefreq;             /* TRUE change the baseline base composition of 3 models  */
  int     cod;	                  /* TRUE create a COD-like  alignment                      */
  int     null;	                  /* TRUE create a null-like  alignment                     */
  int     rna;	                  /* TRUE create a RNA-like  alignment                      */
  double  tfactor;                /* evolutionary time factor                               */
  double  tinc, tinit, tmax;      /* time increment, min amd max times considered           */
  int     traceback;              /* TRUE to traceback alignment                            */
  int     pedantic;               /* TRUE do some checks for evolutionary models to debug   */
  int     verbose;                /* TRUE to be extremely verbose to debug                  */

  char   *outfile;                /* where to send the output                               */
  FILE   *ofp;	                  /* open output file                                       */
  char   *optname;
  char   *optarg; 
  char   *string_name;
  int     num = 0, num_align;
  int     optind;	
  int     seed;

  double  id;                     /* percentage identities in the alignments                */
  double  gap;                    /* percentage of gaps in the alignments                   */
  double  mut;                    /* percentage mutations in the alignments                 */


  /* re-seed the random number generator.
   */
  seed = (int) time ((time_t *) NULL);
  sre_srandom(seed); /* reinit sre_random each time you shuffle a sequence */

  /* Parse command line
   */
  tinc     =  0.01;    /* time increments (if tinit < tmax)  */
  tinit    =  0.00;    /* default minimum evolutionary time  */
  tmax     = 12.00;    /* default maximum evolutionary time  */

  alignment  = FALSE;      /* TRUE  ==  prints alignment                      */
  changefreq = FALSE;      /* TRUE  ==  change the base comp of 3 models      */
  cod        = FALSE;      /* TRUE  ==  create a COD-like  alignment          */
  null       = FALSE;      /* TRUE  ==  create a null-like  alignment         */
  rna        = FALSE;      /* TRUE  ==  create a RNA-like  alignment          */
  traceback  = FALSE;      /* TRUE  ==  traceback alignment                   */
  pedantic   = FALSE;      /* TRUE  ==  check your evolutionary models        */
  verbose    = FALSE;      /* TRUE  ==  for debuging                          */
  Lmax       = 500000;
  
  cfgfile   = NULL;
  pamfile   = "BLOSUM62";
  outfile   = NULL;
  leg       = Lmax;
  num_align = 1;

  while (Getopt(argc, argv, OPTIONS, NOPTIONS, usage,
		&optind, &optname, &optarg))
    {
      if      (strcmp(optname, "-a") == 0)  alignment  = TRUE;
      else if (strcmp(optname, "-C") == 0)  cod        = TRUE;
      else if (strcmp(optname, "-c") == 0)  cfgfile    = optarg;
      else if (strcmp(optname, "-F") == 0)  changefreq = TRUE;
      else if (strcmp(optname, "-i") == 0)  tinit      = atof(optarg);
      else if (strcmp(optname, "-l") == 0)  leg        = atoi(optarg);
      else if (strcmp(optname, "-m") == 0)  tmax       = atof(optarg);
      else if (strcmp(optname, "-n") == 0)  null       = TRUE;
      else if (strcmp(optname, "-o") == 0)  outfile    = optarg;
      else if (strcmp(optname, "-p") == 0)  pamfile    = optarg;
      else if (strcmp(optname, "-q") == 0)  num_align  = atoi(optarg);
      else if (strcmp(optname, "-R") == 0)  rna        = TRUE;
      else if (strcmp(optname, "-t") == 0)  traceback  = TRUE;
      else if (strcmp(optname, "-v") == 0)  verbose    = TRUE;
      else if (strcmp(optname, "-h") == 0) 
	{
	  puts(banner);
	  printf("          %s (%s)", RELEASE, RELEASEDATE);
	  printf(" using squid %s (%s)\n", squid_version, squid_date);
	  puts(usage);
	  exit(0);
	}
    }

  tfactor  = tinit; /* set the time to start with */

  if (argc - optind != 0)
    Die("Incorrect number of command line arguments.\n%s\n", usage);

 if (cfgfile == NULL) /* SCFG is not provided */
    cfgfile = FileConcat("/nfs/wol2/people/elena/qrna/db/rna", "mix_tied_linux.cfg");
 
  if (leg > Lmax) {
    Warn ("too long, Lmax = %d\n", Lmax);
    leg = Lmax;
  }
  
  /* Open output file 
   */
  ofp = stdout;
  if (outfile != NULL && (ofp = fopen(outfile, "w")) == NULL)
    Die("Failed to open output file %s", outfile);

  /* Allocate space for the sequences
   */
  AllocCharSeqs(leg, &seqX, &seqY);
  AllocIntSeqs(leg, &isegX, &isegY);
  AllocIntSeqs(leg, &iseqX, &iseqY);
  AllocCt(leg, &ct);
  AllocAli(leg, &ali);

  while (tfactor <= tmax) {
    /* Contruct the models 
     *    at a given evolutionary time "tfactor"
     */
    ConstructModels(ofp, hexapfile, pamfile, cfgfile, &model, tfactor, FALSE, pedantic, verbose);
    
    id  = 0.0;
    gap = 0.0;
    mut = 0.0;

    while (num++ < num_align) {
      if (null) {
	string_name = "null";
	SimulateNullAlign(isegX, isegY, leg, leg, model->null);
      }
      else if (cod) {
	string_name = "cod";
	SimulateCODAlign(ofp, &sqinfoX, isegX, &sqinfoY, isegY, 0, leg, model->cod, model->null, ali, 
			 traceback, alignment, string_name);    
      }
      else if (rna) {
	string_name = "rna";
	SimulateRNAAlign(ofp, &sqinfoX, isegX, &sqinfoY, isegY, 0, leg, model->rna, model->null, ali, 
			 ct, traceback, alignment, string_name);    
      }
      else {
	string_name = "oth";
	SimulateOTHAlign(ofp, &sqinfoX, isegX, &sqinfoY, isegY, 0, leg, model->oth, ali, 
			 traceback, alignment, string_name);
      }
      
      AlignStat(ofp, &sqinfoX, isegX, &sqinfoY, isegY, leg, &id, &gap, &mut);
      PrintAlignSequences(ofp, &sqinfoX, &sqinfoY, string_name, tfactor, 0, leg, isegX, isegY);
    }

    printf("# align %d\n", num_align);
    printf("# time  #id    #gap   #mut  #nogaps\n");
    printf("%.2f %.4f %.4f %.4f %.4f \n", tfactor, id /num_align, gap/num_align, 
	   mut/num_align, (id+mut)/num_align);

    num = 0;
    tfactor += tinc; /* increment the evolutionary time */
    FreeModels(model);
  }
  
  
  /* Cleanup
  if (outfile != NULL) fclose(ofp);
  free(seqX);
  free(seqY);
  free(isegX);
  free(isegY);
  free(iseqX);
  free(iseqY);
  free(ct);
  FreeAli(ali);
   */
  return EXIT_SUCCESS;
}









