/************************************************************
 * QRNA - Comparative analysis of biological sequences 
 *        with pair hidden Markov models and pair stochastic context-free grammars
 * Copyright (C) 2000-Howard Hughes Medical Institute
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* rnacreatealign.c
 *
 * ER, Wed Jun  2 09:43:57 CDT 1999 [St. Louis]
 * 
 * creates an alignment acoding to the rnamodel
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

#include "funcs.h"
#include "globals.h"
#include "squid.h"
#include "structs.h"

#ifdef MEMDEBUG
#include "dbmalloc.h"
#endif

static void coemitIS1(FILE *ofp, int *seqX, int *seqY, struct rna_loop5 *is1, int L, int j, int d, double *pnode_is1,
		      struct tracekn_s *cur_tr, struct traceknstack_s *dolist, struct ali_s *ali, 
		      int traceback, int *ct);
static void coemitIS2(FILE *ofp, int *seqX, int *seqY, struct rna_loop5 *is2, int L, int j, int d, 
		      int d1, int d2, double *pnode_is2,
		      struct tracekn_s *cur_tr, struct traceknstack_s *dolist, struct ali_s *ali, 
		      int traceback, int *ct);
static void coemitV(FILE *ofp, int *seqX, int *seqY, struct rna_v5 *v, int L, int j, int d, double *pnode_v,
		    double *pnode_is1, double *pnode_is2, struct tracekn_s *cur_tr, struct traceknstack_s *dolist, 
		    struct ali_s *ali, int traceback, int *ct);
static void coemitW(FILE *ofp, int *seqX, int *seqY, struct rna_w5 *w, int L, int j, int d, double *pnode_w,
		    struct tracekn_s *cur_tr, struct traceknstack_s *dolist, struct ali_s *ali, 
		    int traceback, int *ct);

static void emitIS1(FILE *ofp, int *seq, int *seqX, int *seqY, struct rna_loop5 *is1, int L, int j, int d, 
		    double *pnode_is1, struct tracekn_s *cur_tr, struct traceknstack_s *dolist, 
		    struct ali_s *ali, int traceback, int *ct);
static void emitIS2(FILE *ofp, int *seq, int *seqX, int *seqY, struct rna_loop5 *is2, int L, int j, int d, 
		    int d1, int d2, double *pnode_is2,
		    struct tracekn_s *cur_tr, struct traceknstack_s *dolist, struct ali_s *ali, 
		    int traceback, int *ct);
static void emitV(FILE *ofp, int *seq, int *seqX, int *seqY, struct rna_v5 *v, int L, int j, int d, double *pnode_v,
		  double *pnode_is1, double *pnode_is2, struct tracekn_s *cur_tr, struct traceknstack_s *dolist, 
		  struct ali_s *ali, int traceback, int *ct);
static void emitW(FILE *ofp, int *seq, int *seqX, int *seqY, struct rna_w5 *w, int L, int j, int d, double *pnode_w,
		  struct tracekn_s *cur_tr, struct traceknstack_s *dolist, struct ali_s *ali, 
		  int traceback, int *ct);

/* Function: SimulateNUSSequences()
 * Date:     ER, Wed Jun  9 17:21:30 CDT 1999 [St. Louis]
 *
 * Purpose:  Simulated an ungapped alignment of two sequences
 *           generated by the RNA model. Not a complete generation:
 *           rather, is given a structure (implying a given parse tree)
 *           and just generates residues.
 *           
 * Args:     ct     : Zuker .ct notation for an RNA structure 
 *                      e.g. 0..L-1 values if paired -1 if not          
 *           L      : length of structure
 *           ret_sX : RETURN: string X
 *           ret_sY : RETURN: string Y
 *
 * Returns:  void
 */
void
SimulateNUSSequences(struct rnamodel_s *rna, int *ct, int L, char **ret_sX, char **ret_sY)
{
  char   *sX, *sY;
  double  pair[256];
  double  single[16];
  int     z;
  int     x1,x2,y1,y2;
  int     i,j;
  double  totsc;

  sX = MallocOrDie(sizeof(char) * (L+1));
  sY = MallocOrDie(sizeof(char) * (L+1));

  for (z = 0; z < 16; z++) 
    {
      x1 = z / 4;
      y1 = z % 4;
      single[z] = rna->nus->p[idx(x1,y1)];
    }
  for (z = 0; z < 256; z++)
    {
      x1 = z / 64;
      x2 = (z % 64) / 16;
      y1 = (z % 16) / 4;
      y2 = z % 4;
      pair[z] = rna->nus->pp[idx(x1,y1)][idx(x2,y2)];
    }

  totsc = 0.0;
  for (i = 0; i < L; i++)
    {
      if (ct[i] == -1)		/* singlet */
	{
	  z  = DChoose(single, 16);
	  x1 = z / 4;
	  y1 = z % 4;

	  sX[i] = DNAAlphabet[x1];
	  sY[i] = DNAAlphabet[y1];
	  
	  totsc += rna->nus->p[idx(x1,y1)];

	}
      else if (ct[i] > i)	/* doublet that we haven't done yet */
	{
	  j = ct[i];
	  z = DChoose(pair, 256);
	  x1 = z / 64;
	  x2 = (z % 64) / 16;
	  y1 = (z % 16) / 4;
	  y2 = z % 4;
	  
	  sX[i] = DNAAlphabet[x1];
	  sY[i] = DNAAlphabet[y1];
	  sX[j] = DNAAlphabet[x2];
	  sY[j] = DNAAlphabet[y2];

	  totsc += rna->nus->pp[idx(x1,y1)][idx(x2,y2)];
	}
    }    


  /* specific correction for tRNAPhe parse tree
   * (-82.23)
   */
  totsc += 34.0 * rna->nus->tl;
  totsc += 21.0 * rna->nus->tv;
  totsc += 2.0  * rna->nus->tw;
  totsc += 3.0  * rna->t[TRNAROE];

  /* printf("simulated RNAs should score at least %f\n", totsc); */
  sX[L] = '\0';
  sY[L] = '\0';
  *ret_sX = sX;
  *ret_sY = sY;
  return;
}


/* Function: SimulateRNAAlign()
 * 
 * Date:     ER, Thu Feb 10 10:18:06 CST 2000 [St. Louis]
 * 
 * Purpose:  Create an alignment of length L according to the RNA model.
 *           
 * Args:     ofp       - print to file
 *           sqinfoX   - info about sequence X
 *           seqX      - sequence X, in integer form (A=0, C=1, G=2, U=3)
 *           sqinfoY   - info about sequence Y
 *           seqY      - sequence Y, in integer form (A=0, C=1, G=2, U=3)
 *           start     - starting position in the arrays
 *           L         - length of alignment, which is 0..len-1
 *           rna       - the SCFG model
 *           null      - the Null model
 *           ali       - struture of arrays to store alingmend sequences
 *           traceback - if TRUE print walk through model
 *           alignment - if TRUE print  alignment
 *           ctX       - array to store structure
 *
 * Return:   void. seqX, seqY, ali, ct are filled here.
 */
void
SimulateRNAAlign(FILE *ofp, SQINFO *sqinfoX, int *seqX, SQINFO *sqinfoY, int *seqY, int start, 
		 int L, struct rnamodel_s *rna, struct nullmodel_s *null, struct ali_s *ali, 
		 int *ct, int traceback, int alignment, char *string_name)
{
  struct tracekn_s      *tr;                   /* the traceback tree under construction         */
  struct tracekn_s      *cur_tr;               /* ptr to node of tr we're working on            */
  struct traceknstack_s *dolist;               /* pushdown stack of active tr nodes             */
  double                 pnode[NDPS][NNODES];  /* stores the transition probabilies of the SCFG */
  int                    state, node;          /* state and node we are at                      */
  int                    i,j,k,l;              /* coords in mtx's                               */
  int                    d, d1, d2;            /* coords in mtx's                               */
  int                    size;                 /* size of loops                                 */

  for (state = 0; state < NDPS; state++)
    for (node = 0; node < NNODES; node++)
      pnode[state][node] = -BIGFLOAT;

  pnode[V][dpcS1]  = rna->pi2->v->t1;
  pnode[V][dpcS2S] = rna->pi2->v->t2s;
  pnode[V][dpcS2B] = rna->pi2->v->t2b;
  pnode[V][dpcS2I] = rna->pi2->v->t2i;
  pnode[V][dpcMV]  = rna->pi2->v->t3;
  
  pnode[W][dpcL]   = rna->pi2->w->tl;
  pnode[W][dpcR]   = rna->pi2->w->tr;
  pnode[W][dpcP]   = rna->pi2->w->tv;
  pnode[W][dpcBW]  = rna->pi2->w->tw;

  pnode[WB][dpcL]  = rna->pi2->wb->tl;
  pnode[WB][dpcR]  = rna->pi2->wb->tr;
  pnode[WB][dpcP]  = rna->pi2->wb->tv;
  pnode[WB][dpcBW] = rna->pi2->wb->tw;

  for (size = 0; size < MAXRNALOOP; size++) {
    pnode[IS1][dpcLEN(size)] = rna->pi2->is1->tn[size];
    pnode[IS2][dpcLEN(size)] = rna->pi2->is2->tn[size];
  }  
  
  /* Initialize.
   * Start at j = L-1, d = L-1, d1 = (int)(d/2), d2 = d - (int)(d/2) - 1.
   */
  tr     = InitTracekn();       /* start a trace tree */
  dolist = InitTraceknstack();	/* start a stack for traversing the trace tree */

  if (L == 0) return;
  else d = L - 1;
  
  j = start + d;
  i = j - d;
  k = i + (int)(d/2);
  l = k + 1;
  
  if (traceback) fprintf(ofp,"---------------------------------------------------\n");

 /* assigned state to (i,j):  V 
   *                          W 
   *                          WB
   *                          IS1
   *                          IS2
   */

  cur_tr = AttachTracekn(tr, i, j, k, l, W, dpcEW); 
  PushTraceknstack(dolist, cur_tr);

  /* Recursion. While there's active nodes in the stack, trace from them.
   * 
  */
  while ((cur_tr = PopTraceknstack(dolist)) != NULL)
   {
     /* get some useful numbers, mostly for clarity */
      i = cur_tr->emiti;
      j = cur_tr->emitj;
      k = cur_tr->emitk;
      l = cur_tr->emitl;
      
      state = cur_tr->type;

      d  = j - i;
      
      if (d == 0) continue;

      d1 = k - i;
      d2 = j - l;
      
      if (d < 5) 
	{
	  if(state == V) {
	    PushTraceknstack(dolist, AttachTracekn(cur_tr, i, j, k, k+1, IS1, dpcEIS));
	    if (traceback) fprintf(ofp," emit IS1 (%d)  \n", d); 
	    continue;
	  }
	}
      
      if (traceback) fprintf(ofp,"---------------------------------------------------\n");
      if (traceback) fprintf(ofp,"%s %d %d %d %d  \n", stNAME[cur_tr->type], j, d, d1, d2); 
      
      switch (state){
      case IS1: 
	coemitIS1(ofp, seqX, seqY, rna->pi2->is1, L, j, d, pnode[state], cur_tr, dolist, ali, 
		  traceback, ct);
	break;
	
      case IS2: 
	coemitIS2(ofp, seqX, seqY, rna->pi2->is2, L, j, d, d1, d2, pnode[state], cur_tr, dolist, ali, 
		  traceback, ct);
	break;
	
      case V: 
	coemitV(ofp, seqX, seqY, rna->pi2->v, L, j, d, pnode[state], pnode[IS1], pnode[IS2], 
		cur_tr, dolist, ali, traceback, ct);
	break;
	
      case W: 
	coemitW(ofp, seqX, seqY, rna->pi2->w, L, j, d, pnode[state], cur_tr, dolist, ali, 
		traceback, ct);
	break;
	
      case WB: 
	coemitW(ofp, seqX, seqY, rna->pi2->wb, L, j, d, pnode[state], cur_tr, dolist, ali, 
		traceback, ct);
	break;
	
     default:
	Die("invalid matrix assignement in SimulateRNAAlign()");
      }
   } /* while something is in the trace stack */

  if (alignment) PrintCtAlign(ofp, sqinfoX, sqinfoY, start, L, ali, ct, string_name);
  
  FreeTracekn(cur_tr);
  FreeTraceknstack(dolist);
}


/* Function: SimulateRNASequence()
 * 
 * Date:     ER, Mon Mar 27 13:49:09 CST 2000 [St. Louis]
 * 
 * Purpose:  Given a sequence, simulate a second sequence
 *           related to it by the RNA model.
 *           
 * Args:     ofp       - print to file
 *           sqinfoX   - info about sequence X
 *           seqX      - sequence X, in integer form (A=0, C=1, G=2, U=3)
 *           sqinfoY   - info about sequence Y
 *           seqY      - sequence Y, in integer form (A=0, C=1, G=2, U=3)
 *           start     - starting position in the arrays
 *           L         - length of alignment, which is 0..len-1
 *           rna       - the SCFG model
 *           null      - the Null model
 *           ali       - struture of arrays to store alingmend sequences
 *           traceback - if TRUE print walk through model
 *           alignment - if TRUE print  alignment
 *           ctX       - array to store structure
 *
 * Return:   void. seqX, seqY, ali, ct are filled here.
 */
void
SimulateRNASequence(FILE *ofp, int *seq, SQINFO *sqinfoX, int *seqX, SQINFO *sqinfoY, int *seqY, int start, 
		    int L, struct rnamodel_s *rna, struct nullmodel_s *null, struct ali_s *ali, 
		    int *ct, int traceback, int alignment, char *string_name)
{
  struct tracekn_s      *tr;                   /* the traceback tree under construction         */
  struct tracekn_s      *cur_tr;               /* ptr to node of tr we're working on            */
  struct traceknstack_s *dolist;               /* pushdown stack of active tr nodes             */
  double                 pnode[NDPS][NNODES];  /* stores the transition probabilies of the SCFG */
  int                    state, node;          /* state and node we are at                      */
  int                    i,j,k,l;              /* coords in mtx's                               */
  int                    d, d1, d2;            /* coords in mtx's                               */
  int                    size;                 /* size of loops                                 */

  for (state = 0; state < NDPS; state++)
    for (node = 0; node < NNODES; node++)
      pnode[state][node] = -BIGFLOAT;

  pnode[V][dpcS1]  = rna->pi2->v->t1;
  pnode[V][dpcS2S] = rna->pi2->v->t2s;
  pnode[V][dpcS2B] = rna->pi2->v->t2b;
  pnode[V][dpcS2I] = rna->pi2->v->t2i;
  pnode[V][dpcMV]  = rna->pi2->v->t3;
  
  pnode[W][dpcL]   = rna->pi2->w->tl;
  pnode[W][dpcR]   = rna->pi2->w->tr;
  pnode[W][dpcP]   = rna->pi2->w->tv;
  pnode[W][dpcBW]  = rna->pi2->w->tw;

  pnode[WB][dpcL]  = rna->pi2->wb->tl;
  pnode[WB][dpcR]  = rna->pi2->wb->tr;
  pnode[WB][dpcP]  = rna->pi2->wb->tv;
  pnode[WB][dpcBW] = rna->pi2->wb->tw;

  for (size = 0; size < MAXRNALOOP; size++) {
    pnode[IS1][dpcLEN(size)] = rna->pi2->is1->tn[size];
    pnode[IS2][dpcLEN(size)] = rna->pi2->is2->tn[size];
  }  
  
 /* Initialize.
   * Start at j = L-1, d = L-1, d1 = (int)(d/2), d2 = d - (int)(d/2) - 1.
   */
  tr     = InitTracekn();       /* start a trace tree */
  dolist = InitTraceknstack();	/* start a stack for traversing the trace tree */

  if (L == 0) return;
  else d = L - 1;
  
  j = start + d;
  i = j - d;
  k = i + (int)(d/2);
  l = k + 1;
  
  if (traceback) fprintf(ofp,"---------------------------------------------------\n");

 /* assigned state to (i,j):  V 
   *                          W 
   *                          WB
   *                          IS1
   *                          IS2
   */

  cur_tr = AttachTracekn(tr, i, j, k, l, W, dpcEW); 
  PushTraceknstack(dolist, cur_tr);

  /* Recursion. While there's active nodes in the stack, trace from them.
   * 
  */
  while ((cur_tr = PopTraceknstack(dolist)) != NULL)
   {
     /* get some useful numbers, mostly for clarity */
      i = cur_tr->emiti;
      j = cur_tr->emitj;
      k = cur_tr->emitk;
      l = cur_tr->emitl;
      
      state = cur_tr->type;

      d  = j - i;
      
      if (d == 0) continue;

      d1 = k - i;
      d2 = j - l;
      
      if (d < 5) 
	{
	  if (state == V) {
	    PushTraceknstack(dolist, AttachTracekn(cur_tr, i, j, k, k+1, IS1, dpcEIS));
	    if (traceback) fprintf(ofp," emit IS1 (%d)  \n", d); 
	    continue;
	  }
	}
      
      if (traceback) fprintf(ofp,"---------------------------------------------------\n");
      if (traceback) fprintf(ofp,"%s %d %d %d %d  \n", stNAME[cur_tr->type], j, d, d1, d2); 
      
      switch (state){
      case IS1: 
	emitIS1(ofp, seq, seqX, seqY, rna->pi2->is1, L, j, d, pnode[state], cur_tr, dolist, ali, 
		traceback, ct);
	break;
	
      case IS2: 
	emitIS2(ofp, seq, seqX, seqY, rna->pi2->is2, L, j, d, d1, d2, pnode[state], cur_tr, dolist, ali, 
		traceback, ct);
	break;

      case V: 
	emitV(ofp, seq, seqX, seqY, rna->pi2->v, L, j, d, pnode[state], pnode[IS1], pnode[IS2], 
	      cur_tr, dolist, ali, traceback, ct);
	break;
	
      case W: 
	emitW(ofp, seq, seqX, seqY, rna->pi2->w, L, j, d, pnode[state], cur_tr, dolist, ali, 
	      traceback, ct);
	break;
      
      case WB: 
	emitW(ofp, seq, seqX, seqY, rna->pi2->wb, L, j, d, pnode[state], cur_tr, dolist, ali, 
	      traceback, ct);
	break;
      
     default:
	Die("invalid matrix assignement in SimulateRNAAlign()");
      }
   } /* while something is in the trace stack */

  if (alignment) PrintCtAlign(ofp, sqinfoX, sqinfoY, start, L, ali, ct, string_name);

  FreeTracekn(cur_tr);
  FreeTraceknstack(dolist);
}


void
coemitIS1(FILE *ofp, int *seqX, int *seqY, struct rna_loop5 *is1, int L, int j, int d, double *pnode_is1,
	struct tracekn_s *cur_tr, struct traceknstack_s *dolist, struct ali_s *ali, 
	int traceback, int *ct)
{
  int    i, r;
  int    pos_r;
  int    cur_xr, cur_yr;

  i = j - d;

 if (traceback && d > 1) fprintf(ofp," loop: ");
   for (r = 1; r < d ; r++) {
    pos_r = DLog2Choose(is1->ps, 25);
    
    cur_xr = pos_r / 5;
    cur_yr = pos_r % 5;
    
    seqX[i+r] = cur_xr;
    seqY[i+r] = cur_yr;
    
    ct[i+r] = -1;
    
    ali->charX[i+r] = (cur_xr < 4)? DNAAlphabet[cur_xr] : '.';
    ali->charY[i+r] = (cur_yr < 4)? DNAAlphabet[cur_yr] : '.';

    if (traceback) fprintf(ofp,"[%d %d] ", cur_xr, cur_yr); 
    }
  if (traceback) fprintf(ofp,"\n");
}

void
emitIS1(FILE *ofp, int *seq, int *seqX, int *seqY, struct rna_loop5 *is1, int L, int j, int d, double *pnode_is1,
	struct tracekn_s *cur_tr, struct traceknstack_s *dolist, struct ali_s *ali, 
	int traceback, int *ct)
{
  int    i, r;
  int    cur_xr, cur_yr;
  int    y;
  double pxy[5][5], py[5];

  i = j - d;

  /* calculate pxy[x][y]= P(y|x) = P(x,y) / P(x)
   */
  for (y = 0; y < 5; y++) 
    py[y] = 0.;

  for (y = 0; y < 25; y++)
    py[y/5] += EXP2(is1->ps[y]);

  for (y = 0; y < 5; y++)
    py[y] = LOG2(py[y]);
  
  for (y = 0; y < 25; y++) 
    pxy[y/5][y%5] = is1->ps[y] - py[y/5];

  if (traceback && d > 1) fprintf(ofp," loop: ");

   for (r = 1; r < d ; r++) {
     cur_xr = seq[i+r];
     
     cur_yr = DLog2Choose(pxy[cur_xr], 5);
     
     seqX[i+r] = cur_xr;
     seqY[i+r] = cur_yr;
     
     ct[i+r] = -1;
     
     ali->charX[i+r] = (cur_xr < 4)? DNAAlphabet[cur_xr] : '.';
     ali->charY[i+r] = (cur_yr < 4)? DNAAlphabet[cur_yr] : '.';
     
     if (traceback) fprintf(ofp,"[%d %d] ", cur_xr, cur_yr); 
   }
   if (traceback) fprintf(ofp,"\n");
}

void
coemitIS2(FILE *ofp, int *seqX, int *seqY, struct rna_loop5 *is2, int L, int j, int d, int d1, int d2, 
	double *pnode_is2, struct tracekn_s *cur_tr, struct traceknstack_s *dolist, struct ali_s *ali, 
	int traceback, int *ct)
{
  int    i, r;
  int    pos_r;
  int    cur_xr, cur_yr;

  i = j - d;

  if (traceback && d1 > 1) fprintf(ofp," loop l: ");
  for (r = 1; r < d1 ; r++) {
    pos_r = DLog2Choose(is2->ps, 25);
    
    cur_xr = pos_r / 5;
    cur_yr = pos_r % 5;
    
    seqX[i+r] = cur_xr;
    seqY[i+r] = cur_yr;
    
    ct[i+r] = -1;
    
    ali->charX[i+r] = (cur_xr < 4)? DNAAlphabet[cur_xr] : '.';
    ali->charY[i+r] = (cur_yr < 4)? DNAAlphabet[cur_yr] : '.';

    if (traceback) fprintf(ofp,"[%d %d] ", cur_xr, cur_yr); 
  }
  
  if (traceback && d2 > 1) fprintf(ofp,"\n loop r: ");
  for (r = 1; r < d2 ; r++) {
    pos_r = DLog2Choose(is2->ps, 25);
    
    cur_xr = pos_r / 5;
    cur_yr = pos_r % 5;
    
    seqX[j-r] = cur_xr;
    seqY[j-r] = cur_yr;
    
    ct[j-r] = -1;
    
    ali->charX[j-r] = (cur_xr < 4)? DNAAlphabet[cur_xr] : '.';
    ali->charY[j-r] = (cur_yr < 4)? DNAAlphabet[cur_yr] : '.';

    if (traceback) fprintf(ofp,"[%d %d] ", cur_xr, cur_yr); 
  }
  if (traceback) fprintf(ofp,"\n");
}

void
emitIS2(FILE *ofp, int *seq, int *seqX, int *seqY, struct rna_loop5 *is2, int L, int j, int d, int d1, int d2, 
	double *pnode_is2, struct tracekn_s *cur_tr, struct traceknstack_s *dolist, struct ali_s *ali, 
	int traceback, int *ct)
{
  int    i, r;
  int    cur_xr, cur_yr;
  int    y;
  double pxy[5][5], py[5];

  /* calculate pxy[x][y]= P(y|x) = P(x,y) / P(x)
   */
  for (y = 0; y < 5; y++) 
    py[y] = 0.;

  for (y = 0; y < 25; y++)
    py[y/5] += EXP2(is2->ps[y]);

  for (y = 0; y < 5; y++)
    py[y] = LOG2(py[y]);
  
  for (y = 0; y < 25; y++) 
    pxy[y/5][y%5] = is2->ps[y] - py[y/5];


  i = j - d;

  if (traceback && d1 > 1) fprintf(ofp," loop l: ");
  for (r = 1; r < d1 ; r++) {
    cur_xr = seq[i+r];
    cur_yr = DLog2Choose(pxy[cur_xr], 5);
    
    seqX[i+r] = cur_xr;
    seqY[i+r] = cur_yr;
    
    ct[i+r] = -1;
    
    ali->charX[i+r] = (cur_xr < 4)? DNAAlphabet[cur_xr] : '.';
    ali->charY[i+r] = (cur_yr < 4)? DNAAlphabet[cur_yr] : '.';

    if (traceback) fprintf(ofp,"[%d %d] ", cur_xr, cur_yr); 
  }
  
  if (traceback && d2 > 1) fprintf(ofp,"\n loop r: ");
  for (r = 1; r < d2 ; r++) {
    cur_xr = seqX[i+r];
    cur_yr = DLog2Choose(pxy[cur_xr], 5);
    
    seqY[j-r] = cur_yr;
    
    ct[j-r] = -1;
    
    ali->charX[j-r] = (cur_xr < 4)? DNAAlphabet[cur_xr] : '.';
    ali->charY[j-r] = (cur_yr < 4)? DNAAlphabet[cur_yr] : '.';

    if (traceback) fprintf(ofp,"[%d %d] ", cur_xr, cur_yr); 
  }
  if (traceback) fprintf(ofp,"\n");
}

void 
coemitV(FILE *ofp, int *seqX, int *seqY, struct rna_v5 *v, int L, int j, int d, double *pnode_v, 
	double *pnode_is1, double *pnode_is2, struct tracekn_s *cur_tr, struct traceknstack_s *dolist, 
	struct ali_s *ali, int traceback, int *ct)
{
  int    i, k, l, r;
  int    pos_k, pos_l, pos_kl;
  int    cur_xk, cur_yk;
  int    cur_xl, cur_yl;
  int    state;
  int    size, size_i, size_j;
  double pp[625];
  double pn[5];

  i = j - d;
  k = i + (int) (d/2);
  l = k + 1;

  state = cur_tr->type;

  if (d < 5) {
    pn[0] = 0.;
    pn[1] = -BIGFLOAT;
    pn[2] = -BIGFLOAT;
    pn[3] = -BIGFLOAT;
    pn[4] = -BIGFLOAT;
  }

  else if (d < 10){
    pn[0] = pnode_v[dpcS1]  + pnode_is1[dpcLEN(d)];
    pn[1] = pnode_v[dpcS2S] + pnode_is2[dpcLEN(2)];
    pn[2] = pnode_v[dpcS2B] + DLog2Sum(pnode_is2 + DpNodeidx[IS2] + 3, d-2);
    pn[3] = pnode_v[dpcS2I] + DLog2Sum(pnode_is2 + DpNodeidx[IS2] + 4, d-3);
    pn[4] = -BIGFLOAT;
  }

  else {
    pn[0] = (d<MAXRNALOOP)? pnode_v[dpcS1]  + pnode_is1[dpcLEN(d)] : -BIGFLOAT;
    pn[1] = pnode_v[dpcS2S] + pnode_is2[dpcLEN(2)];
    pn[2] = (d<MAXRNALOOP)? pnode_v[dpcS2B] + DLog2Sum(pnode_is2 + DpNodeidx[IS2] + 3, d-2) : -BIGFLOAT;
    pn[3] = (d<MAXRNALOOP)? pnode_v[dpcS2I] + DLog2Sum(pnode_is2 + DpNodeidx[IS2] + 4, d-3) : -BIGFLOAT;
    pn[4] = pnode_v[dpcMV];
  }
  
  for (pos_kl = 0; pos_kl < 625; pos_kl++)
    pp[pos_kl] = v->pp[pos_kl/25][pos_kl%25];

  cur_tr->node = DpNodeidx[state] + DENChoose(pn, NodesPerDp[state]);

  switch (cur_tr->node) {
  case dpcS1:
    PushTraceknstack(dolist, AttachTracekn(cur_tr, i, j, k, k+1, IS1, dpcEIS));
    if (traceback) fprintf(ofp," emit IS1 (%d)  \n", d); 
    break;
    
  case dpcS2S:
    pos_kl = DLog2Choose(pp, 625);
    
    k = i+1;
    l = j-1;

    pos_k = pos_kl / 25;
    pos_l = pos_kl % 25;

    cur_xk = pos_k / 5;
    cur_yk = pos_k % 5;
    cur_xl = pos_l / 5;
    cur_yl = pos_l % 5;
    
    seqX[k] = cur_xk;
    seqY[k] = cur_yk;
    seqX[l] = cur_xl;
    seqY[l] = cur_yl;

    ct[k] = l;
    ct[l] = k;

    ali->charX[k] = (cur_xk < 4)? DNAAlphabet[cur_xk] : '.';
    ali->charY[k] = (cur_yk < 4)? DNAAlphabet[cur_yk] : '.';
   
    ali->charX[l] = (cur_xl < 4)? DNAAlphabet[cur_xl] : '.';
    ali->charY[l] = (cur_yl < 4)? DNAAlphabet[cur_yl] : '.';
   
    d = l - k;
    r = k+(int)(d/2);
    if (d > 0) PushTraceknstack(dolist, AttachTracekn(cur_tr, k, l, r, r+1, V, dpcEV));
    PushTraceknstack(dolist, AttachTracekn(cur_tr, i, j, k, l, IS2, dpcEIS));
    if (traceback) fprintf(ofp," emit IS2S [%d %d] [%d %d]  \n", cur_xk, cur_yk, cur_xl, cur_yl); 
    break;
    
  case dpcS2B:
    size = 3 + DENChoose(pnode_is2 + DpNodeidx[IS2] + 3, d-2);

    pos_kl = DLog2Choose(pp, 625);
   
    k = i + size - 1;
    l = j - 1;

    pos_k = pos_kl / 25;
    pos_l = pos_kl % 25;

    cur_xk = pos_k / 5;
    cur_yk = pos_k % 5;
    cur_xl = pos_l / 5;
    cur_yl = pos_l % 5;
    
    seqX[k] = cur_xk;
    seqY[k] = cur_yk;
    seqX[l] = cur_xl;
    seqY[l] = cur_yl;

    ct[k] = l;
    ct[l] = k;

    ali->charX[k] = (cur_xk < 4)? DNAAlphabet[cur_xk] : '.';
    ali->charY[k] = (cur_yk < 4)? DNAAlphabet[cur_yk] : '.';
   
    ali->charX[l] = (cur_xl < 4)? DNAAlphabet[cur_xl] : '.';
    ali->charY[l] = (cur_yl < 4)? DNAAlphabet[cur_yl] : '.';

    d = l - k;
    r = k + (int)(d/2);
   if (d > 0)  PushTraceknstack(dolist, AttachTracekn(cur_tr, k, l, r, r+1, V, dpcEV));
    PushTraceknstack(dolist, AttachTracekn(cur_tr, i, j, k, l, IS2, dpcEIS));
    if (traceback) fprintf(ofp," emit IS2B(%d %d) [%d %d] [%d %d]  \n", 
			   size, 1, cur_xk, cur_yk, cur_xl, cur_yl); 
    break;
    
  case dpcS2I:
    size = 4 + DENChoose(pnode_is2 + DpNodeidx[IS2] + 4, d-3);

    size_i = (int) (size * sre_random());
    size_j = size - size_i;

    k = i + size_i;
    l = j - size_j;

    pos_kl = DLog2Choose(pp, 625);
    
    pos_k = pos_kl / 25;
    pos_l = pos_kl % 25;

    cur_xk = pos_k / 5;
    cur_yk = pos_k % 5;
    cur_xl = pos_l / 5;
    cur_yl = pos_l % 5;
    
    seqX[k] = cur_xk;
    seqY[k] = cur_yk;
    seqX[l] = cur_xl;
    seqY[l] = cur_yl;

    ct[k] = l;
    ct[l] = k;

    ali->charX[k] = (cur_xk < 4)? DNAAlphabet[cur_xk] : '.';
    ali->charY[k] = (cur_yk < 4)? DNAAlphabet[cur_yk] : '.';
   
    ali->charX[l] = (cur_xl < 4)? DNAAlphabet[cur_xl] : '.';
    ali->charY[l] = (cur_yl < 4)? DNAAlphabet[cur_yl] : '.';

    d = l - k;
    r = k + (int) (d/2);
    if (d > 0) PushTraceknstack(dolist, AttachTracekn(cur_tr, k, l, r, r+1, V, dpcEV));
    PushTraceknstack(dolist, AttachTracekn(cur_tr, i, j, k, l, IS2, dpcEIS));
    if (traceback) fprintf(ofp," emit IS2I(%d %d) [%d %d] [%d %d]  \n", 
			   size_i, size_j, cur_xk, cur_yk, cur_xl, cur_yl); 
    break;
    
  case dpcMV:
    k = i + 1 + (int)(sre_random()*(d-2));   /* choose the breaking point at randon */
    l = i + 1 + (int)((k-i-1)/2);
    r = k + 1 + (int)((j-k-2)/2);

    PushTraceknstack(dolist, AttachTracekn(cur_tr, i+1, k, l, l+1, WB, dpcEW));
    PushTraceknstack(dolist, AttachTracekn(cur_tr, k+1, j-1, r, r+1, WB, dpcEW));
    if (traceback) fprintf(ofp," emit W_b W_b \n"); 
    break;
    
  default:
    Die("invalid matrix assignement in emitV()");
  }
}


void 
emitV(FILE *ofp, int *seq, int *seqX, int *seqY, struct rna_v5 *v, int L, int j, int d, double *pnode_v, 
      double *pnode_is1, double *pnode_is2, struct tracekn_s *cur_tr, struct traceknstack_s *dolist, 
      struct ali_s *ali, int traceback, int *ct)
{
  int    i, k, l, r;
  int    pos_k, pos_l, pos_kl;
  int    cur_xk, cur_yk;
  int    cur_xl, cur_yl;
  int    pos_x, pos_y;
  int    state;
  int    size, size_i, size_j;
  double pp[25][25];
  double pn[5];

  i = j - d;
  k = i + (int) (d/2);
  l = k + 1;

  state = cur_tr->type;

  if (d < 5) {
    pn[0] = 0.;
    pn[1] = -BIGFLOAT;
    pn[2] = -BIGFLOAT;
    pn[3] = -BIGFLOAT;
    pn[4] = -BIGFLOAT;
  }

  else if (d < 10){
    pn[0] = pnode_v[dpcS1]  + pnode_is1[dpcLEN(d)];
    pn[1] = pnode_v[dpcS2S] + pnode_is2[dpcLEN(2)];
    pn[2] = pnode_v[dpcS2B] + DLog2Sum(pnode_is2 + DpNodeidx[IS2] + 3, d-2);
    pn[3] = pnode_v[dpcS2I] + DLog2Sum(pnode_is2 + DpNodeidx[IS2] + 4, d-3);
    pn[4] = -BIGFLOAT;
  }

  else {
    pn[0] = pnode_v[dpcS1]  + pnode_is1[dpcLEN(d)];
    pn[1] = pnode_v[dpcS2S] + pnode_is2[dpcLEN(2)];
    pn[2] = pnode_v[dpcS2B] + DLog2Sum(pnode_is2 + DpNodeidx[IS2] + 3, d-2);
    pn[3] = pnode_v[dpcS2I] + DLog2Sum(pnode_is2 + DpNodeidx[IS2] + 4, d-3);
    pn[4] = pnode_v[dpcMV];
  }
  
  for (pos_kl = 0; pos_kl < 625; pos_kl++) {

    pos_k = pos_kl / 25;
    pos_l = pos_kl % 25;

    cur_xk = pos_k / 5;
    cur_yk = pos_k % 5;
    cur_xl = pos_l / 5;
    cur_yl = pos_l % 5;

    pos_x = cur_xk * 5 + cur_xl;
    pos_y = cur_yk * 5 + cur_yl;

    pp[pos_x][pos_y] = v->pp[pos_k][pos_l];
  }

  cur_tr->node = DpNodeidx[state] + DENChoose(pn, NodesPerDp[state]);

  switch (cur_tr->node) {
  case dpcS1:
    PushTraceknstack(dolist, AttachTracekn(cur_tr, i, j, k, k+1, IS1, dpcEIS));
    if (traceback) fprintf(ofp," emit IS1 (%d)  \n", d); 
    break;
    
  case dpcS2S:
    k = i+1;
    l = j-1;

    cur_xk = seq[k];
    cur_xl = seq[l];

    pos_x = cur_xk * 5 + cur_xl;
    pos_y = DLog2Choose(pp[pos_x], 25);
    
    cur_yk = pos_y % 5;
    cur_yl = pos_y % 5;
    
    seqX[k] = cur_xk;
    seqY[k] = cur_yk;
    seqX[l] = cur_xl;
    seqY[l] = cur_yl;

    ct[k] = l;
    ct[l] = k;

    ali->charX[k] = (cur_xk < 4)? DNAAlphabet[cur_xk] : '.';
    ali->charY[k] = (cur_yk < 4)? DNAAlphabet[cur_yk] : '.';
   
    ali->charX[l] = (cur_xl < 4)? DNAAlphabet[cur_xl] : '.';
    ali->charY[l] = (cur_yl < 4)? DNAAlphabet[cur_yl] : '.';
   
    d = l - k;
    r = k+(int)(d/2);
    if (d > 0) PushTraceknstack(dolist, AttachTracekn(cur_tr, k, l, r, r+1, V, dpcEV));
    PushTraceknstack(dolist, AttachTracekn(cur_tr, i, j, k, l, IS2, dpcEIS));
    if (traceback) fprintf(ofp," emit IS2S [%d %d] [%d %d]  \n", cur_xk, cur_yk, cur_xl, cur_yl); 
    break;
    
  case dpcS2B:
    size = 3 + DENChoose(pnode_is2 + DpNodeidx[IS2] + 3, d-2);

    k = i + size - 1;
    l = j - 1;

    cur_xk = seq[k];
    cur_xl = seq[l];

    pos_x = cur_xk * 5 + cur_xl;
    pos_y = DLog2Choose(pp[pos_x], 25);
    
    cur_yk = pos_y % 5;
    cur_yl = pos_y % 5;
    
    seqX[k] = cur_xk;
    seqY[k] = cur_yk;
    seqX[l] = cur_xl;
    seqY[l] = cur_yl;

    ct[k] = l;
    ct[l] = k;

    ali->charX[k] = (cur_xk < 4)? DNAAlphabet[cur_xk] : '.';
    ali->charY[k] = (cur_yk < 4)? DNAAlphabet[cur_yk] : '.';
   
    ali->charX[l] = (cur_xl < 4)? DNAAlphabet[cur_xl] : '.';
    ali->charY[l] = (cur_yl < 4)? DNAAlphabet[cur_yl] : '.';

    d = l - k;
    r = k + (int)(d/2);
   if (d > 0)  PushTraceknstack(dolist, AttachTracekn(cur_tr, k, l, r, r+1, V, dpcEV));
    PushTraceknstack(dolist, AttachTracekn(cur_tr, i, j, k, l, IS2, dpcEIS));
    if (traceback) fprintf(ofp," emit IS2B(%d %d) [%d %d] [%d %d]  \n", 
			   size, 1, cur_xk, cur_yk, cur_xl, cur_yl); 
    break;
    
  case dpcS2I:
    size = 4 + DENChoose(pnode_is2 + DpNodeidx[IS2] + 4, d-3);

    size_i = (int) (size * sre_random());
    size_j = size - size_i;

    k = i + size_i;
    l = j - size_j;

    cur_xk = seq[k];
    cur_xl = seq[l];

    pos_x = cur_xk * 5 + cur_xl;
    pos_y = DLog2Choose(pp[pos_x], 25);
    
    cur_yk = pos_y % 5;
    cur_yl = pos_y % 5;
    
    seqX[k] = cur_xk;
    seqY[k] = cur_yk;
    seqX[l] = cur_xl;
    seqY[l] = cur_yl;

    ct[k] = l;
    ct[l] = k;

    ali->charX[k] = (cur_xk < 4)? DNAAlphabet[cur_xk] : '.';
    ali->charY[k] = (cur_yk < 4)? DNAAlphabet[cur_yk] : '.';
   
    ali->charX[l] = (cur_xl < 4)? DNAAlphabet[cur_xl] : '.';
    ali->charY[l] = (cur_yl < 4)? DNAAlphabet[cur_yl] : '.';

    d = l - k;
    r = k + (int) (d/2);
    if (d > 0) PushTraceknstack(dolist, AttachTracekn(cur_tr, k, l, r, r+1, V, dpcEV));
    PushTraceknstack(dolist, AttachTracekn(cur_tr, i, j, k, l, IS2, dpcEIS));
    if (traceback) fprintf(ofp," emit IS2I(%d %d) [%d %d] [%d %d]  \n", 
			   size_i, size_j, cur_xk, cur_yk, cur_xl, cur_yl); 
    break;
    
  case dpcMV:
    k = i + 1 + (int)(sre_random()*(d-2));   /* choose the breaking point at randon */
    l = i + 1 + (int)((k-i-1)/2);
    r = k + 1 + (int)((j-k-2)/2);

    PushTraceknstack(dolist, AttachTracekn(cur_tr, i+1, k, l, l+1, WB, dpcEW));
    PushTraceknstack(dolist, AttachTracekn(cur_tr, k+1, j-1, r, r+1, WB, dpcEW));
    if (traceback) fprintf(ofp," emit W_b W_b \n"); 
    break;
    
  default:
    Die("invalid matrix assignement in emitV()");
  }
}


void
coemitW(FILE *ofp, int *seqX, int *seqY, struct rna_w5 *w, int L, int j, int d, double *pnode_w,
	struct tracekn_s *cur_tr, struct traceknstack_s *dolist, struct ali_s *ali, 
	int traceback, int *ct)
{
  int    i, k, l, r;
  int    pos_i, pos_j, pos_ij;
  int    cur_xi, cur_yi;
  int    cur_xj, cur_yj;
  int    state;
  double pp[625];

  i = j - d;
  k = i + (int) (d/2);
  l = k + 1;

  state = cur_tr->type;

  if (d < 5) {
    PushTraceknstack(dolist, AttachTracekn(cur_tr, i, j, k, k+1, V, dpcEV));
    if (traceback) fprintf(ofp," emit P [%d %d] [%d %d] \n", seqX[i], seqY[i], seqX[j], seqY[j]); 
    return;
  }
  
  for (pos_ij = 0; pos_ij < 625; pos_ij++)
    pp[pos_ij] = w->pp[pos_ij/25][pos_ij%25];

  cur_tr->node = DpNodeidx[state] + DLog2Choose(pnode_w + DpNodeidx[state], NodesPerDp[state]);

  switch (cur_tr->node) {
  case dpcL:
    pos_i = DLog2Choose(w->pl, 25);

    cur_xi = pos_i / 5;
    cur_yi = pos_i % 5;

    seqX[i] = cur_xi;
    seqY[i] = cur_yi;

    ct[i] = -1;

    ali->charX[i] = (cur_xi < 4)? DNAAlphabet[cur_xi] : '.';
    ali->charY[i] = (cur_yi < 4)? DNAAlphabet[cur_yi] : '.';
    
    k = i + 1 + (int) ((d-1)/2);
    PushTraceknstack(dolist, AttachTracekn(cur_tr, i+1, j, k, k+1, state, dpcEW));
    if (traceback) fprintf(ofp," emit L [%d %d] \n", cur_xi, cur_yi); 
    break;
    
  case dpcR:
    pos_j = DLog2Choose(w->pr, 25);
    
    cur_xj = pos_j / 5;
    cur_yj = pos_j % 5;
    
    seqX[j] = cur_xj;
    seqY[j] = cur_yj;

    ct[j] = -1;

    ali->charX[j] = (cur_xj < 4)? DNAAlphabet[cur_xj] : '.';
    ali->charY[j] = (cur_yj < 4)? DNAAlphabet[cur_yj] : '.';
   
    k = i + (int) ((d-1)/2);
    PushTraceknstack(dolist, AttachTracekn(cur_tr, i, j-1, k, k+1, state, dpcEW));
    if (traceback) fprintf(ofp," emit R [%d %d] \n", cur_xj, cur_yj); 
    break;
    
  case dpcP:
    pos_ij = DLog2Choose(pp, 625);
    
    pos_i = pos_ij / 25;
    pos_j = pos_ij % 25;

    cur_xi = pos_i / 5;
    cur_yi = pos_i % 5;
    cur_xj = pos_j / 5;
    cur_yj = pos_j % 5;
    
    seqX[i] = cur_xi;
    seqY[i] = cur_yi;
    seqX[j] = cur_xj;
    seqY[j] = cur_yj;

    ct[i] = j;
    ct[j] = i;

    ali->charX[i] = (cur_xi < 4)? DNAAlphabet[cur_xi] : '.';
    ali->charY[i] = (cur_yi < 4)? DNAAlphabet[cur_yi] : '.';
   
    ali->charX[j] = (cur_xj < 4)? DNAAlphabet[cur_xj] : '.';
    ali->charY[j] = (cur_yj < 4)? DNAAlphabet[cur_yj] : '.';
   

    if (d > 1) PushTraceknstack(dolist, AttachTracekn(cur_tr, i, j, k, k+1, V, dpcEV));
    if (traceback) fprintf(ofp," emit P [%d %d] [%d %d] \n", cur_xi, cur_yi, cur_xj, cur_yj); 
    break;
    
  case dpcBW:
    k = i + (int)(sre_random()*d);   /* choose the breaking point at randon */
    l = i + (int)((k-i)/2);
    r = k + 1 + (int)((j-k-1)/2);

    PushTraceknstack(dolist, AttachTracekn(cur_tr, i, k, l, l+1, state, dpcEW));
    PushTraceknstack(dolist, AttachTracekn(cur_tr, k+1, j, r, r+1, state, dpcEW));
    if (traceback) fprintf(ofp," emit WW\n"); 
    break;
    
  default:
    Die("invalid matrix assignement in emitW()");
  }
}

void
emitW(FILE *ofp, int *seq, int *seqX, int *seqY, struct rna_w5 *w, int L, int j, int d, double *pnode_w,
      struct tracekn_s *cur_tr, struct traceknstack_s *dolist, struct ali_s *ali, 
      int traceback, int *ct)
{
  int    i, k, l, r;
  int    pos_i, pos_j, pos_ij;
  int    cur_xi, cur_yi;
  int    cur_xj, cur_yj;
  int    pos_x, pos_y;
  int    state;
  double pp[25][25];
  double pl[5][5], pr[5][5];

  i = j - d;
  k = i + (int) (d/2);
  l = k + 1;

  state = cur_tr->type;

  if (d < 5) {
    PushTraceknstack(dolist, AttachTracekn(cur_tr, i, j, k, k+1, V, dpcEV));
    if (traceback) fprintf(ofp," emit P [%d %d] [%d %d] \n", seqX[i], seqY[i], seqX[j], seqY[j]); 
    return;
  }
  
  for (pos_i = 0; pos_i < 25; pos_i++) 
    pl[pos_i/5][pos_i%5] = w->pl[pos_i];
  
  for (pos_j = 0; pos_j < 25; pos_j++) 
    pr[pos_j/5][pos_j%5] = w->pr[pos_j];
  
  for (pos_ij = 0; pos_ij < 625; pos_ij++) {
    pos_i = pos_ij / 25;
    pos_j = pos_ij % 25;
    
    cur_xi = pos_i / 5;
    cur_yi = pos_i % 5;
    cur_xj = pos_j / 5;
    cur_yj = pos_j % 5;
    
    pos_x = cur_xi * 5 + cur_xj;
    pos_y = cur_yi * 5 + cur_yj;
    
    pp[pos_x][pos_y] = w->pp[pos_i][pos_j];
  }

  cur_tr->node = DpNodeidx[state] + DLog2Choose(pnode_w + DpNodeidx[state], NodesPerDp[state]);
  
  switch (cur_tr->node) {
  case dpcL:
    cur_xi = seq[i];
    cur_yi = DLog2Choose(pl[cur_xi], 5);

    seqX[i] = cur_xi;
    seqY[i] = cur_yi;

    ct[i] = -1;

    ali->charX[i] = (cur_xi < 4)? DNAAlphabet[cur_xi] : '.';
    ali->charY[i] = (cur_yi < 4)? DNAAlphabet[cur_yi] : '.';
    
    k = i + 1 + (int) ((d-1)/2);
    PushTraceknstack(dolist, AttachTracekn(cur_tr, i+1, j, k, k+1, state, dpcEW));
    if (traceback) fprintf(ofp," emit L [%d %d] \n", cur_xi, cur_yi); 
    break;
    
  case dpcR:
    cur_xj = seq[j];
    cur_yj = DLog2Choose(pr[cur_xj], 5);

    seqX[j] = cur_xj;
    seqY[j] = cur_yj;

    ct[j] = -1;

    ali->charX[j] = (cur_xj < 4)? DNAAlphabet[cur_xj] : '.';
    ali->charY[j] = (cur_yj < 4)? DNAAlphabet[cur_yj] : '.';
   
    k = i + (int) ((d-1)/2);
    PushTraceknstack(dolist, AttachTracekn(cur_tr, i, j-1, k, k+1, state, dpcEW));
    if (traceback) fprintf(ofp," emit R [%d %d] \n", cur_xj, cur_yj); 
    break;
    
  case dpcP:
    cur_xi = seq[i];
    cur_xj = seq[j];
    
    pos_x = cur_xi * 5 + cur_xj;
    pos_y = DLog2Choose(pp[pos_x], 25);

    cur_yi = pos_y / 5;
    cur_yj = pos_y % 5;
    
    seqX[i] = cur_xi;
    seqY[i] = cur_yi;
    seqX[j] = cur_xj;
    seqY[j] = cur_yj;

    ct[i] = j;
    ct[j] = i;

    ali->charX[i] = (cur_xi < 4)? DNAAlphabet[cur_xi] : '.';
    ali->charY[i] = (cur_yi < 4)? DNAAlphabet[cur_yi] : '.';
   
    ali->charX[j] = (cur_xj < 4)? DNAAlphabet[cur_xj] : '.';
    ali->charY[j] = (cur_yj < 4)? DNAAlphabet[cur_yj] : '.';
   

    if (d > 1) PushTraceknstack(dolist, AttachTracekn(cur_tr, i, j, k, k+1, V, dpcEV));
    if (traceback) fprintf(ofp," emit P [%d %d] [%d %d] \n", cur_xi, cur_yi, cur_xj, cur_yj); 
    break;
    
  case dpcBW:
    k = i + (int)(sre_random()*d);   /* choose the breaking point at randon */
    l = i + (int)((k-i)/2);
    r = k + 1 + (int)((j-k-1)/2);

    PushTraceknstack(dolist, AttachTracekn(cur_tr, i, k, l, l+1, state, dpcEW));
    PushTraceknstack(dolist, AttachTracekn(cur_tr, k+1, j, r, r+1, state, dpcEW));
    if (traceback) fprintf(ofp," emit WW\n"); 
    break;
    
  default:
    Die("invalid matrix assignement in emitW()");
  }
}




