/************************************************************
 * QRNA - Comparative analysis of biological sequences 
 *        with pair hidden Markov models and pair stochastic context-free grammars
 * Copyright (C) 2000-Howard Hughes Medical Institute
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* rnamodel.c
 *
 * E. Rivas [St. Louis]
 * 
 * 9 april 1999.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

#include "funcs.h"
#include "globals.h"
#include "squid.h"
#include "structs.h"

#ifdef MEMDEBUG
#include "dbmalloc.h"
#endif


/* Function: AllocNUSModel()
 * Date:     ER, Wed Jun  9 09:59:08 CDT 1999 [St. Louis]
 *
 * Purpose:  Allocates memory for the transition and emission probs of the nusmodel
 *
 * Args:     nusmodel - nusmodel structure
 *
 * Returns:  void
 *           allocates nus->t[], nus->ROB, nus->ROJ, nus->ROE, nus->wx 
 *           (which must be free'd by caller).
 */
struct nusmodel_s *
AllocNUSModel(void)
{
  struct nusmodel_s *nus;  /* nusmodel structure           */

  nus = (struct nusmodel_s *)  MallocOrDie (sizeof(struct nusmodel_s));
  
  return nus;
}

/* Function: AllocPI2Model()
 * Date:     ER,  Tue Jun  8 14:15:46 CDT 1999 [St. Louis]
 *
 * Purpose:  Allocates memory for the transition and emission probs of the rnamodel
 *
 * Args:     rnamodel - rnamodel structure
 *
 * Returns:  void
 *           allocates rna->t[], rna->ROB, rna->ROJ, rna->ROE, 
 *                     rna->vx, rna->wx, rna->wbx, rna->is1, rna->is2, 
 *           (which must be free'd by caller).
 */
struct pi2model_s *
AllocPI2Model(void)
{
  struct pi2model_s *pi2;   /* rnamodel structure           */

  pi2        = (struct pi2model_s *)  MallocOrDie (sizeof(struct pi2model_s));
  pi2->v     = (struct rna_v5     *)  MallocOrDie (sizeof(struct rna_v5));
  pi2->w     = (struct rna_w5     *)  MallocOrDie (sizeof(struct rna_w5));
  pi2->wb    = (struct rna_w5     *)  MallocOrDie (sizeof(struct rna_w5));
  pi2->is1   = (struct rna_loop5  *)  MallocOrDie (sizeof(struct rna_loop5));
  pi2->is2   = (struct rna_loop5  *)  MallocOrDie (sizeof(struct rna_loop5));

  pi2->Rloop = AllocOTHModel();
  
  return pi2;
}

/* Function: AllocRNAModel()
 * Date:     ER,  Tue Jun  8 14:15:46 CDT 1999 [St. Louis]
 *
 * Purpose:  Allocates memory for the transition and emission probs of the rnamodel
 *
 * Args:     rnamodel - rnamodel structure
 *
 * Returns:  void
 *           allocates rna->t[], rna->ROB, rna->ROJ, rna->ROE, 
 *                     rna->vx, rna->wx, rna->wbx, rna->is1, rna->is2, 
 *           (which must be free'd by caller).
 */
struct rnamodel_s *
AllocRNAModel(void)
{
  struct rnamodel_s *rna;   /* rnamodel structure           */

  rna      = (struct rnamodel_s *)  MallocOrDie (sizeof(struct rnamodel_s));
  rna->t   = (double *)             MallocOrDie (sizeof(double) * RTRANS);

  rna->ROB = AllocOTHModel();
  rna->ROJ = AllocOTHModel();
  rna->ROE = AllocOTHModel();

  rna->nus = AllocNUSModel();
  rna->pi2 = AllocPI2Model();
  
  PatternRNAModel(rna);

  return rna;
}


/* Function: CheckRNAProbs()
 * Date:     ER, Thu Jun 10 11:59:42 CDT 1999 [St. Louis]
 *
 * Purpose:  Verify that transition prob's of a RNAmodel add up to one
 *
 * Args:     rna - the structure for an pi2model
 *
 * Returns:  void. 
 */
void
CheckRNAProbs(struct rnamodel_s *rna)
{
  int    idx;               /* index for transition prob's      +*/
  int    st;                /* state we are at                  +*/
  double sum;

  /* check transition prob's add up to one
   */
  for (st = 0; st < RSTATES; st++) {
    sum = 0.0; 
    for (idx = IdxTransRNA[st]; idx < IdxTransRNA[st]+TransPerRNAState[st]; idx++)
      sum += rna->t[idx];
    
    if (sum > 1.01 || sum < 0.99) 
      Die("transition prob's sum for state %d(%s) is %f \n", st, rstNAME[st], sum);
  }
}


/* Function: ConstructRNAModel()
 * Date:     ER, Tue Jun  8 13:30:43 CDT 1999 [St. Louis]
 *
 * Purpose:  Build the RNA model from the null model under
 *           certain simplifying assumptions; namely, that
 *           P(x1 y1 x2 y2) = P(x1 y1) P (x2 y2) P(x1 x2) / (P(x1) P(x2))
 *
 * Args:     pammodel[64][64]
 *           cfg      -- the file with the SCFG model
 *           rnamodel -- pointer to alloc'ed rnamodel
 *
 * Returns:  void
 *           fills in rnamodel
 */
void
ConstructRNAModel(double *mutpxy, double *mut5pxy, double ***cfg_node, 
		  double *pair5, double tfactor, struct rnamodel_s *rna, struct nullmodel_s *null, 
		  double *freqX, double *freqY, int changefreq, int pedantic)
{
  double             tr;                       /* transition prob's                        +*/
  double            *pair5prob;                /* base-pair probabilities at t=0 (no gaps) +*/
  double            *pair5prob_zero;           /* base-pair probabilities at t=0 (no gaps) +*/
  double            *pair5condXY;              /* base-pair probabilities                  +*/
  double            *pair5condYX;              /* base-pair probabilities                  +*/
  double            *pmarX;                    /* marginalization of the pair5probs        +*/
  double            *pmarY;                    /* marginalization of the pair5probs        +*/
  double            *px;                       /* single nt probabilities                  +*/
  double            *py;                       /* single nt probabilities                  +*/
  int                idx;                      /* index for transition prob's              +*/
  int                i,j;                      /* symbols for emission prob's              +*/
  int                xi,xj,yi,yj;
  int                len;                      /* length of a loop (< MAXRNALOOP)          +*/
  double             pp;
  double             sum;

  px = (double *) MallocOrDie (sizeof(double)*5);
  py = (double *) MallocOrDie (sizeof(double)*5);

  for (i = 0; i < 5; i++) { px[i] = 0.0; py[i] = 0.0; }

  ConstructOTHModel(mutpxy, ROBparam,   tfactor, rna->ROB,        pedantic);
  ConstructOTHModel(mutpxy, ROJparam,   tfactor, rna->ROJ,        pedantic);
  ConstructOTHModel(mutpxy, ROEparam,   tfactor, rna->ROE,        pedantic);
  ConstructOTHModel(mutpxy, Rloopparam, tfactor, rna->pi2->Rloop, pedantic);

 /* Calculate transition prob's: rna.t[]
   */
  for (idx = 0; idx < RTRANS; idx++) {
   switch (idx) {
    case TROBRNA: tr = RNAparam.psi;       break;
    case TROBROE: tr = 1.0 - RNAparam.psi; break;
    case TRNAROJ: tr = 1.0 - RNAparam.eta; break;
    case TRNAROE: tr = RNAparam.eta;       break;
    case TROJRNA: tr = 1.0;                break;
    case TROEROE: tr = 1.0;                break;
    default:  Die("transition (%d) does not exits", idx);
    }

   if (tr > 1.0 || tr < 0.0) Die ("transition %s in RNA model is wrong", rtrNAME[idx]);
   else rna->t[idx] = tr;
  }
  
 /* single nt probs by marginalization of mutation probs
   */
  for (i = 0; i < 5; i++) {
    px[i] = 0.;
    py[i] = 0.;
  }

  for (i = 0; i < 4; i++) 
    for (j = 0; j < 5; j++) {
      px[i] += mut5pxy[idx5(i,j)];
      py[i] += mut5pxy[idx5(j,i)];
    }
  px[4] = GAPP;
  py[4] = GAPP;

                                 /* normalize */
  sum = 0.;
  for (i = 0; i < 5; i++) 
    sum += px[i];
  for (i = 0; i < 5; i++) 
    px[i] /= sum;
  sum = 0.;
  for (i = 0; i < 5; i++) 
    sum += py[i];
  for (i = 0; i < 5; i++) 
    py[i] /= sum;

  /* 
   */
  pair5prob      = (double *) MallocOrDie (sizeof(double) * 25);
  pair5prob_zero = (double *) MallocOrDie (sizeof(double) * 25);

  for (i = 0; i < 5; i++) 
    for (j = 0; j < 5; j++) {
      pair5prob[idx5(i,j)]      = pair5[idx5(i,j)];
      pair5prob_zero[idx5(i,j)] = pair5[idx5(i,j)];
    }

  /* pair5prob_zero are defined as Conditional probabilites.
   *
   * I just assume that at time zero there are not gaps, 
   * so I give a 1.0 to gap-to-gap pair, and 0.0 to nt-to-gap pair.
   */
  for (i = 0; i < 4; i++) {
    sum = 0.;
    for (j = 0; j < 4; j++) 
      sum += pair5prob_zero[idx5(i,j)];
    for (j = 0; j < 4; j++) 
      pair5prob_zero[idx5(i,j)] /= sum;
  }
  pair5prob_zero[idx5(4,4)] = 1.0;
  
  /* To change the overal base composition, [we do this before dealing with gaps]
   *
   * calculate:         P(i|j) = P(i,j) / P(j)
   *
   * then use:         ^P(j) = f(j)
   *
   * to recalculate:   ^P(i,j) = P(i|j) * ^P(j)
   */
  if (changefreq) {
    pair5condXY = (double *) MallocOrDie (sizeof(double) * 25);
    pair5condYX = (double *) MallocOrDie (sizeof(double) * 25);
    pmarX       = (double *) MallocOrDie (sizeof(double) * 5);
    pmarY       = (double *) MallocOrDie (sizeof(double) * 5);
    
    for (i = 0; i < 5; i++) { 
      pmarX[i] = 0.0;
      pmarY[i] = 0.0;
    }
    
    for (i = 0; i < 5; i++) 
      for (j = 0; j < 5; j++) {
	pmarX[i] += pair5prob[idx5(i,j)];
	pmarY[j] += pair5prob[idx5(i,j)];
      }
    
    /* these loops go just to 3, because we are not dealing with gaps just yet */
    for (i = 0; i < 4; i++)
      for (j = 0; j < 4; j++) {
	pair5condXY[idx5(i,j)] = pair5prob[idx5(i,j)] / pmarY[j];
	pair5condYX[idx5(j,i)] = pair5prob[idx5(i,j)] / pmarX[i];
      }
    
    for (i = 0; i < 4; i++)
      for (j = 0; j < 4; j++)
	pair5prob[idx5(i,j)] = pair5condXY[idx5(i,j)]*freqY[j]/2.0 + pair5condYX[idx5(j,i)]*freqX[i]/2.0;
    
    free(pmarX);
    free(pmarY);
    free(pair5condXY);
    free(pair5condYX);
  }

  /* OK, NT-2-GAP AND GAP-2-GAP Watson-Crick-Pair-Probabilities cannot be justified
   *
   * This is how I deal with them:
   */
                              /* add pair-to-gap */
  for (i = 0; i < 4; i++) {
    pair5prob[idx5(i,4)] = px[i] * py[4];
    pair5prob[idx5(4,i)] = px[4] * py[i];
  }
  pair5prob[idx5(4,4)] = 1.02*px[4]*py[4];/* this guarantees a decent instant evolutionary matrix.
					   * otherwise I get probabilities that go slighly negative
					   * for t ~ 0.4.
					   */

                                 /* normalize */
  sum = 0.;
  for (i = 0; i < 25; i++) 
    sum += pair5prob[i];

  for (i = 0; i < 25; i++) 
    pair5prob[i] /= sum;
  CheckSingleProb(pair5prob, 25); 

  if (FALSE) {
    fprintf(stdout, "Ppair5(i,j, %f) Joint-probabilities\n", tfactor);
    PrintProbs(stdout, pair5prob,5);
    fprintf(stdout, "Ppair5(i,j, 0) Cond-probabilities\n");
    PrintProbs(stdout, pair5prob_zero,5);
  }
  

  /* Convert 
   *         joint probabilities pair5prob(ab, t_o) 
   *   to 
   *         joint probabilities pair5prob(ab, tfactor*t_o)
   */
  Joint2Joint(pair5prob, pair5prob_zero, 5, tfactor, pedantic, FALSE);
  if (FALSE) PrintMut5Probs(pair5prob, null->xem);

  /* PI2 model
   */
  rna->pi2->v->t1  = 0.;		
  rna->pi2->v->t2s = 0.;		
  rna->pi2->v->t2b = 0.;		
  rna->pi2->v->t2i = 0.;		
  rna->pi2->v->t3  = 0.;

  for (idx = 0; idx < Idx[V]; idx++) {
    rna->pi2->v->t1  += EXP2(cfg_node[V][idx][dpcS1  - DpNodeidx[V]]);		

    rna->pi2->v->t2s += EXP2(cfg_node[V][idx][dpcS2S - DpNodeidx[V]]);		
    rna->pi2->v->t2b += EXP2(cfg_node[V][idx][dpcS2B - DpNodeidx[V]]);		
    rna->pi2->v->t2i += EXP2(cfg_node[V][idx][dpcS2I - DpNodeidx[V]]);		
    rna->pi2->v->t3  += EXP2(cfg_node[V][idx][dpcMV  - DpNodeidx[V]]);
  }	
  rna->pi2->v->t1  /= Idx[V];
  rna->pi2->v->t2s /= Idx[V];
  rna->pi2->v->t2b /= Idx[V];
  rna->pi2->v->t2i /= Idx[V];
  rna->pi2->v->t3  /= Idx[V];

  sum = rna->pi2->v->t1+rna->pi2->v->t2s+rna->pi2->v->t2b+rna->pi2->v->t2i+rna->pi2->v->t3;
  if (sum > 1.01 || sum < 0.99) 
    Die ("pi2->v transtitions don't add up to one (sum_v = %f)\n", sum);

  /* tie L and R emission
  */
  rna->pi2->w->tl = EXP2(cfg_node[W][0][dpcL   - DpNodeidx[W]]);		
  rna->pi2->w->tr = EXP2(cfg_node[W][0][dpcR   - DpNodeidx[W]]);
  rna->pi2->w->tl = 0.5 * (rna->pi2->w->tl + rna->pi2->w->tr);
  rna->pi2->w->tr = rna->pi2->w->tl;
  rna->pi2->w->tv = EXP2(cfg_node[W][0][dpcP   - DpNodeidx[W]]);
  rna->pi2->w->tw = EXP2(cfg_node[W][0][dpcBW  - DpNodeidx[W]]);

  rna->pi2->w->tl = 0.1;		
  rna->pi2->w->tr = 0.1;
  rna->pi2->w->tv = 0.75;
  rna->pi2->w->tw = 0.05;

 sum = rna->pi2->w->tl + rna->pi2->w->tr + rna->pi2->w->tv + rna->pi2->w->tw;
  if (sum > 1.1 || sum < 0.99) 
    Die ("pi2->w transtitions don't add up to one (sum_w = %f)\n", sum);

 
  rna->pi2->wb->tl = EXP2(cfg_node[WB][0][dpcL  - DpNodeidx[WB]]);		
  rna->pi2->wb->tr = EXP2(cfg_node[WB][0][dpcR  - DpNodeidx[WB]]);
  rna->pi2->wb->tl = 0.5 * (rna->pi2->wb->tl+rna->pi2->wb->tr);
  rna->pi2->wb->tr = rna->pi2->wb->tl;
  rna->pi2->wb->tv = EXP2(cfg_node[WB][0][dpcP  - DpNodeidx[WB]]);
  rna->pi2->wb->tw = EXP2(cfg_node[WB][0][dpcBW - DpNodeidx[WB]]);

  rna->pi2->wb->tl = 0.1;		
  rna->pi2->wb->tr = 0.1;
  rna->pi2->wb->tv = 0.75;
  rna->pi2->wb->tw = 0.05;

  sum = rna->pi2->wb->tl + rna->pi2->wb->tr + rna->pi2->wb->tv + rna->pi2->wb->tw;
  if (sum > 1.1 || sum < 0.99) 
    Die ("pi2->wb transtitions don't add up to one (sum_wb = %f)\n", sum);

  
  for (len = 0; len < MAXRNALOOP; len++) {
    rna->pi2->is1->tn[len] = EXP2(cfg_node[IS1][0][dpcLEN(len)-DpNodeidx[IS1]]);
    rna->pi2->is2->tn[len] = EXP2(cfg_node[IS2][0][dpcLEN(len)-DpNodeidx[IS2]]);
  }

				/* copy singlets */
  for (i = 0; i < 25; i++) {
    rna->pi2->w->pl[i]   = mut5pxy[i];
    rna->pi2->w->pr[i]   = mut5pxy[i];
    rna->pi2->wb->pl[i]  = mut5pxy[i];
    rna->pi2->wb->pr[i]  = mut5pxy[i];
    rna->pi2->is1->ps[i] = mut5pxy[i];
    rna->pi2->is2->ps[i] = mut5pxy[i];
  }
  
				/* calculate pairs */
  sum = 0.0;
  for (xi = 0; xi < 5; xi++)
    for (yi = 0; yi < 5; yi++)
      for (xj = 0; xj < 5; xj++)
	for (yj = 0; yj < 5; yj++)
	  {
	     pp = pair5prob[idx5(xi,xj)] * pair5prob[idx5(yi,yj)]  *
	       ( (mut5pxy[idx5(xi,yi)] / (px[xi] * py[yi])) + 
		 (mut5pxy[idx5(xj,yj)] / (px[xj] * py[yj]))   ) * 0.5;
	     
	     rna->pi2->v->pp [idx5(xi,yi)][idx5(xj,yj)] = pp;
	     rna->pi2->w->pp [idx5(xi,yi)][idx5(xj,yj)] = pp;
	     rna->pi2->wb->pp[idx5(xi,yi)][idx5(xj,yj)] = pp;
	     sum += pp;
	  }
				/* renormalize to clean up a bit */
  for (i = 0; i < 25; i++)
    for (j = 0; j < 25; j++) {
	  rna->pi2->v->pp [i][j] /= sum;  
	  rna->pi2->w->pp [i][j] /= sum;  
	  rna->pi2->wb->pp[i][j] /= sum;
    }  

  /* NUS model
   */
  rna->nus->tl = 0.1; 
  rna->nus->tr = 0.1; 
  rna->nus->tv = 0.7;
  rna->nus->tw = 0.095;
  rna->nus->te = 0.005;

  sum = rna->nus->tl + rna->nus->tr + rna->nus->tv + rna->nus->tw + rna->nus->te;
  if (sum > 1.02 || sum < 0.998) 
    Die ("nus->w transtitions don't add up to one (sum_w = %f)\n", sum);

  for (i = 0; i < 25; i++) {
    rna->nus->p[i] = rna->pi2->wb->pl[i];

    for (j = 0; j < 25; j++)
      rna->nus->pp[i][j] = rna->pi2->v->pp[i][j];
  }

  /* check prob's add up to one
   */
  CheckRNAProbs(rna);

  /* convert to Log2 form
   */
  RNAToLog2(rna, null);

  free(px);
  free(py);
  free(pair5prob);
  free(pair5prob_zero);
}

void
FreeNUSModel(struct nusmodel_s *nus)
{
  free(nus);
}

void
FreePI2Model(struct pi2model_s *pi2)
{
  FreeOTHModel(pi2->Rloop);

  free(pi2->v);
  free(pi2->w);
  free(pi2->wb);

  free(pi2->is1);
  free(pi2->is2);

  free(pi2);
}

void
FreeRNAModel(struct rnamodel_s *rna)
{
  FreeOTHModel(rna->ROB);
  FreeOTHModel(rna->ROJ);
  FreeOTHModel(rna->ROE);

  FreeNUSModel(rna->nus);
  FreePI2Model(rna->pi2);

  free(rna->t);
  free(rna);
}

/* Function: NUSToLog2()
 * Date:     ER, Wed Jun  9 17:48:13 CDT 1999 [St. Louis]
 *
 * Purpose:  Converts transition and emission prob's of a nusmodel to log2 form
 *
 * Args:     nus - the structure for an nusmodel
 *
 * Returns:  void. 
 */
void
NUSToLog2(struct nusmodel_s *nus)
{
  int x1,x2;       /* symbols for emission prob's      +*/

  /* transition prob's 
   */
  nus->tl = LOG2(nus->tl);		
  nus->tr = LOG2(nus->tr);
  nus->tv = LOG2(nus->tv);
  nus->tw = LOG2(nus->tw);
  nus->te = LOG2(nus->te);

  /* emission prob's 
   */
  for (x1 = 0; x1 < 25; x1++) {
    nus->p[x1] = LOG2(nus->p[x1]);
    
    for (x2 = 0; x2 < 25; x2++)
      nus->pp[x1][x2] = LOG2(nus->pp[x1][x2]);
  }
}


/* Function: NUSLog2ToOdds()
 * Date:     ER, Wed Jun  9 17:46:37 CDT 1999 [St. Louis]
 *
 * Purpose:  Converts transition and emission prob's of a nusmodel
 *           from log2 to log2odds form
 *
 * Args:     rna - the structure for an rnamodel
 *
 * Returns:  void. 
 */
void
NUSLog2ToOdds(struct nusmodel_s *nus, struct nullmodel_s *null)
{
  int x1,x2;
  double nullx1, nullx2;

  /* transition prob's 
   */
  nus->tl -= 2.0 * null->meta;
  nus->tr -= 2.0 * null->meta;
  nus->tv -= 4.0 * null->meta;

  /* emission prob's 
   */
  for (x1 = 0; x1 < 25; x1++) {
    nullx1  = (x1/5<4)? null->xem[x1/5] : -null->meta;
    nullx1 += (x1%5<4)? null->yem[x1%5] : -null->meta;

    nus->p[x1] -= nullx1;

   for (x2 = 0; x2 < 25; x2++) {
      nullx2  = (x2/5<4)? null->xem[x2/5] : -null->meta;
      nullx2 += (x2%5<4)? null->yem[x2%5] : -null->meta;

      nus->pp[x1][x2] -= nullx1 + nullx2; 
    }
  }
}

void     
PatternRNAModel(struct rnamodel_s *rna)
{
  int st;

  for (st = 0; st < RSTATES; st++) rna->t[st] = 0.0;

  PatternOTHModel(rna->ROB);
  PatternOTHModel(rna->ROJ);
  PatternOTHModel(rna->ROE);

  PatternOTHModel(rna->pi2->Rloop);

}
/* Function: PI2ToLog2()
 * Date:     ER, Tue Jun  8 12:46:08 CDT 1999 [St. Louis]
 *
 * Purpose:  Converts transition and emission prob's of a pi2model to log2 form
 *
 * Args:     rna - the structure for an rnamodel
 *
 * Returns:  void. 
 */
void
PI2ToLog2(struct pi2model_s *pi2, struct nullmodel_s *null)
{
  int len;               /* length of a loop (< MAXRNALOOP)  +*/
  int i,j;               /* symbols for emission prob's      +*/
  double sum1, sum2;
  double corr;
  double loop_3, loop_4, loop_10;

  /* transition prob's 
   */
  pi2->v->t1  = LOG2(pi2->v->t1);		
  pi2->v->t2s = LOG2(pi2->v->t2s);		
  pi2->v->t2b = LOG2(pi2->v->t2b);		
  pi2->v->t2i = LOG2(pi2->v->t2i);		
  pi2->v->t3  = LOG2(pi2->v->t3);
		
  pi2->w->tl = LOG2(pi2->w->tl);		
  pi2->w->tr = LOG2(pi2->w->tr);
  pi2->w->tv = LOG2(pi2->w->tv);
  pi2->w->tw = LOG2(pi2->w->tw);

  pi2->wb->tl = LOG2(pi2->wb->tl);		
  pi2->wb->tr = LOG2(pi2->wb->tr);
  pi2->wb->tv = LOG2(pi2->wb->tv);
  pi2->wb->tw = LOG2(pi2->wb->tw);
  
  /* this is a hack. 
   * The contribution of  IS1, IS2 seems to work better if I substitute
   * 
   *   tn[l] --> tn[l]*(1-\eta)^l / sum_{l'}(tn[l']*(1-\eta)^l')
   *
   * in this way, larger lengths get a lower probability
   */
  sum1 = 0.;
  sum2 = 0.;
  
  corr = LOG2(0.15);
  for (len = 1; len < MAXRNALOOP; len++) {
    pi2->is1->tn[len] = LOG2(pi2->is1->tn[len]) + ((len>0)? (len-1.)*corr : 0.); 
    sum1 += EXP2(pi2->is1->tn[len]);
  }
  corr = LOG2(0.15);
  for (len = 2; len < MAXRNALOOP; len++) {
    pi2->is2->tn[len] = LOG2(pi2->is2->tn[len]) + (len-2.)*corr;
    sum2 += EXP2(pi2->is2->tn[len]);
  }
  
  for (len = 0; len < MAXRNALOOP; len++) {
    pi2->is1->tn[len] -=  LOG2(sum1);
    pi2->is2->tn[len] -=  LOG2(sum2);
  }
  
  loop_3  = DLog2Sum(pi2->is1->tn+1,  3);
  loop_4  = DLog2Sum(pi2->is1->tn+4,  200);
  loop_10 = DLog2Sum(pi2->is1->tn+200, MAXRNALOOP-200);
  
  /* group in bins
  for (len = 1; len < MAXRNALOOP; len++) {
    if      (len < 4)  pi2->is1->tn[len] = loop_3;
    else if (len < 201) pi2->is1->tn[len] = loop_4;
    else               pi2->is1->tn[len] = loop_10;
  }
   */
  
  
  /* emission prob's 
   */
  for (i = 0; i < 25; i++) {
    pi2->w->pl  [i] = LOG2(pi2->w->pl  [i]);
    pi2->w->pr  [i] = LOG2(pi2->w->pr  [i]);
    pi2->wb->pl [i] = LOG2(pi2->wb->pl [i]);
    pi2->wb->pr [i] = LOG2(pi2->wb->pr [i]);
    pi2->is1->ps[i] = LOG2(pi2->is1->ps[i]);
    pi2->is2->ps[i] = LOG2(pi2->is2->ps[i]);
    
    for (j = 0; j < 25; j++) {
      pi2->v->pp [i][j] = LOG2(pi2->v->pp [i][j]);
      pi2->w->pp [i][j] = LOG2(pi2->w->pp [i][j]);
      pi2->wb->pp[i][j] = LOG2(pi2->wb->pp[i][j]);
    }
  }
}

/* Function: PI2Log2ToOdds()
 * Date:     ER, Tue Jun  8 12:49:42 CDT 1999 [St. Louis]
 *
 * Purpose:  Converts transition and emission prob's of a pi2model
 *           from log2 to log2odds form
 *
 * Args:     rna - the structure for an pi2model
 *
 * Returns:  void. 
 */
void
PI2Log2ToOdds(struct pi2model_s *pi2, struct nullmodel_s *null)
{
  int    len;               /* length of a loop (< MAXRNALOOP)  +*/
  int    x1,x2;
  double nullx1, nullx2;

  OTHLog2ToOdds(pi2->Rloop, null);
  
  /* transition prob's 
   */
  pi2->v->t2s -= 4.0 * null->meta;
  pi2->v->t2b -= 4.0 * null->meta;
  pi2->v->t2i -= 4.0 * null->meta;

  pi2->w->tl  -= 2.0 * null->meta;
  pi2->w->tr  -= 2.0 * null->meta;
  pi2->w->tv  -= 4.0 * null->meta;

  pi2->wb->tl -= 2.0 * null->meta;
  pi2->wb->tr -= 2.0 * null->meta;
  pi2->wb->tv -= 4.0 * null->meta;
  
  for (len = 0; len < MAXRNALOOP; len++) { 
    pi2->is1->tn[len] -= (len-1)>0? (len-1.)*2.*null->meta : .0; /* pi2->is1->tn[len] emits len-1 nts */
    pi2->is2->tn[len] -= (len-2)>0? (len-2.)*2.*null->meta : .0; /* pi2->is2->tn[len] emits len-2 nts */
  }  
  
  /* emission prob's 
   */
  for (x1 = 0; x1 < 25; x1++) {
    nullx1  = (x1/5<4)? null->xem[x1/5] : -null->meta;
    nullx1 += (x1%5<4)? null->yem[x1%5] : -null->meta;
    
    pi2->w->pl  [x1] -= nullx1;
    pi2->w->pr  [x1] -= nullx1;
    pi2->wb->pl [x1] -= nullx1;
    pi2->wb->pr [x1] -= nullx1;
    pi2->is1->ps[x1] -= nullx1;
    pi2->is2->ps[x1] -= nullx1;
    
    for (x2 = 0; x2 < 25; x2++) {
      nullx2  = (x2/5<4)? null->xem[x2/5] : -null->meta;
      nullx2 += (x2%5<4)? null->yem[x2%5] : -null->meta;
      
      pi2->v->pp [x1][x2] -= nullx1 + nullx2; 
      pi2->w->pp [x1][x2] -= nullx1 + nullx2; 
      pi2->wb->pp[x1][x2] -= nullx1 + nullx2; 
    }
  }
}

/* Function: RNAToLog2()
 * Date:     ER, Tue Jun  8 12:46:08 CDT 1999 [St. Louis]
 *
 * Purpose:  Converts transition and emission prob's of a rnamodel to log2 form
 *
 * Args:     rna - the structure for an rnamodel
 *
 * Returns:  void. 
 */
void
RNAToLog2(struct rnamodel_s *rna, struct nullmodel_s *null)
{
  int idx;               /* index for transition prob's      +*/

  NUSToLog2(rna->nus);
  PI2ToLog2(rna->pi2, null);

  /* transition prob's 
   */
  for (idx = 0; idx < RTRANS; idx++) 
    rna->t[idx] = LOG2(rna->t[idx]);
  
}

/* Function: RNALog2ToOdds()
 * Date:     ER, Tue Jun  8 12:49:42 CDT 1999 [St. Louis]
 *
 * Purpose:  Converts transition and emission prob's of a rnamodel
 *           from log2 to log2odds form
 *
 * Args:     rna - the structure for an pi2model
 *
 * Returns:  void. 
 */
void
RNALog2ToOdds(struct rnamodel_s *rna, struct nullmodel_s *null)
{

  OTHLog2ToOdds(rna->ROB, null);
  OTHLog2ToOdds(rna->ROJ, null);
  OTHLog2ToOdds(rna->ROE, null);
  
  NUSLog2ToOdds(rna->nus, null);
  PI2Log2ToOdds(rna->pi2, null);
}

/* Function: PrintNUSModel()
 * Date:     ER, Wed Jun  9 17:43:05 CDT 1999 [St. Louis]
 *
 * Purpose:  print base pair alignment stuff for NUS model (given in log2 form)
 *
 * Args:     nusmodel, nullmodel
 *
 * Returns:  void, prints stuff.
 */
void
PrintNUSModel(struct nusmodel_s *nus)
{
  int    x1,x2,y1,y2;
  double mutp;
  double info   = 0.0;
  double expect = 0.0;

  printf("\nNUS MODEL -- Transition probabilities\n");
  printf("t[M->aM]  = %f\n", EXP2(nus->tl));		
  printf("t[M-> Ma] = %f\n", EXP2(nus->tr));
  printf("t[M->aMb] = %f\n", EXP2(nus->tv));
  printf("t[M->M M] = %f\n", EXP2(nus->tw));
  printf("t[M-> e]  = %f\n", EXP2(nus->te));

  printf("\nNUS MODEL -- Emission probabilities\n");
  for (x1 = 0; x1 < 5; x1++)
    for (y1 = 0; y1 < 5; y1++) {
      printf("%c*%c wl = %12f\n ", 
	     (x1<4)?DNAAlphabet[x1]:'-',    (y1<4)?DNAAlphabet[y1]:'-', 
	     EXP2(nus->p[idx5(x1,y1)]));
      
      for (x2 = 0; x2 < 5; x2++)
	for (y2 = 0; y2 < 5; y2++) {
	  mutp = nus->p[idx5(x1,y1)];
	  printf("%c*%c %c*%c w = %12f \n",
		 (x1<4)?DNAAlphabet[x1]:'-', (y1<4)?DNAAlphabet[y1]:'-',
		 (x2<4)?DNAAlphabet[x2]:'-', (y2<4)?DNAAlphabet[y2]:'-',
		 EXP2(nus->pp[idx5(x1,y1)][idx5(x2,y2)]));
	  info += EXP2(nus->pp[idx5(x1,y1)][idx5(x2,y2)]) * 
	    (nus->pp[idx5(x1,y1)][idx5(x2,y2)] - mutp);
	  expect += EXP2(mutp) * 
	    (nus->pp[idx5(x1,y1)][idx5(x2,y2)] - mutp);
	}
    }
  
  printf("NUS model info   = %f bits\n", info);
  printf("NUS model expect = %f bits\n", expect);
}

/* Function: PrintPI2Model()
 * Date:     ER, Tue Jun  8 13:32:19 CDT 1999 [St. Louis]
 *
 * Purpose:  print base pair alignment stuff for PI2 model (given in log2 form)
 *
 * Args:     rnamodel, nullmodel
 *
 * Returns:  void, prints stuff.
 */
void
PrintPI2Model(struct pi2model_s *pi2)
{
  int    x1,x2,y1,y2;
  int    len;               /* length of a loop (< MAXRNALOOP)  +*/
  double mutp;
  double info   = 0.0;
  double expect = 0.0;

  printf("\nRloop MODEL \n");
  PrintOTHModel(pi2->Rloop);

  printf("\nPI2 MODEL -- Transition probabilities\n");
  printf("t[V->IS1]  \t = %f\n", EXP2(pi2->v->t1));		
  printf("t[V->IS2S] \t = %f\n", EXP2(pi2->v->t2s));		
  printf("t[V->IS2B] \t = %f\n", EXP2(pi2->v->t2b));		
  printf("t[V->IS2I] \t = %f\n", EXP2(pi2->v->t2i));		
  printf("t[V->WB WB]\t = %f\n\n", EXP2(pi2->v->t3));
		
  printf("t[W->aW] \t = %f\n", EXP2(pi2->w->tl));		
  printf("t[W-> Wa]\t = %f\n", EXP2(pi2->w->tr));
  printf("t[W->aVb]\t = %f\n", EXP2(pi2->w->tv));
  printf("t[W->W W]\t = %f\n\n", EXP2(pi2->w->tw));

  printf("t[WB->aWB]  \t = %f\n", EXP2(pi2->wb->tl));		
  printf("t[WB-> WBa] \t = %f\n", EXP2(pi2->wb->tr));
  printf("t[WB->aVb]  \t = %f\n", EXP2(pi2->wb->tv));
  printf("t[WB->WB WB]\t = %f\n\n", EXP2(pi2->wb->tw));
  
  for (len = 1; len < MAXRNALOOP; len++) 
     printf("t[IS1->%d]\t = %f\n", len, EXP2(pi2->is1->tn[len]));
  printf("\n");
  for (len = 2; len < MAXRNALOOP; len++) 
     printf("t[IS2->%d]\t = %f\n", len ,EXP2(pi2->is2->tn[len]));

  printf("\nPI2 MODEL -- Emission probabilities\n");
  for (x1 = 0; x1 < 5; x1++)
    for (y1 = 0; y1 < 5; y1++) {
      printf("%c*%c    wl = %12f wr  = %12f\n      wbl = %12f wbr = %12f\n      is1 = %12f is2 = %12f \n", 
	     (x1<4)?DNAAlphabet[x1]:'-',     (y1<4)?DNAAlphabet[y1]:'-', 
	     EXP2(pi2->w->pl  [idx5(x1,y1)]), EXP2(pi2->w->pr  [idx5(x1,y1)]),
	     EXP2(pi2->wb->pl [idx5(x1,y1)]), EXP2(pi2->wb->pr [idx5(x1,y1)]),
	     EXP2(pi2->is1->ps[idx5(x1,y1)]), EXP2(pi2->is2->ps[idx5(x1,y1)]));
      
      for (x2 = 0; x2 < 5; x2++)
	for (y2 = 0; y2 < 5; y2++) {
	  mutp = pi2->w->pl[idx5(x1,y1)] + pi2->w->pl[idx5(x2,y2)];
	  printf("%c*%c %c*%c v = %12f \n",
		 (x1<4)?DNAAlphabet[x1]:'-', (y1<4)?DNAAlphabet[y1]:'-',
		 (x2<4)?DNAAlphabet[x2]:'-', (y2<4)?DNAAlphabet[y2]:'-',
		 EXP2(pi2->v->pp [idx5(x1,y1)][idx5(x2,y2)]));

	  info += EXP2(pi2->v->pp[idx5(x1,y1)][idx5(x2,y2)]) * 
	    (pi2->v->pp[idx5(x1,y1)][idx5(x2,y2)] - mutp);
	  expect += EXP2(mutp) * 
	    (pi2->v->pp[idx5(x1,y1)][idx5(x2,y2)] - mutp);
	}
    }
  
  printf("PI2 model info   = %f bits\n", info);
  printf("PI2 model expect = %f bits\n", expect);
}

/* Function: PrintRNAModel()
 * Date:     ER, Tue Jun  8 13:32:19 CDT 1999 [St. Louis]
 *
 * Purpose:  print base pair alignment stuff for RNA model (given in log2 form)
 *
 * Args:     rnamodel, nullmodel
 *
 * Returns:  void, prints stuff.
 */
void
PrintRNAModel(struct rnamodel_s *rna)
{
  int idx;

  printf("\nROB MODEL \n");
  PrintOTHModel(rna->ROB);

  printf("\nROJ MODEL \n");
  PrintOTHModel(rna->ROJ);

  printf("\nROE MODEL \n");
  PrintOTHModel(rna->ROE);

  printf("\nRNA MODEL -- Transition probabilities\n");
  for (idx = 0; idx < RTRANS; idx++)
    printf("t[%s]\t = %f\n", rtrNAME[idx], EXP2(rna->t[idx])); 

  printf("\nNUS MODEL -- Probabilities\n");
  PrintNUSModel(rna->nus);

  printf("\nPI2 MODEL -- Probabilities\n");
  PrintPI2Model(rna->pi2);

}
















