/************************************************************
 * QRNA - Comparative analysis of biological sequences 
 *        with pair hidden Markov models and pair stochastic context-free grammars
 * Copyright (C) 2000-Howard Hughes Medical Institute
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* scorewithmodels.c
 *
 * ER, Fri Feb 25 22:00:06 CST 2000 [St. Louis]
 * 
 * 
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

#include "funcs.h"
#include "globals.h"
#include "squid.h"
#include "structs.h"

#ifdef MEMDEBUG
#include "dbmalloc.h"
#endif

static void cal_structure (FILE *ofp,
			    SQINFO sqinfoX, int *seqX,  
			    SQINFO sqinfoY, int *seqY, 
			    int start, int L, 
			    struct rnamodel_s *rnamodel, 
			    struct rnascfg_s  *mx, 
			    struct sc3_s      *sc, 
			    int cyk, int logodds, int do_nus, int rnass, int traceback, struct end_s *rnaends, int isrnarev);
static void RNAbanner (FILE *ofp, int cyk);
static void score_with_null(FILE *ofp, 
			    int *isegX, int *isegY, int len, 
			    struct nullmodel_s *null, 
			    double nullsc, 
			    double nullrevsc, 
			    int verbose);

static int score_global(FILE *ofp, 
			SQINFO sqinfoX, int *isegX, 
			SQINFO sqinfoY, int *isegY, 
			int start, int win, 
			struct dos_s      d, 
			struct model_s   *model, 
			struct dpd_s     *dpd, 
			struct rnascfg_s *mx, 
			struct sc3_s     *sc, 
			struct ali_s     *ali, 
			int *ret_scfg_status, int *ret_scfgrv_status,
			int cyk, int logodds, int rnass, int sweep, int traceback, struct end3_s *ends);

static int score_viterbi_diag(FILE *ofp, 
			      SQINFO sqinfoX, int *isegX, 
			      SQINFO sqinfoY, int *isegY,  
			      int start, int win,
			      struct dos_s      d, 
			      struct model_s   *model,
			      struct dpd_s     *dpd, 
			      struct rnascfg_s *mx, 
			      struct sc3_s     *sc, 
			      struct ali_s     *ali, 
			      int *ret_scfg_status, int *ret_scfgrv_status,
			      int alignment, int cyk, int logodds, int rnass, int sweep, int traceback, 
			      struct end3_s *ends);

static int score_forward_diag(FILE *ofp, 
			      SQINFO sqinfoX, int *isegX, 
			      SQINFO sqinfoY, int *isegY, 
			      int start, int win, 
			      struct dos_s      d, 
			      struct model_s   *model, 
			      struct dpd_s     *dpd, 
			      struct rnascfg_s *mx, 
			      struct sc3_s     *sc, 
			      struct dpsc3_s   *ardiag, 
			      struct ali_s     *ali, 
			      int *ret_scfg_status, int *ret_scfgrv_status,
			      int cyk, int logodds, int sweep, struct end3_s *ends);

static int score_semi(FILE *ofp, 
		      SQINFO sqinfoX, int *iseqX, int startX, int lenX, 
		      SQINFO sqinfoY, int *iseqY, int startY, int lenY,
		      int start, int *ret_legnew, int Lw, 
		      struct dos_s      d,  
		      struct model_s   *model, 
		      struct dpd_s     *dpd, 
		      struct othdpf_s  *othdpf, 
		      struct rnascfg_s *mx, 
		      struct scores_s  *sc, 
		      struct ali_s     *ali, 
		      int alignment, int cyk, int logodds, int rnass, int sweep, int traceback, 
		      struct end3_s *ends);

static int score_full(FILE *ofp, 
		      SQINFO sqinfoX, int *iseqX, int startX, int lenX, 
		      SQINFO sqinfoY, int *iseqY, int startY, int lenY,
		      int start, int *ret_legnew, int Lw, 
		      struct dos_s     d, 
		      struct model_s   *model, 
		      struct dpf_s     *dpf, 
		      struct rnascfg_s *mx, 
		      struct scores_s  *sc, 
		      struct ali_s     *ali, 
		      int alignment, int cyk, int logodds, int rnass, int sweep, int traceback, 
		      struct end3_s *ends);

static void test_OTH_model(FILE *ofp, 
			   SQINFO sqinfoX, int *isegX, 
			   SQINFO sqinfoY, int *isegY, 
			   int win,  
			   struct othmodel_s *oth, 
			   struct othdpd_s   *othdp, 
			   double            *othsc, 
			   struct ali_s      *ali, 
			   int alignment, int traceback, 
			   struct end_s *othends);

/* Function: AllocEnds()
 * Date:     ER, Wed Sep  5 13:10:37 CDT 2001  [St. Louis]
 *
 * Purpose:  Allocates memory for structure end3_s 
 *
 * Returns:  end3_s are allocated
 */
struct end3_s *
AllocEnds()
{
  struct end3_s *ends;   
   
  ends = (struct end3_s *) MallocOrDie (sizeof(struct end3_s));

  ends->cod = (struct end_s *) MallocOrDie (sizeof(struct end_s));
  ends->oth = (struct end_s *) MallocOrDie (sizeof(struct end_s));
  ends->rna = (struct end_s *) MallocOrDie (sizeof(struct end_s));

  /* initialize */
  ends->cod->lend = -1;
  ends->cod->rend = -1;
  ends->oth->lend = -1;
  ends->oth->rend = -1;
  ends->rna->lend = -1;
  ends->rna->rend = -1;

  return ends;
}

/* Function: AllocScores()
 * Date:     ER, Sat Feb 26 13:53:21 CST 2000 [St. Louis]
 *
 * Purpose:  Allocates memory for all scores
 *
 * Returns:  scores are allocated
 */
struct scores_s *
AllocScores()
{
  struct scores_s *sc;       /* structure with dp matrices   */

  sc = (struct scores_s *) MallocOrDie (sizeof(struct scores_s));

  sc->global = AllocSc3();
  sc->vidiag = AllocSc3();
  sc->frdiag = AllocSc3();
  sc->vifull = AllocSc3();
  sc->frfull = AllocSc3();
  sc->visemi = AllocSc3();
  sc->frsemi = AllocSc3();

  sc->ardiag = (struct dpsc3_s *) MallocOrDie (sizeof(struct dpsc3_s));
  sc->arfull = (struct dpsc3_s *) MallocOrDie (sizeof(struct dpsc3_s));

  return sc;
}

struct sc3_s *
AllocSc3()
{
  struct sc3_s *sc3;

  sc3 = (struct sc3_s *) MallocOrDie (sizeof(struct sc3_s));

  return sc3;
}

void
FreeEnds(struct end3_s *ends)
{
  free(ends->oth);
  free(ends->cod);
  free(ends->rna);

  free(ends);
}

void
FreeSc3(struct sc3_s *sc3)
{
  free(sc3);
}

void
FreeScores(struct scores_s *sc)
{
  FreeSc3(sc->global);
  FreeSc3(sc->vidiag);
  FreeSc3(sc->frdiag);
  FreeSc3(sc->vifull);
  FreeSc3(sc->frfull);
  FreeSc3(sc->visemi);
  FreeSc3(sc->frsemi);

  free(sc->ardiag);
  free(sc->arfull);

  free(sc);
}

/* Function: PrintScores()
 * Date:     ER, Mon 12 April 1999 [St. Louis]
 *
 * Purpose:  print scores
 *
 * Args:     ofp, sqinfoX, sqinfoY, nullsc, pamsc, rnasc
 *
 * Returns:  void.
 */
void
PrintScores(FILE *ofp, struct sc3_s *sc, struct end3_s *ends)
{
  char   *winn;
  double  othsc, codsc, rnasc;
  double  post_o, post_c, post_r;  /* posterior probs in bits */

  winn = WinnerScore(sc);

  othsc = LOG2(EXP2(sc->oth->pl) + EXP2(sc->oth->mn)) - 1.0;
  codsc = LOG2(EXP2(sc->cod->pl) + EXP2(sc->cod->mn)) - 1.0;
  rnasc = LOG2(EXP2(sc->rna->pl) + EXP2(sc->rna->mn)) - 1.0;

  fprintf(ofp, "winner = %s \n", winn);
 
  fprintf(ofp, "OTH  ends = %d %d\n", ends->oth->lend, ends->oth->rend);
  fprintf(ofp, "COD  ends = %d %d\n", ends->cod->lend, ends->cod->rend);
  fprintf(ofp, "RNA  ends = %d %d\n", ends->rna->lend, ends->rna->rend);
 
  fprintf(ofp, "              OTH = %12.3f             COD = %12.3f             RNA = %12.3f \n", 
	  othsc, codsc, rnasc);

  post_o = PosteriorLog2(othsc, codsc, rnasc);
  post_c = PosteriorLog2(codsc, othsc, rnasc);
  post_r = PosteriorLog2(rnasc, othsc, codsc);

  fprintf(ofp, "   logoddspostOTH = %12.3f  logoddspostCOD = %12.3f  logoddspostRNA = %12.3f \n\n", 
	  post_o - post_o, post_c - post_o, post_r - post_o);
}

/* Function: WinnerScore()
 * Date:     ER, Mon Sep  3 16:03:32 CDT 2001    [St. Louis]
 *
 * Purpose:  calculates the winner score
 *
 * Args:     ofp, sqinfoX, sqinfoY, nullsc, pamsc, rnasc
 *
 * Returns:  winner
 */
char *
WinnerScore(struct sc3_s *sc)
{
  char *win;
  double othsc, codsc, rnasc;

  othsc = LOG2(EXP2(sc->oth->pl) + EXP2(sc->oth->mn)) - 1.0;
  codsc = LOG2(EXP2(sc->cod->pl) + EXP2(sc->cod->mn)) - 1.0;
  rnasc = LOG2(EXP2(sc->rna->pl) + EXP2(sc->rna->mn)) - 1.0;

  if      (othsc >= codsc && othsc >= rnasc) win = "OTH";
  else if (codsc >= othsc && codsc >= rnasc) win = "COD";
  else if (rnasc >= othsc && rnasc >= codsc) win = "RNA";
  else Die ("who is the winner?");

  return win;
}

/* Function: ScoreWithModels()
 * Date:     ER, Sat Feb 26 10:39:26 CST 2000 [St. Louis]
 *
 * Purpose:  decide which scoring method to use. Called by main().
 *
 * Args:     
 *
 * Returns:  void.
 */
void
ScoreWithModels(FILE *ofp, 
		SQINFO sqinfoX, int *isegX, int *iseqX, 
		SQINFO sqinfoY, int *isegY, int *iseqY, 
		int start, int win, int Lw, 
		struct dos_s      d, 
		struct model_s   *model, 
		struct dpd_s     *dpd, 
		struct dpf_s     *dpf, 
		struct rnascfg_s *mx, 
		struct scores_s  *sc, 
		struct ali_s     *ali, 
		int alignment, int cyk, int logodds, int rnass, int shuffle, int sweep, 
		int traceback, int verbose)
{
  struct end3_s *ends;
  int            startX, endX;        /* first and last position of analysis for seqX   */
  int            startY, endY;        /* first and last position of analysis for seqY   */
  int            lenX, lenY;          /* len of seq's  without gaps                     */
  int            fstX, fstY;          /* mapping of startX and startY to the seq without gaps */
  int            leg_new;
  int            is_significant;
  int            scfg_status;
  int            scfgrv_status;
  int           *iseqshX, *iseqshY;
  int           *seqX,    *seqY;
  double        *freqX,   *freqY;
  double         id;

  ends = AllocEnds();

  /* default values */
  
  is_significant = FALSE; /* TRUE = OTH score is significant, continue scoring with the COD and RNA models        */
  scfg_status    = FALSE; /* TRUE = the scfg part of the RNA model has already been calculated -- given strand    */
  scfgrv_status  = FALSE; /* TRUE = the scfg part of the RNA model has already been calculated -- reverse strand  */

  lenX = win;
  lenY = win; 
  
  if (sweep) startX = 0;
  else       startX = start;
  startY = start;
  endX = startX + lenX - 1;
  endY = startY + lenY - 1;

  freqX = AllocBaseFreq();
  freqY = AllocBaseFreq();
  
  
  /* (0) leg = minimum length of aligment 
     RemoveJointGaps(lenX, isegX+startX, lenY, isegY+startY, &win); 
     
     *
     * this function can be passed to analyze_2_sequences() in "qrna.c" if 
     * one wants to remove the common gaps of the two alignments globally.
     * The difference is only relevant when using a window.
     */
  
  /*if (verbose) 
    test_OTH_model(ofp, sqinfoX, isegX+startX, sqinfoY, isegY+startY, win, model->cod->COJ, 
		   dpd->oth, sc->ardiag->oth, ali, alignment, traceback, ends->oth);
  */

  /* if (shuffle == TRUE) shuffle aligment */
  if (shuffle) {
    AllocIntSeqs(win, &iseqshX, &iseqshY);
    DupIntSeq(isegX, iseqshX, startX+win-1, win-1);
    DupIntSeq(isegY, iseqshY, startY+win-1, win-1);  

    Shuffle2IntSequences(iseqshX, iseqshY, win, win-1, win-1, verbose);

    seqX = iseqshX;
    seqY = iseqshY;
  }
  else {
    seqX = isegX+startX;
    seqY = isegY+startY;
  }

  id = PerIdSeqs(seqX, seqY, win-1, win-1);

  /* (1) duplicate the region (start, start+win-1) of iseg into iseq. 
   *     Remove the gaps, and calculate lenX, and lenY.
   */
  DupIntSeq(seqX, iseqX, win-1, win-1);
  DupIntSeq(seqY, iseqY, win-1, win-1);  
      
  RemoveGaps(iseqX, win, &lenX, verbose); /* len of seqX without gaps */
  RemoveGaps(iseqY, win, &lenY, verbose); /* len of seqY without gaps */

  fstX = PosNoGaps(isegX, startX);
  fstY = PosNoGaps(isegY, startY);

  if (!sweep) {
    fprintf(ofp, "length alignment: %d (id=%.2f)", win, id);
    if (shuffle) fprintf(ofp, "(shuffled)\n");
    else         fprintf(ofp, "\n");
    
    BaseComp(ofp, iseqX, lenX-1, lenX-1, freqX);
    BaseComp(ofp, iseqY, lenY-1, lenY-1, freqY);

    fprintf(ofp, "posX: %d-%d [%d-%d](%d) -- (%.2f %.2f %.2f %.2f) \n", 
	    startX, endX, fstX, fstX+lenX-1, lenX, freqX[0], freqX[1], freqX[2], freqX[3]);
    fprintf(ofp, "posY: %d-%d [%d-%d](%d) -- (%.2f %.2f %.2f %.2f) \n", 
	    startY, endY, fstY, fstY+lenY-1, lenY, freqY[0], freqY[1], freqY[2], freqY[3]);
  }

  /* Calculate the null-model scores
   */
  score_with_null(ofp, seqX, seqY, win, model->null, sc->null, sc->nullrev, verbose);
  
  /* GLOBAL (global alignment   */
  if (d.global) 
    is_significant = 
      score_global(ofp, sqinfoX, seqX, sqinfoY, seqY, start, win, d, model, dpd, 
		   mx, sc->global, ali, &scfg_status, &scfgrv_status, 
		   cyk, logodds, rnass, sweep, traceback, ends);
  
  /* VITERBI (diag local aligmnent) 
   */
  if (d.vidiag) 
    is_significant = 
      score_viterbi_diag(ofp, sqinfoX, seqX, sqinfoY, seqY, start, win, d, model, dpd, 
			 mx, sc->vidiag, ali, &scfg_status, &scfgrv_status, 
			 alignment, cyk, logodds, rnass, sweep, traceback, ends);
  
  /* FORWARD (diag local aligmnent)
   */
  if (d.frdiag) 
    is_significant = 
      score_forward_diag(ofp, sqinfoX, seqX, sqinfoY, seqY, start, win, d, model, dpd, 
			 mx, sc->frdiag, sc->ardiag, ali, &scfg_status, &scfgrv_status, cyk, logodds, sweep, ends);

  /* SEMI-DIAG local aligmnent. [Here is where we use iseq]  
   */
  if (d.visemi || d.frsemi)
    is_significant = 
      score_semi(ofp, sqinfoX, iseqX, 0, lenX, sqinfoY, iseqY, 0, lenY, start, &leg_new, Lw, d, model, 
		 dpd, dpf->oth, mx, sc, ali, alignment, cyk, logodds, rnass, sweep, traceback, ends);
  

  /* FULL local aligmnent.  [Here is where we use iseq]
   */
  if (d.vifull || d.frfull)
    is_significant = 
      score_full(ofp, sqinfoX, iseqX, 0, lenX, sqinfoY, iseqY, 0, lenY, start, &leg_new, Lw, d, model,
		 dpf, mx, sc, ali, alignment, cyk, logodds, rnass, sweep, traceback, ends);
  
  if (is_significant && sweep) {
    fprintf(ofp, "length alignment: %d ", win);
    if (shuffle) fprintf(ofp, "(shuffled)\n");
    else         fprintf(ofp, "\n");
    
    BaseComp(ofp, iseqX, lenX-1, lenX-1, freqX);
    BaseComp(ofp, iseqY, lenY-1, lenY-1, freqY);

   fprintf(ofp, "posX: %d-%d [%d-%d](%d) -- (%.2f %.2f %.2f %.2f) \n", 
	   startX, endX, fstX, fstX+lenX-1, lenX, freqX[0], freqX[1], freqX[2], freqX[3]);
   fprintf(ofp, "posY: %d-%d [%d-%d](%d) -- (%.2f %.2f %.2f %.2f) \n", 
	   startY, endY, fstY, fstY+lenY-1, lenY, freqY[0], freqY[1], freqY[2], freqY[3]);
  }
  
  if (!d.vidiag && !d.vifull && !d.visemi && alignment) 
    Print2Seqs(ofp, &sqinfoX, seqX, &sqinfoY, seqY, 0, win);

  if (shuffle) {
    free(iseqshX);
    free(iseqshY);
  }

  free(freqX);
  free(freqY);
  FreeEnds(ends);
}

void
cal_structure (FILE *ofp,
	       SQINFO sqinfoX, int *seqX,  
	       SQINFO sqinfoY, int *seqY, 
	       int start, int L, 
	       struct rnamodel_s *rnamodel, 
	       struct rnascfg_s  *mx, 
	       struct sc3_s      *sc, 
	       int cyk, int logodds, int do_nus, int rnass, int traceback, struct end_s *rnaends, int isrnarev)
{
  int d;
  
  d = abs(rnaends->rend - rnaends->lend);
  
  if (d > 0) {
    if (cyk && ((rnass && (WinnerScore(sc) == "RNA")) || traceback))
      CYKTracebackRNA(ofp, sqinfoX, seqX, sqinfoY, seqY, start, L, rnamodel, mx, do_nus, logodds, traceback, rnaends, isrnarev);
    
    if (!cyk && rnass && WinnerScore(sc) == "RNA") 
      PosteriorRNA(ofp, sqinfoX, seqX, sqinfoY, seqY, start, L, rnamodel, mx, do_nus, logodds, traceback, rnaends, isrnarev);
  }
}

void 
RNAbanner (FILE *ofp, int cyk)
{
  if (cyk) 
    fprintf(ofp, "[CYK SCFG]\n");
  else 
    fprintf(ofp, "[Inside SCFG]\n");
}

void
score_with_null(FILE *ofp, int *isegX, int *isegY, int len, struct nullmodel_s *null, 
		double nullsc, double nullrevsc, int verbose)
{
  nullsc    = ScoreWithNullDiag(isegX, isegY, 0, len, null);
  nullrevsc = ScoreWithNullRevDiag(isegX, isegY, 0, len, null);
  if (verbose) PrintNull(ofp, null, nullsc, nullrevsc);
}

int
score_global(FILE *ofp, 
	     SQINFO sqinfoX, int *isegX, 
	     SQINFO sqinfoY, int *isegY, 
	     int start, int win,
	     struct dos_s      d, 
	     struct model_s   *model, 
	     struct dpd_s     *dpd, 
	     struct rnascfg_s *mx, 
	     struct sc3_s     *sc, 
	     struct ali_s     *ali, 
	     int *ret_scfg_status, int *ret_scfgrv_status,
	     int cyk, int logodds, int rnass, int sweep, int traceback, struct end3_s *ends)
{
  int isrnarev = FALSE;
  int scfg_status;
  int scfgrv_status;
 
  scfg_status   = *ret_scfg_status;
  scfgrv_status = *ret_scfgrv_status;

  sc->oth = ScoreWithOTH2(ofp, isegX, isegY, 0, win, model->oth, traceback);

  sc->cod = ScoreWithCOD(ofp, isegX, isegY, 0, win, model->cod, model->null, 
			 dpd->cod, logodds, traceback);
  sc->rna = ScoreWithRNA(ofp, sqinfoX, isegX, sqinfoY, isegY, 0, win, model, dpd->oth, mx, 
			 ret_scfg_status, ret_scfgrv_status, cyk, logodds, d.nus, traceback);

  if (sc->rna->mn > sc->rna->pl) isrnarev = TRUE;
 /* since this is "global", ends are the total ends
   */
  if (sc->cod->pl > sc->cod->mn) {
    ends->cod->lend = 0;
    ends->cod->rend = win - 1;
  }
  else {
    ends->cod->lend = win - 1;
    ends->cod->rend = 0;
  }

  if (sc->oth->pl > sc->oth->mn) {
    ends->oth->lend = 0;
    ends->oth->rend = win - 1;
  }
  else {
    ends->oth->lend = win - 1;
    ends->oth->rend = 0;
  }

  if (sc->rna->pl > sc->rna->mn) {
    ends->rna->lend = 0;
    ends->rna->rend = win - 1;
  }
  else {
    ends->rna->lend = win - 1;
    ends->rna->rend = 0;
  }
  
  cal_structure(ofp, sqinfoX, isegX, sqinfoY, isegY, 0, win, model->rna, mx, sc, 
		cyk, logodds, d.nus, rnass, traceback, ends->rna, isrnarev);

  /* add absolute position to ends
   */
  ends->cod->lend += start;
  ends->oth->lend += start;
  ends->rna->lend =+ start;

  ends->cod->rend += start;
  ends->oth->rend += start;
  ends->rna->rend += start;

  fprintf(ofp, "GLOBAL -- ");
  RNAbanner(ofp, cyk);

  PrintScores(ofp, sc, ends);

  *ret_scfg_status   = scfg_status;
  *ret_scfgrv_status = scfgrv_status;
  return 1;
}

int
score_viterbi_diag(FILE *ofp, 
		   SQINFO sqinfoX, int *isegX, 
		   SQINFO sqinfoY, int *isegY,  
		   int start, int win,
		   struct dos_s      d, 
		   struct model_s   *model, 
		   struct dpd_s     *dpd,  
		   struct rnascfg_s *mx, 
		   struct sc3_s     *sc, 
		   struct ali_s     *ali, 
		   int *ret_scfg_status,int *ret_scfgrv_status,
		   int alignment, int cyk, int logodds, int rnass, int sweep, int traceback, struct end3_s *ends)
{
  int   isrnarev = FALSE;
  int   scfg_status;
  int   scfgrv_status;

  scfg_status   = *ret_scfg_status;
  scfgrv_status = *ret_scfgrv_status;

  sc->oth = ViterbiOTHDiag(ofp, sqinfoX, isegX, sqinfoY, isegY, 0, win, 
			   model->oth, dpd->oth, ali, alignment, sweep, traceback, ends->oth);

  /*if (sweep && sc->oth < 0.0) return 0;*/

  sc->cod = ViterbiCODDiag(ofp, sqinfoX, isegX, sqinfoY, isegY, 0, win,
			   model->cod, dpd, model->null, ali,   
			   alignment, logodds, traceback, ends);  
  sc->rna = ViterbiRNADiag(ofp, sqinfoX, isegX, sqinfoY, isegY, 0, win, 
			   model, dpd, mx, ali, &scfg_status, &scfgrv_status, 
			   alignment, cyk, logodds, d.nus, traceback, ends);

 if (sc->rna->mn > sc->rna->pl) isrnarev = TRUE;
  
  cal_structure(ofp, sqinfoX, isegX, sqinfoY, isegY, 0, win, model->rna, mx, sc, 
		cyk, logodds, d.nus, rnass, traceback, ends->rna, isrnarev);
  
  /* add absolute position to ends
   */
  ends->cod->lend += start;
  ends->oth->lend += start;
  ends->rna->lend += start;
  
  ends->cod->rend += start;
  ends->oth->rend += start;
  ends->rna->rend += start;
  
  fprintf(ofp, "LOCAL_DIAG_VITERBI -- ");
  RNAbanner(ofp, cyk);
  PrintScores(ofp, sc, ends);

  *ret_scfg_status   = scfg_status;
  *ret_scfgrv_status = scfgrv_status;
  return 1;
}

int
score_forward_diag(FILE *ofp, 
		   SQINFO sqinfoX, int *isegX, 
		   SQINFO sqinfoY, int *isegY,
		   int start, int win, 
		   struct dos_s      d, 
		   struct model_s   *model, 
		   struct dpd_s     *dpd, 
		   struct rnascfg_s *mx, 
		   struct sc3_s     *sc, 
		   struct dpsc3_s   *ardiag, 
		   struct ali_s     *ali, 
		   int *ret_scfg_status, int *ret_scfgrv_status,
		   int cyk, int logodds, int sweep, struct end3_s *ends)
{   
  int  scfg_status;
  int  scfgrv_status;

  scfg_status   = *ret_scfg_status;
  scfgrv_status = *ret_scfgrv_status;

  sc->oth = ForwardOTHDiag(ofp, isegX, isegY, 0, win, model->oth, dpd->oth, ardiag->oth);
  /*if (sweep && sc->oth < 0.0) return 0;*/

  sc->cod = ForwardCODDiag(ofp, isegX, isegY, 0, win, model->cod, dpd, model->null, logodds, ardiag);
  sc->rna = ForwardRNADiag(ofp, sqinfoX, isegX, sqinfoY, isegY, 0, win, model, dpd, mx, ardiag, 
			   &scfg_status, &scfgrv_status, cyk, logodds, d.nus);

  fprintf(ofp, "LOCAL_DIAG_FORWARD -- ");
  RNAbanner(ofp, cyk);
  PrintScores(ofp, sc, ends);

  *ret_scfg_status   = scfg_status;
  *ret_scfgrv_status = scfgrv_status;
  return 1;
}


/* Function: score_semi()
 * Date:     ER, Sat Feb 26 16:34:13 CST 2000 [St. Louis]
 *
 * Purpose:  this is a little trick, to not to have to implement the full COD/RNA.
 *      
 *           I take the "new aligment" generated by 
 *           ViterbiOTH_L2() and use it to calculate COD and RNA scores
 *           as if it where a "given alignment" therefore as a diag DP
 *
 * Args:     
 *
 * Returns:  void.
 */
int
score_semi(FILE *ofp, 
	   SQINFO sqinfoX, int *iseqX, int startX, int lenX, 
	   SQINFO sqinfoY, int *iseqY, int startY, int lenY,
	   int start, int *ret_legnew, int Lw, 
	   struct dos_s      d, 
	   struct model_s   *model, 
	   struct dpd_s     *dpd, 
	   struct othdpf_s  *othdpf, 
	   struct rnascfg_s *mx, 
	   struct scores_s  *sc, 
	   struct ali_s     *ali, 
	   int alignment, int cyk, int logodds, int rnass, int sweep, int traceback, 
	   struct end3_s *ends)
{
  int     Ltot;       /* total space necessary to allocate a possible alignment of the two sequences */
  int     legnew;     /* length of new aligment generated by full viterbi                            */
  int     isrnarev      = FALSE;
  int     scfg_status   = FALSE;
  int     scfgrv_status = FALSE;
  double  id;

  /* (1) apply full viterbi according to the OTH model
   */
  sc->visemi->oth = ViterbiOTHFull(ofp, sqinfoX, iseqX, sqinfoY, iseqY, startX, startY, lenX, lenY,
				   &legnew, Lw, model->oth, othdpf, ali, alignment, sweep, traceback, ends->oth);
  
  /*if (sweep && sc->visemi->oth < -300.0) return 0;*/

  /* (2) dump the alignments in iseq and calculate the length of the alignment
   */

  Ltot = (lenX > lenY)? 2*lenX : 2*lenY;
  IntizeGapAsequence(ali->charX, Ltot-legnew, legnew, iseqX, FALSE);
  IntizeGapAsequence(ali->charY, Ltot-legnew, legnew, iseqY, FALSE);
  
  id = PerIdSeqs(iseqX, iseqY, legnew-1, legnew-1);

  if (legnew > 2*Lw) 
    Die ("alignment is larger than allocated space\n");
    
  if (d.visemi) 
    {
      sc->visemi->cod = ViterbiCODDiag(ofp, sqinfoX, iseqX, sqinfoY, iseqY, 0, legnew, 
				       model->cod, dpd, model->null, ali, alignment, logodds, traceback, ends);

      /*if (sweep && sc->visemi->oth < 0.0) return 0;*/

      sc->visemi->rna = ViterbiRNADiag(ofp, sqinfoX, iseqX, sqinfoY, iseqY, 0, legnew, model, dpd, mx, ali, 
				       &scfg_status, &scfgrv_status, alignment, cyk, logodds, d.nus, traceback, ends);
      
      if (sc->visemi->rna->mn > sc->visemi->rna->pl) isrnarev = TRUE;

      cal_structure(ofp, sqinfoX, iseqX, sqinfoY, iseqY, 0, legnew, model->rna, mx, sc->visemi, 
		    cyk, logodds, d.nus, rnass, traceback, ends->rna, isrnarev);
  
      /* add absolute position to ends
       */
      ends->cod->lend += start;
      ends->oth->lend += start;
      ends->rna->lend += start;
      
      ends->cod->rend += start;
      ends->oth->rend += start;
      ends->rna->rend += start;

      fprintf(ofp, "LOCAL_SEMI_VITERBI -- ");
      RNAbanner(ofp, cyk);
      fprintf(ofp, "length alignment: %d (id=%.2f)\n", legnew, id);
      PrintScores(ofp, sc->visemi, ends);    
    }

  if (d.frsemi) 
    {
      sc->frsemi->oth = ForwardOTHDiag(ofp, iseqX, iseqY, 0, legnew, model->oth, 
				       dpd->oth, sc->ardiag->oth);
      /*if (sweep && sc->frsemi->oth < -300.0) return 0;*/

      sc->frsemi->cod = ForwardCODDiag(ofp, iseqX, iseqY, 0, legnew, model->cod, dpd, model->null, 
				       logodds, sc->ardiag);
      sc->frsemi->rna = ForwardRNADiag(ofp, sqinfoX, iseqX, sqinfoY, iseqY, 0, legnew, model, dpd, mx, sc->ardiag, 
				       &scfg_status, &scfgrv_status, cyk, logodds, d.nus);

      fprintf(ofp, "LOCAL_SEMI_FORWARD -- ");
      RNAbanner(ofp, cyk);
      PrintScores(ofp, sc->frsemi, ends);
    }

  *ret_legnew = legnew;
  return 1;
}

int
score_full(FILE *ofp, 
	   SQINFO sqinfoX, int *iseqX, int startX, int lenX, 
	   SQINFO sqinfoY, int *iseqY, int startY, int lenY,
	   int start, int *ret_legnew, int Lw, 
	   struct dos_s      d, 
	   struct model_s   *model, 
	   struct dpf_s     *dpf, 
	   struct rnascfg_s *mx, 
	   struct scores_s  *sc, 
	   struct ali_s     *ali, 
	   int alignment, int cyk, int logodds, int rnass, int sweep, int traceback, struct end3_s *ends)
{
  int legnew;     /* length of new aligmnt generated by full viterbi  */

 /* apply full-viterbi/full-froward according to the three  models
   */
  if (d.vifull) {
    if (!d.visemi)
      sc->vifull->oth = ViterbiOTHFull(ofp, sqinfoX, iseqX, sqinfoY, iseqY, 
				       startX, startY, lenX, lenY, &legnew, Lw, 
				       model->oth, dpf->oth, 
				       ali, alignment, sweep, traceback, ends->oth);
    else {
      sc->vifull->oth = sc->visemi->oth;
      legnew = *ret_legnew;
    }

  /* this one compares the different ways of calculating viterbiOTHFull
   */
  
   if (FALSE) 
      CompareOTHFull(ofp, sqinfoX, iseqX, sqinfoY, iseqY, startX, lenX, startY, lenY, 
		     &legnew, Lw, model->cod->COJ, dpf->oth, ali, ends->oth);

    sc->vifull->cod = ViterbiCODFull(ofp, sqinfoX, iseqX, sqinfoY, iseqY,  
				     startX, startY, lenX, lenY, &legnew, Lw,
				     model->cod, dpf, model->null, 
				     ali, alignment, logodds, traceback, ends);
     
    sc->vifull->rna->pl = 0.0;
    sc->vifull->rna->mn = 0.0;
    
    /* add absolute position to ends
     */
    ends->cod->lend += start;
    ends->oth->lend += start;
    ends->rna->lend += start;
    
    ends->cod->rend += start;
    ends->oth->rend += start;
    ends->rna->rend += start;
    
    fprintf(ofp, "LOCAL_FULL_VITERBI --under construction -- ");
    RNAbanner(ofp, cyk);
    fprintf(ofp, "length of alignment: %d\n", legnew);
    PrintScores(ofp, sc->vifull, ends);
  }
  
  if (d.frfull) {
    sc->frfull->oth = ForwardOTHFull(ofp, iseqX, iseqY, startX, startY, lenX, lenY, Lw, model->oth, 
				    dpf->oth, sc->arfull->oth);

    sc->frfull->cod->pl = 0.0;
    sc->frfull->cod->mn = 0.0;
    sc->frfull->rna->pl = 0.0;
    sc->frfull->rna->mn = 0.0;
    
    fprintf(ofp, "LOCAL_FULL_FORWARD -- under construction -- ");
    RNAbanner(ofp, cyk);
    PrintScores(ofp, sc->frfull, ends);
  }
  *ret_legnew = legnew;
  return 1;
}

void
test_OTH_model(FILE *ofp, 
	       SQINFO sqinfoX, int *isegX, 
	       SQINFO sqinfoY, int *isegY, 
	       int win,
	       struct othmodel_s *oth, 
	       struct othdpd_s   *othdp, 
	       double            *othsc, 
	       struct ali_s      *ali, 
	       int alignment, int traceback, struct end_s *othends)
{
  TestViterbiOTH(ofp, sqinfoX, isegX, sqinfoY, isegY, 0, win, oth, othdp, 
		 ali, alignment, traceback, othends);
  TestForwardOTH(ofp, isegX, isegY, 0, win, oth, othdp, othsc);
  CompareOTH(ofp, sqinfoX, isegX, sqinfoY, isegY, 0, win, oth, othdp, 
	     othsc, ali, alignment, traceback);
}









