/************************************************************
 * QRNA - Comparative analysis of biological sequences 
 *        with pair hidden Markov models and pair stochastic context-free grammars
 * Copyright (C) 2000-Howard Hughes Medical Institute
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

#ifndef STRUCTSH_INCLUDED
#define STRUCTSH_INCLUDED

/* structs.h
 * header file for the three models (other, conding, rna).
 * Defines structures and macros.
 *
 * ER, Fri May 28 11:36:57 CDT 1999 [STL]
 *
 * 
 */ 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

#include "squid.h"
#include "globals.h"

struct ali_s { /* arrays for an alignment of two seqs */
  char *charX;
  char *charY; /* arrays for symbols */
};

struct dos_s {                        /* list of all possible combinations of scoring systems */
  int frdiag;   /* TRUE  ==  calculates forward diagonal           */
  int frfull;   /* TRUE  ==  calculates forward full               */
  int frsemi;   /* TRUE  ==  calculates forward semi-full          */
  int fulldp;   /* TRUE  ==  calculates forward full               */
  int global;   /* TRUE  ==  calculates global alignments          */
  int nus;      /* TRUE  ==  uses NUS model for rna                */
  int semidp;   /* TRUE  ==  calculates semi-full dp scores        */
  int vidiag;   /* TRUE  ==  calculates viterbi diagonal           */
  int vifull;   /* TRUE  ==  calculates viterbi full               */
  int visemi;   /* TRUE  ==  calculates viterbi semi-full          */
  int forward;  /* TRUE  ==  calculates forward dp scores          */
  int twodiag;  /* TRUE  ==  both Forward and Viterbi diagonal dp  */
  int twofull;  /* TRUE  ==  both Forward and Viterbi full dp      */
  int twosemi;  /* TRUE  ==  both Forward and Viterbi semi-full dp */
};

struct dpd_s { /* arrays for DP matrices -  diagonal version */
  struct coddp_s  *cod;
  struct othdpd_s *oth;
  struct rnadp_s  *rna;
};

struct dpf_s { /* arrays for DP matrices -  full version */
  struct coddp_s  *cod;
  struct othdpf_s *oth;
  struct rnadp_s  *rna;
};

struct end_s { 
  int lend;
  int rend;
};

struct end3_s { 
  struct end_s  *cod; /* the ends for COD */
  struct end_s  *oth; /* the ends for OTH */
  struct end_s  *rna; /* the ends for RNA */
};

struct model_s { /* arrays for the three models + null model */
  struct codmodel_s  *cod; /* the model for COD */
  struct othmodel_s  *oth; /* the model for OTH */
  struct rnamodel_s  *rna; /* the model for RNA */

  struct nullmodel_s *null;
};

struct dpsc3_s { /* arrays for forward scores */
  double *oth;
  double *cod;
  double *rna;
};

struct sc2_s { /* to display plus and minus strand scores for a given model */
  double pl;
  double mn;
};
struct sc3_s { /* to display the three model scores */
  struct sc2_s *oth;
  struct sc2_s *cod;
  struct sc2_s *rna;
};


struct scores_s { /* to display all possible scores */
  double null;                  /* null model score (given   strands)                     */
  double nullrev;               /* null model score (reverse strands)                     */
  
  struct sc3_s   *global;       /* global    dp LOD scores                                */
  struct sc3_s   *vidiag;       /* diagonal  dp LOD scores (viterbi)                      */
  struct sc3_s   *frdiag;       /* diagonal  dp LOD scores (forward)                      */
  struct sc3_s   *vifull;       /* full      dp LOD scores (viterbi)                      */
  struct sc3_s   *frfull;       /* full      dp LOD scores (forward)                      */
  struct sc3_s   *visemi;       /* semi-full dp LOD scores (viterbi)                      */
  struct sc3_s   *frsemi;       /* semi-full dp LOD scores (forward)                      */
  
  struct dpsc3_s *ardiag;       /* arrays to store scores for forward calculation         */
  struct dpsc3_s *arfull;       /* arrays to store scores for forward calculation         */
};

/**********************************************************************
 * Single-nucleotide probabilities. [To set the base composition of the different models]
 **********************************************************************/
struct singlep_s {
  double pa;                 /* p[a]  */
  double pc;                 /* p[c]  */
  double pg;                 /* p[g]  */
};
extern struct singlep_s  singlep;     /* OTH model single-emission probs */


/**********************************************************************
 * NULL-model
 **********************************************************************/
struct nullparam_s {
  double eta;                 /* S->N
                                 N->B  */
};

struct nullmodel_s {
  double  eta;          /* transition prob's.  t = eta         +*/
  double  meta;         /* transition prob's.  t = 1-eta        +*/
  double *xem;          /* RX insert emissions. xem[0..3]       +*/
  double *yem;          /* RY insert emissions. yem[0..3]       +*/
};

extern struct nullparam_s Nullparam; /* parameter for the null model */

/**********************************************************************
 * OTH-model
 **********************************************************************/
/* STATES othmodel 
 */
#define stFL      0
#define stB       1
#define stM       2
#define stX       3
#define stY       4
#define stE       5
#define stFJ      6
#define stFR      7
#define OSTATES   8

/* Indices for OTH model state transitions.
 * Used for indexing oth->t[]
 * mnemonic: Transition from Match to Match = TMM
 */
#define TFLB     0
#define TFLFR    1
#define TBM      2
#define TBX      3
#define TBY      4
#define TMM      5
#define TMX      6
#define TMY      7
#define TME      8
#define TXM      9
#define TXX     10
#define TXY     11
#define TXE     12
#define TYM     13
#define TYX     14
#define TYY     15
#define TYE     16
#define TEFJ    17
#define TEFR    18
#define TFJB    19
#define TFRFR   20
#define OTRANS  21

extern int   EmitsPerOTHState[OSTATES];    /* allowed emissions per state in the othmodel       */
extern int   IdxTransOTH[OSTATES];         /* starting point in transition indexing             */
extern int   OConnects[OSTATES][OSTATES];  /* allowed transitions in the othmodel               */
extern char *ostNAME[OSTATES];             /* ASCII states     names                            */
extern char *otrNAME[OTRANS];              /* ASCII transition names                            */
extern int   TransPerOTHState[OSTATES];    /* non-zero transitions per state in the othmodel    */

/* Structures for othmodel 
 */
struct othparam_s {
  struct nullparam_s FLN; /* parameter for the flanking left   null model */
  struct nullparam_s FJN; /* parameter for the flanking middle null model */
  struct nullparam_s FRN; /* parameter for the flanking right  null model */

  double delta;                 /* M->X|Y         transitions   */
  double epsilon;               /* X->X   
				   Y->Y           transitions   */
  double eta;                   /* E->F_R         transitions   */
  double gamma;                 /* X->Y
				   Y->X           transitions   */
  double kappa;                 /* B->X|Y         transitions   */
  double tau;                   /* M->E 
				   X->E
				   Y->E           transitions   */
  double xi;                    /* B->E           transitions   */
};

extern struct othparam_s  OTHparam;     /* parameter for the othmodel                  */

struct othmodel_s {

  /*   Transition probabilities are usually accessed as a
   *   two-D array: oth->t[TMM], for instance. They are allocated
8   *   such that they can also be stepped through in 1D by pointer
   *   manipulations, for efficiency in DP algorithms.
   */
  double *t;            /* transition prob's.  t[0..OTRNAS]    +*/
  double *mem;          /* match    emissions. mem[0..15]      +*/ 
  double *xem;          /* X insert emissions. xem[0..3]       +*/
  double *yem;          /* Y insert emissions. yem[0..3]       +*/

  struct nullmodel_s *FLN; /* Flanking model to the left       +*/
  struct nullmodel_s *FJN; /* Flanking model to the middle     +*/
  struct nullmodel_s *FRN; /* Flanking model to the right      +*/
};

struct othdpd_s {
  double  *flmx;
  double **fjmx;
  double  *frmx;

  double *bmx;
  double *mmx;
  double *xmx;
  double *ymx;
  double *emx;
};

struct othdpf_s {
  double  *flxmx;
  double **fjxmx;
  double  *frxmx;

  double  *flymx;
  double **fjymx;
  double  *frymx;

  double *bmx;
  double *mmx;
  double *xmx;
  double *ymx;
  double *emx;
};

/**********************************************************************
 * COD-model
 **********************************************************************/

/* STATES codmodel 
 */
#define stCOB     0
#define stCb      1
#define stC33     2
#define stC34     3
#define stC32     4
#define stC43     5
#define stC44     6
#define stC42     7
#define stC23     8
#define stC24     9
#define stC22    10
#define stC30    11
#define stC40    12
#define stC20    13
#define stC03    14
#define stC04    15
#define stC02    16
#define stCe     17
#define stCOJ    18
#define stCOE    19
#define CSTATES  20 

/* Indices for COD model state transitions.
 * Used for indexing cod->t[]
 * mnemonic: Transition from Match to Match = TMM
 */
#define TCOBCb    0
#define TCOBCOE   1
#define TCbC33    2
#define TCbC34    3
#define TCbC32    4
#define TCbC43    5
#define TCbC44    6
#define TCbC42    7
#define TCbC23    8
#define TCbC24    9
#define TCbC22   10
#define TCbC30   11
#define TCbC40   12
#define TCbC20   13
#define TCbC03   14
#define TCbC04   15
#define TCbC02   16
#define TCbCe    17
#define TC33Ce   18
#define TC34Ce   19
#define TC32Ce   20
#define TC43Ce   21
#define TC44Ce   22
#define TC42Ce   23
#define TC23Ce   24
#define TC24Ce   25
#define TC22Ce   26
#define TC30Ce   27
#define TC40Ce   28
#define TC20Ce   29
#define TC03Ce   30
#define TC04Ce   31
#define TC02Ce   32
#define TCeCOJ   33
#define TCeCOE   34
#define TCOJCb   35
#define TCOECOE  36
#define CTRANS   37

extern int   EmitsPerCODState[CSTATES];  /* allowed emissions per state in the codmodel     */
extern int   IdxTransCOD[CSTATES];       /* starting point in transiton indexing            */
extern char *cstNAME[CSTATES];           /* ASCII states     names                          */
extern char *ctrNAME[CTRANS];            /* ASCII transition names                          */
extern int   TransPerCODState[CSTATES];  /* non-zero transitions per state in the codmodel  */

/* Structures for codmodel 
 */
struct codparam_s {
  double psi;         /* COB -> Cb   */
  double eta;         /* Ce  -> COE  */
  double tau;         /* Cb  -> Ce   */
  double xi[4][4];    /* Cb  -> Cij  */
};

extern struct othparam_s  COBparam;      /* parameter for the othmodel O_J in codmodel  */
extern struct othparam_s  COJparam;      /* parameter for the othmodel O_J in codmodel  */
extern struct othparam_s  COEparam;      /* parameter for the othmodel O_E in codmodel  */

extern struct codparam_s  CODparam;      /* parameter for the codmodel                  */

struct codmodel_s {
  double  *t;           /* transition prob's.  t[0..OTRNAS]            +*/
  double  *pmut;        /* mutation  prob's. pmut[16]                  +*/
  double **pcodon;      /* codon     emission  prob's. pcodon[64][64]  +*/
  double **phexa;       /* dicodon   prob's. phexa[64][64]             +*/

  struct othmodel_s *COB;
  struct othmodel_s *COJ;
  struct othmodel_s *COE;

};

struct coddp_s {

  double **COJ;

  double  *cbmx;
  double  *c33mx;
  double  *cemx;
  double  *cobmx;
  double  *cojmx;
};

struct pamcond_s {
  double **pXY;
  double **pYX;
};


/**********************************************************************
 * RNA-model
 **********************************************************************/
#define GAPP 0.40     /* for paired-to-a-gap  */
#define INDL 0.60     /* for mutates-to-a-gap */

/* STATES rnamodel 
 */
#define stROB    0
#define stRNA    1
#define stROJ    2
#define stROE    3
#define RSTATES  4 

/* Indices for RNA model state transitions.
 * Used for indexing rna->t[]
 * mnemonic: Transition from Match to Match = TMM
 */
#define TROBRNA  0
#define TROBROE  1
#define TRNAROJ  2
#define TRNAROE  3
#define TROJRNA  4
#define TROEROE  5
#define RTRANS   6

extern int   IdxTransRNA[RSTATES];      /* starting point in transiton indexing            */
extern char *rstNAME[RSTATES];          /* ASCII states     names                          */
extern char *rtrNAME[RTRANS];           /* ASCII transition names                          */
extern int   TransPerRNAState[RSTATES]; /* non-zero transitions per state in the rnamodel  */

struct rnaparam_s {
  double psi;                 /* ROB->RNA */
  double eta;                 /* RNA->ROE */
};

extern struct othparam_s  ROBparam;      /* parameter for the othmodel O_B  in rnamodel  */
extern struct othparam_s  ROJparam;      /* parameter for the othmodel O_J  in rnamodel  */
extern struct othparam_s  ROEparam;      /* parameter for the othmodel O_E  in rnamodel  */
extern struct othparam_s  Rloopparam;    /* parameter for the othmodel loop in rnamodel  */

extern struct rnaparam_s  RNAparam;      /* parameter for the rnamodel                  */

struct rnadp_s {

  double **ROJ;

  double  *rbmx;
  double  *rrmx;
  double  *remx;
  double  *rjmx;
};

struct rnamodel_s {

  double  *t;

  struct othmodel_s *ROB;
  struct othmodel_s *ROJ;
  struct othmodel_s *ROE;

  struct nusmodel_s *nus;
  struct pi2model_s *pi2;
};

struct rnascfg_s {
  double  *sc;    /* to keep the scores */
  double  *vp;    /* to calculate internal loops in L^3 */

  struct rnamtx_s  *in;   /* Inside a lgorithm dp matrices  */
  struct rnamtx_s  *inrv; /* Inside  algorithm dp matrices -- revese strand */

  struct rnamtx_s  *ou;   /* Outside algorithm dp matrices -- only one because I will do the posterior only for the winner strand */
  
  struct nrnscfg_s *nrn;  /* matrices  for non-redundant nussinov (robin's) grammar*/

 };

struct rnamtx_s {
  double **rnaj;  /* to calculate single stranded inside an IS with a geometrical distribution */

  /* the 3 states of the grammar 
   */
  double **vx; 
  double **wx;
  double **wbx;
};

struct nrnscfg_s {
  double **sx;
  double **lx;
  double **rx;
};

/* Structures for PI2model (this is the probabilistic model for rna up to Irred surf of order 2)
 */
struct pi2model_s {

  struct othmodel_s *Rloop;

  struct rna_v5 *v;
  struct rna_w5 *w;
  struct rna_w5 *wb;

  struct rna_loop5 *is1;
  struct rna_loop5 *is2;
};


/* Nussinov model (this is the probabilistic model for rna up to Irred surf of order 0)
 */
struct nusmodel_s {
  double tl;		    /* W->L transition           */
  double tr;		    /* W->R transition           */
  double tv;		    /* W->V transition           */
  double tw;		    /* W->WW bifurcation         */
  double te;		    /* W->e  end in a hairpin    */

  double  p[25];	    /* P(x1,y1) for left-right  singlets   */
  double  pp[25][25];       /* P(x1,y1;x2,y2) for paired emissions */
};

struct rna_w5 {
  double tl;		    /* W->L transition       */
  double tr;		    /* W->R transition       */
  double tv;		    /* W->V transition       */
  double tw;		    /* W->WW bifurcation     */

  double  pl[25];	    /* P(x1,y1) for left singlets          */
  double  pr[25];	    /* P(x1,y1) for right singlets         */
  double  pp[25][25];       /* P(x1,y1;x2,y2) for paired emissions */
};
struct rna_w {
  double tl;		    /* W->L transition       */
  double tr;		    /* W->R transition       */
  double tv;		    /* W->V transition       */
  double tw;		    /* W->WW bifurcation     */

  double  pl[16];	    /* P(x1,y1) for left singlets          */
  double  pr[16];	    /* P(x1,y1) for right singlets         */
  double  pp[16][16];       /* P(x1,y1;x2,y2) for paired emissions */
};
struct rna_v5 { 
  double t1;		          /* V->IS1 transition                 */
  double t2s;		          /* V->IS2 transition  (stems)        */
  double t2b;		          /* V->IS2 transition  (bulges)       */
  double t2i;		          /* V->IS2 transition  (intloops)     */
  double t3;		          /* V->WB WB transition               */

  double pp[25][25];              /* P(i,j) for paired emissions       */
};
struct rna_v { 
  double t1;		          /* V->IS1 transition                 */
  double t2s;		          /* V->IS2 transition  (stems)        */
  double t2b;		          /* V->IS2 transition  (bulges)       */
  double t2i;		          /* V->IS2 transition  (intloops)     */
  double t3;		          /* V->WB WB transition               */

  double pp[16][16];              /* P(i,j) for paired emissions       */
  double p2p[16][16][16][16];     /* P(i,j ; k,l) for 2 pair emissions */
};
struct rna_loop {
  double tn[MAXRNALOOP];
 
  double ps[16];	   /* P(x1,y1) for singlets in loop  */
};
struct rna_loop5 {
  double tn[MAXRNALOOP];
 
  double ps[25];	   /* P(x1,y1) for singlets in loop  */
};


/* non-redundant R.Dowell's nussinov grammar for posteriors
 */
struct nrn_s {
  struct nrnS_s *S;
  struct nrnL_s *L;
  struct nrnR_s *R;
};

struct nrnS_s {
  double   tss;
  double   tsl;
  double   tsr;
  double   tsb;
  double   tse;

  double **pp;
  double  *ps;
};

struct nrnL_s {
  double   tls;
  double   tll;
  double   tle;

  double **pp;
  double  *ps;
};

struct nrnR_s {
  double   trr;
  double   tre;

  double  *ps;
};

/* ungapped
 */
struct rnamodelungapped_s {

  double  *t;

  struct othmodel_s *ROB;
  struct othmodel_s *ROJ;
  struct othmodel_s *ROE;

  struct rna_v *v;
  struct rna_w *w;
  struct rna_w *wb;

  struct rna_loop *is1;
  struct rna_loop *is2;
};

struct nusmodelungapped_s {
  double  *t;

  struct othmodel_s *ROB;
  struct othmodel_s *ROJ;
  struct othmodel_s *ROE;

  struct rna_w *w;
};

struct rna_s {
  double tsl;		     /* W->L transition       */
  double tsr;		     /* W->R transition       */
  double tsp;		     /* W->V transition       */
  double tss;		     /* W->WW bifurcation     */
  double tse;		     /* W->e b                */

  double  px[4][4];	     /* P(x1,y1) for singlets               */
  double  pxy[4][4][4][4];   /* P(x1,y1;x2,y2) for paired emissions */
};

#endif /* STRUCTSH_INCLUDED */
















